!
! Module: ellipse
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines to perform geometric
! manipulations of ellipses.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module ellipse

  use types
  use constants

  implicit none 

  private 

  public initializeEllipseT
  public ellipsePhaseToAngle
  public ellipseAngleToPhase
  public pointInEllipseWithPhase
  public pointInEllipseWithAngle
  public calcEllipseFromAngle
  public calcEllipseFromPhase
  public dumpEllipseT

  interface initializeEllipseT
    module procedure initializeEllipseTScalar
    module procedure initializeEllipseTVector
  end interface

!  interface ellipseAngleToPhase
!    module procedure ellipseAngleToPhaseScalar
!    module procedure ellipseAngleToPhaseVector
!  end interface

  interface dumpEllipseT
    module procedure dumpEllipseTScalar
  end interface

  type, public :: EllipseT
    real(single) ::&
      xAmp,&
      yAmp,&
      phase,&		! radians
      shortSemiAxis,&
      longSemiAxis,&
      rotatedAngle	! radians
    character(10) :: unit	! eg 'pixels', 'detxy', 'tanxy'
    logical(bool) :: isCircle
  end type EllipseT

contains 

  !---------------------------------------------------------------------
  subroutine initializeEllipseTScalar(ellipse)
  !---------------------------------------------------------------------

    implicit none

    type(EllipseT), intent(out) :: ellipse

    type(EllipseT) :: ellipseVector(1)

    !-------------------------------------------------------------------
    ! Start

    call initializeEllipseTVector(ellipseVector)
    ellipse = ellipseVector(1)

  end subroutine initializeEllipseTScalar

  !---------------------------------------------------------------------
  subroutine initializeEllipseTVector(ellipse)
  !---------------------------------------------------------------------

    implicit none

    type(EllipseT), intent(out) :: ellipse(:)

    !-------------------------------------------------------------------
    ! Start

    ellipse%xAmp = 0.0
    ellipse%yAmp = 0.0
    ellipse%phase = 0.0
    ellipse%shortSemiAxis = 0.0
    ellipse%longSemiAxis = 0.0
    ellipse%rotatedAngle = 0.0
    ellipse%unit = 'none'
    ellipse%isCircle = .false.

  end subroutine initializeEllipseTVector

  !---------------------------------------------------------------------
  subroutine ellipsePhaseToAngle(xAmp, yAmp, phase&
  , shortSemiAxis, longSemiAxis, rotatedAngle)
  !---------------------------------------------------------------------
  ! This and the associated subroutine ellipseAngleToPhase() change
  ! the parameterisation of an ellipse. In the 'cartesian' form, the
  ! ellipse is defined by a long and a short semiaxis, plus an angle about
  ! which the whole ellipse is rotated. The short axis is conceived of as
  ! originally aligned with the x axis of the coordinate system, positive
  ! values of rotatedAngle being anticlockwise, ie towards the y axis. In
  ! terms of the 'cartesian' parameters, the points (x,y) on the ellipse
  ! are perhaps best obtained from the following equations:
  !
  !	      x0^2                y0^2
  !	----------------- + ---------------- = 1		1(a)
  !	 shortSemiAxis^2     longSemiAxis^2
  !
  !	x = x0 * cos(rotatedAngle) - y0 * sin(rotatedAngle)	1(b)
  !
  !	y = y0 * cos(rotatedAngle) + x0 * sin(rotatedAngle)	1(c)
  !
  ! The 'phase' form of the ellipse defines (x,y) in terms of a 'dummy'
  ! parameter theta as follows:
  !
  !	x = xAmp * cos(theta)					2(a)
  !
  !	y = yAmp * cos(theta + phase)				2(b)
  !
  ! Each format has its own advantages for certain applications, hence
  ! the usefulness of routines to convert between them.
  !
  !.....................................................................
  !
  ! The ellipsePhaseToAngle transformation is achieved conceptually by
  ! first combining parametric equations 2(a) and 2(b) into a single
  ! equation giving the radius of points on the ellipse as a function of
  ! theta:
  !
  !	r^2 = xAmp^2 * cos^2(theta) + yAmp^2 * cos^2(theta + phase)	3(a)
  !
  ! This equation for r can be visualized on the complex plane as the
  ! real value of an offset rotating sum of two vectors:
  !
  !	r^2 = 0.5 * (xAmp^2 + yAmp^2
  !	    + Re[xAmp^2 * exp{i*2theta}
  !	    + yAmp^2 * exp{i*(2theta + 2phase)}]).			3(b)
  !
  ! Note that the vectors xAmp^2 * exp{i*2theta} and yAmp^2 * exp{i*(2theta
  ! + 2phase)} maintain a fixed phase relationship, the angle between them
  ! being of course 2*phase. Note also that the vectors are offset along
  ! the real axis by xAmp^2 + yAmp^2, which ensures that their real value
  ! of the total can never be less than 0. Equation 3(b) can be further
  ! manipulated to give r in terms of the resultant vector Z * exp{i*(2theta
  ! + alpha)}:
  !
  !	r^2 = 0.5 * (xAmp^2 + yAmp^2
  !	    + Re[Z * exp{i*(2theta + alpha)}]).				4(a)
  !
  ! Here Z, the length of the resultant vector, is given by
  !
  !	Z^2 = xAmp^4 + yAmp^4 + xAmp^2 * yAmp^2 * cos(2phase)	4(b)
  !
  ! (note the powers of 4), and the phase alpha of the resultant is
  ! obtainable from
  !
  !	                 yAmp^2 * sin(2phase)
  !	tan(alpha) = ----------------------------------.		4(c)
  !	              xAmp^2 + yAmp^2 * cos(2phase)
  !
  ! Consider just the vector
  !
  !	Z * exp{i*(2theta + alpha)}
  !	 = xAmp^2 * exp{i*2theta} + yAmp^2 * exp{i*(2theta + 2phase)}
  !
  ! rotating on the complex plane (Argand diagram). Clearly the largest
  ! value of r (equal to the long semiaxis) occurs when 2*theta = -alpha;
  ! conversely the smallest value of r (= the short semiaxis) occurs when
  ! 2*theta = pi - alpha. From equation 3(a), clearly
  !
  !	 longSemiAxis^2 = xAmp^2 * cos^2(-alpha/2)
  !	                + yAmp^2 * cos^2(-alpha/2 + phase)		5(a)
  ! and
  !
  !	shortSemiAxis^2 = xAmp^2 * cos^2(pi/2 - alpha/2)
  !	                + yAmp^2 * cos^2(pi/2 - alpha/2 + phase);	5(b)
  !
  ! However it is probably easier to use equations 2 to get
  !
  !	x_long  = xAmp * cos(-alpha/2)					6(a)
  !
  !	y_long  = yAmp * cos(-alpha/2 + phase)				6(b)
  !
  !	x_short = xAmp * cos(pi/2 - alpha/2)				6(c)
  !
  !	y_short = yAmp * cos(pi/2 - alpha/2 + phase),			6(d)
  !
  ! and thence
  !
  !	shortSemiAxis^2 = x_short * x_short + y_short * y_short		7(a)
  !
  !	longSemiAxis^2  = x_long  * x_long  + y_long  * y_long		7(b)
  !
  ! because then also
  !
  !	rotatedAngle = atan2(y_short, x_short).				8
  !
  !.....................................................................

    implicit none

    real(single), intent(in)  :: xAmp, yAmp, phase
    real(single), intent(out) :: longSemiAxis, shortSemiAxis, rotatedAngle

    real(single) :: xAmp2, yAmp2, alpha, xShort, yShort, xLong, yLong

    !-------------------------------------------------------------------
    ! Start

    xAmp2 = xAmp * xAmp
    yAmp2 = yAmp * yAmp
    alpha = atan2(yAmp2 * sin(2.0 * phase), xAmp2 + yAmp2 * cos(2.0 * phase))
    xLong  = xAmp * cos(-alpha / 2.0)
    yLong  = yAmp * cos(-alpha / 2.0 + phase)
    xShort = xAmp * cos(PiOn2 - alpha / 2.0)
    yShort = yAmp * cos(PiOn2 - alpha / 2.0 + phase)

    shortSemiAxis = sqrt(xShort * xShort + yShort * yShort)
    longSemiAxis  = sqrt(xLong  * xLong  + yLong  * yLong)

    rotatedAngle = atan2(yShort, xShort)

  end subroutine ellipsePhaseToAngle

  !---------------------------------------------------------------------
  subroutine ellipseAngleToPhase(shortSemiAxis, longSemiAxis, rotatedAngle&
  , xAmp, yAmp, phase)
  !---------------------------------------------------------------------
  ! This and the associated subroutine ellipsePhaseToAngle() change
  ! the parameterisation of an ellipse. In the 'cartesian' form, the
  ! ellipse is defined by a long and a short semiaxis, plus an angle about
  ! which the whole ellipse is rotated. The short axis is conceived of as
  ! originally aligned with the x axis of the coordinate system, positive
  ! values of rotatedAngle being anticlockwise, ie towards the y axis. In
  ! terms of the 'cartesian' parameters, the points (x,y) on the ellipse
  ! are perhaps best obtained from the following equations (repeated from
  ! ellipsePhaseToAngle()):
  !
  !	      x0^2                y0^2
  !	----------------- + ---------------- = 1		1(a)
  !	 shortSemiAxis^2     longSemiAxis^2
  !
  !	x = x0 * cos(rotatedAngle) - y0 * sin(rotatedAngle)	1(b)
  !
  !	y = y0 * cos(rotatedAngle) + x0 * sin(rotatedAngle)	1(c)
  !
  ! The 'phase' form of the ellipse defines (x,y) in terms of a 'dummy'
  ! parameter theta as follows:
  !
  !	x = xAmp * cos(theta)					2(a)
  !
  !	y = yAmp * cos(theta + phase)				2(b)
  !
  ! Each format has its own advantages for certain applications, hence
  ! the usefulness of routines to convert between them.
  !
  !.....................................................................
  !
  ! The ellipseAngleToPhase transformation is achieved conceptually by
  ! first changing equation 1(a) to parametric form, giving
  !
  !	x0(beta) = shortSemiAxis * cos(beta)			9(a)
  !
  !	y0(beta) =  longSemiAxis * sin(beta).			9(a)
  !
  ! Thus equations 1(b) and 1(c) become
  !
  !	x(beta) = shortSemiAxis * cos(beta) * cos(rotatedAngle)
  !	        -  longSemiAxis * sin(beta) * sin(rotatedAngle)	10(a)
  !
  !	y(beta) =  longSemiAxis * sin(beta) * cos(rotatedAngle)
  !	        + shortSemiAxis * cos(beta) * sin(rotatedAngle)	10(b)
  !
  ! Differentiating these yields
  !
  !	                    longSemiAxis
  !	tan(beta_xAmp) = - --------------- tan(rotatedAngle)	11(a)
  !	                    shortSemiAxis
  !
  !	                    longSemiAxis
  !	tan(beta_yAmp) =   --------------- cot(rotatedAngle),	11(a)
  !	                    shortSemiAxis
  !
  ! where beta_xAmp is the value of beta at which x = +-xAmp and beta_yAmp
  ! is the value of beta at which y = +-yAmp. Substitution of beta_xAmp
  ! and beta_yAmp into equations 10 provides the values respectively of
  ! xAmp and yAmp.
  !
  ! It can also be shown (again, how I am not sure!) that
  !
  !	phase = (beta_yAmp - beta_xAmp) modulo pi.		12
  !
  !.....................................................................

    implicit none

    real(single), intent(in)  :: longSemiAxis, shortSemiAxis, rotatedAngle
    real(single), intent(out) :: xAmp, yAmp, phase

    real(single) :: beta_xAmp, beta_yAmp

    !-------------------------------------------------------------------
    ! Start

    beta_xAmp = atan2(-longSemiAxis * sin(rotatedAngle)&
    , shortSemiAxis * cos(rotatedAngle))

    beta_yAmp = atan2(longSemiAxis * cos(rotatedAngle)&
    , shortSemiAxis * sin(rotatedAngle))

    xAmp = abs(shortSemiAxis * cos(beta_xAmp) * cos(rotatedAngle)&
              - longSemiAxis * sin(beta_xAmp) * sin(rotatedAngle))

    yAmp = abs(longSemiAxis * sin(beta_yAmp) * cos(rotatedAngle)&
            + shortSemiAxis * cos(beta_yAmp) * sin(rotatedAngle))

    phase = modulo(beta_yAmp - beta_xAmp, real(Pi))

  end subroutine ellipseAngleToPhase

  !---------------------------------------------------------------------
  function pointInEllipseWithPhase(x, y, xAmp, yAmp, phase)
  !---------------------------------------------------------------------
  ! The ellipse is assumed to be centred about the origin. It is defined
  ! by two amplitudes and a phase.
  !
  ! The function returns -1 if the point (x, y) lies fully within the
  ! ellipse; 0 if it is on the border; and 1 if it is fully outside the
  ! ellipse.
  !.....................................................................

    implicit none

    real(single), intent(in) :: x,&
                                y,&
                                xAmp,&
                                yAmp,&
                                phase

    integer :: pointInEllipseWithPhase
    real(single) :: relX, termA, termB, topY, botY

    !-------------------------------------------------------------------
    ! Start

    relX = x / xAmp

    if (abs(relX) > 1.0) then
      pointInEllipseWithPhase = 1
  return
    else if (abs(relX) == 1.0) then
      pointInEllipseWithPhase = 0
  return
    end if

    termA = relX * cos(phase)
    termB = sqrt(1 - relX * relX) * sin(phase)
    topY = yAmp * (termA + termB)
    botY = yAmp * (termA - termB)

    if (y > topY .or. y < botY) then
      pointInEllipseWithPhase = 1
    else if (y == topY .or. y == botY) then
      pointInEllipseWithPhase = 0
    else
      pointInEllipseWithPhase = -1
    end if

  end function pointInEllipseWithPhase

  !---------------------------------------------------------------------
  function pointInEllipseWithAngle(x, y, xSemiAxis, ySemiAxis, rotatedAngle)
  !---------------------------------------------------------------------
  ! The ellipse is assumed to be centred about the origin. It is defined
  ! by two semi-axes. (The short and long semi-axes are defined respectively
  ! as half the shortest and longest diameters of the ellipse. They are
  ! necessarily orthogonal to each other.) The ellipse is to be thought
  ! of as initially having its semiaxes aligned with the x- and y-axes of
  ! the coordinate system, subsequently having been rotated in the direction
  ! +ve x-axis toward +ve y-axis by an angle rotatedAngle. NOTE that this
  ! angle is must be specified in RADIANS.
  !
  ! The function returns -1 if the point (x, y) lies fully within the
  ! ellipse; 0 if it is on the border; and 1 if it is fully outside the
  ! ellipse.
  !
  ! The algorithm used is to find the coordinates (x1, y1) of the specified
  ! point in the coordinate frame rotated to align with the ellipse
  ! semiaxes, then to insert (x1, y1) into the cartesian equation for
  ! the ellipse in this frame.
  !.....................................................................

    implicit none

    real(single), intent(in) ::&
      x,&
      y,&
      xSemiAxis,&
      ySemiAxis,&
      rotatedAngle

    integer :: pointInEllipseWithAngle
    real(single) :: sinAngle, cosAngle, x1, y1, gauge

    !-------------------------------------------------------------------
    ! Start

    sinAngle = sin(rotatedAngle)
    cosAngle = cos(rotatedAngle)
    x1 = x * cosAngle + y * sinAngle
    y1 = y * cosAngle - x * sinAngle

    gauge = (x1 / xSemiAxis) * (x1 / xSemiAxis)&
    + (y1 / ySemiAxis) * (y1 / ySemiAxis)

    if (gauge .gt. 1.0) then
      pointInEllipseWithAngle = 1
    else if (gauge .eq. 1.0) then
      pointInEllipseWithAngle = 0
    else
      pointInEllipseWithAngle = -1
    end if

  end function pointInEllipseWithAngle

  !---------------------------------------------------------------------
  subroutine calcEllipseFromAngle(xSemi, ySemi, angleDeg, xVals, yVals)
  !---------------------------------------------------------------------
  ! This sub calculates coordinates of samples of an ellipse. The larger
  ! the size of the vectors xVals and yVals, the more accurately is the
  ! ellipse sampled. The intention is, after scaling and addition of
  ! offsets, to send the resulting polynomial to the pgplot routine pgpoly,
  ! viz:
  !
  !   call pgpoly(size(xVals), xVals, yVals)
  !
  ! The ellipse is constructed orthogonal to the x-y axes, then rotated by
  ! angle 'angleDeg'. xSemi is defined as half the diameter of the
  ! (unrotated) ellipse in the direction of the x axis; similarly for ySemi.
  ! The rotation is positive-valued in the direction from +ve x to +ve y
  ! axis.
  !.....................................................................

    implicit none 

    real(single), intent(in)  :: xSemi,&
                                 ySemi,&
                                 angleDeg
    real(single), intent(out) :: xVals(:),&
                                 yVals(size(xVals))

    integer :: point, numPoints
    real(single) :: pointAngle, cosWholeAngle, sinWholeAngle, x, y

    !-------------------------------------------------------------------
    ! Start:

    cosWholeAngle = cos(angleDeg * PiOn180)
    sinWholeAngle = sin(angleDeg * PiOn180)

    numPoints = size(xVals)

    xVals = 0.0
    yVals = 0.0

    do point = 1, numPoints
      pointAngle = (point / real(numPoints)) * TwoPi
      x = xSemi * cos(pointAngle)
      y = ySemi * sin(pointAngle)
      xVals(point) = x * cosWholeAngle - y * sinWholeAngle
      yVals(point) = y * cosWholeAngle + x * sinWholeAngle
    end do

  end subroutine calcEllipseFromAngle

  !---------------------------------------------------------------------
  subroutine calcEllipseFromPhase(xAmp, yAmp, phase, xVals, yVals)
  !---------------------------------------------------------------------
  ! This sub calculates coordinates of samples of an ellipse. The larger
  ! the size of the vectors xVals and yVals, the more accurately is the
  ! ellipse sampled. The intention is, after scaling and addition of
  ! offsets, to send the resulting polynomial to the pgplot routine pgpoly,
  ! viz:
  !
  !   call pgpoly(size(xVals), xVals, yVals)
  !
  !.....................................................................

    implicit none 

    real(single), intent(in)  :: xAmp,&
                                 yAmp,&
                                 phase
    real(single), intent(out) :: xVals(:),&
                                 yVals(size(xVals))

    integer :: point, numPoints
    real(single) :: pointAngle

    !-------------------------------------------------------------------
    ! Start:

    numPoints = size(xVals)

    xVals = 0.0
    yVals = 0.0

    do point = 1, numPoints
      pointAngle = (point / real(numPoints)) * TwoPi
      xVals(point) = xAmp * cos(pointAngle)
      yVals(point) = yAmp * cos(pointAngle + phase)
    end do

  end subroutine calcEllipseFromPhase


  !---------------------------------------------------------------------
  subroutine dumpEllipseTScalar(ellipse, dumpVarMsg)
  !---------------------------------------------------------------------

    implicit none

    type(EllipseT), intent(in) :: ellipse
    character(*), intent(in), optional :: dumpVarMsg

    !-------------------------------------------------------------------
    ! Start

    if (present(dumpVarMsg)) then
      write(*,'("== Dump EllipseT: ",A)') trim(dumpVarMsg)
    else
      write(*,'("== Dump EllipseT: ")')
    end if

    write(*,'("(xAmp,yAmp) : (",ES13.6,", ",ES13.6,")")') ellipse%xAmp, ellipse%yAmp	! real(single)
    write(*,'("phase[radians] : ",F12.6)') ellipse%phase	! real(single)
    write(*,'("(shortSemiAxis,longSemiAxis) : (",ES13.6,", ",ES13.6,")")') &
       ellipse%shortSemiAxis, ellipse%longSemiAxis	! real(single)
    write(*,'("rotatedAngle[radians] : ",F12.6)') ellipse%rotatedAngle	! real(single)
    write(*,'("unit : ",A10)') trim(ellipse%unit)	! character(10)
    write(*,'("isCircle : ",L1)') ellipse%isCircle	! logical(bool)

    write(*,'("=================")')

  end subroutine dumpEllipseTScalar


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module ellipse
