!
! Module: epic_aux
!
! Authors: Masaaki Sakano - SSC/LUX
!          
! Description: A few routines to obtain basic information about EPIC
!   (Ideally they should be supplied by the library supplied by the instrument
!    team, however, they are not at the moment, hence are coded here
!    as a work around).
! 
! Comments: TODO:
! 
! $Id: epic_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module epic_aux

  ! use dal
  use cal
  use types
  use errorHandling

  ! use constants

  implicit none 

  private 

  public getCcdNodeNum
  public inverseCcdNodeNum

  interface getCcdNodeNum
    module procedure getCcdNodeNumInt8, getCcdNodeNumInteger
  end interface

contains 

  !---------------------------------------------------------------------
  function getCcdNodeNumInt8(ccdNum, nodeNum) result(ccdNodeNum)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! Return ccdNodeNum, which is a serial number for ccdNum and nodeNum.
    ! ccdNodeNum is particularly used as a Table name in FITS files
    ! as exposuTabName = 'EXPOSU'//ccdNodeNum .
    !-- Note --
    ! This routine really should be included in some more appropriate package..

    integer(int8), intent(in) :: ccdNum
    integer(int8), intent(in), optional :: nodeNum
    integer(int8) :: nodeNumLocal
    integer(int32) :: ccdNodeNum	! intent(out)

    !-------------------------------------------------------------------
    ! Start:

    nodeNumLocal = 1
    if (present(nodeNum)) nodeNumLocal = nodeNum
    if ((nodeNumLocal .ne. 1) .and. (nodeNumLocal .ne. 2)) then
      write(errStr, '("nodeNum (",I6,") looks wrong in getCcdNodeNum().")') nodeNumLocal
      call warning('strangeNodeNum')
    end if

    ccdNodeNum = 10 * (nodeNumLocal - 1) + ccdNum

  end function getCcdNodeNumInt8


  !---------------------------------------------------------------------
  function getCcdNodeNumInteger(ccdNum, nodeNum) result(ccdNodeNum)
  !---------------------------------------------------------------------
  ! Return ccdNodeNum.  See the comments in getCcdNodeNumInt8().

    implicit none 

    integer, intent(in) :: ccdNum
    integer, intent(in), optional :: nodeNum
    integer :: nodeNumLocal
    integer(int32) :: ccdNodeNum	! intent(out)

    !-------------------------------------------------------------------
    ! Start:

    nodeNumLocal = 1
    if (present(nodeNum)) nodeNumLocal = nodeNum
    if ((nodeNumLocal .ne. 1) .and. (nodeNumLocal .ne. 2)) then
      write(errStr, '("nodeNum (",I6,") looks wrong in getCcdNodeNum().")') nodeNumLocal
      call warning('strangeNodeNum')
    end if

    ccdNodeNum = 10 * (nodeNumLocal - 1) + ccdNum

  end function getCcdNodeNumInteger


  !---------------------------------------------------------------------
  subroutine inverseCcdNodeNum(ccdNodeNum, ccdNum, nodeNum, instrumentId)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! Users should either give instrumentId or set instrumentId in CAL beforehand!
    ! If the given instrumentId is negative, this subroutine treats it as EPN.
    !
    ! Returns ccdNum and nodeNum from ccdNodeNum.
    ! If the instrument is EPN (and nodeNum is given), nodeNum==0 is returned.
    ! See the comment in  getCcdNodeNum()  about ccdNodeNum.
    !-- Note --
    ! This routine really should be included in some more appropriate package..

    integer(int32), intent(in) :: ccdNodeNum
    integer(int8),  intent(out) :: ccdNum
    integer(int8),  intent(out), optional :: nodeNum
    integer(int32), intent(in),  optional :: instrumentId	! if not given, CAL is read.

    integer(int8)  :: nodeNumLocal
    integer(int32) :: instrumentIdLocal

    !-------------------------------------------------------------------
    ! Start:

    !--- Checking the input.
    if ((ccdNodeNum < 1) .or. (ccdNodeNum > 17)) then
      write(errStr, '("ccdNodeNum (",I8,") looks wrong in inverseCcdNodeNum().")') ccdNodeNum
      call warning('strangeCcdNodeNum')	! Should never happen.
    end if

    !--- Setting instrumentId
    if (present(instrumentId)) then
      instrumentIdLocal = instrumentId
    else
      call CAL_getState(instrument=instrumentIdLocal)
    end if

    if (instrumentIdLocal < 0) then
      nodeNumLocal = 0
    else
      select case(instrumentIdLocal)
      case(EPN)
        nodeNumLocal = 0
      case(EMOS1, EMOS2)
        nodeNumLocal =  1
      case default
        call error('badInstrument','in ssclib/epic_aux')
      end select
    end if

    !--- MAIN
    if (nodeNumLocal <= 0) then	! EPN
      ccdNum = ccdNodeNum
    else			! MOS1/2
      ! See getCcdNodeNum() for the algorithm.
      nodeNumLocal = ccdNodeNum/10 + 1	! Integer division
      ccdNum = mod(ccdNodeNum, 10)
    end if

    if (present(nodeNum)) nodeNum = nodeNumLocal

  end subroutine inverseCcdNodeNum


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module epic_aux
