!
! Module: minimizations
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted function minimization routines.
! 
! Comments: TODO:
!*** change y and y(x) data types in newtonsMethod and hybridNewtonsMethod to real(single)
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module minimizations

  use types
  use dss		! for RangeT

  use array_utils	! for getBinNumber, smallestTrueIndex, largestTrueIndex,
			! trapezoidSum, minLoc1d, maxLoc1d.
  use constants		! for GoldenRatio
  use dss_ranges_aux	! for valueIsWithinRange, rangesAreWellFormed
  use ssc_misc		! for fracVal, swap.

  implicit none 

  private 

  public newtonsMethod
  public hybridNewtonsMethod
  public riddersMethod
  public simplexMinimization
  public findXBracketOfMinY
  public findXBracketOfZeroY
  public brentsMethod
  public powellsMethod

  interface findXBracketOfMinY
    module procedure findXBracketOfMinYScalar, findXBracketOfMinYVector
  end interface

  interface findXBracketOfZeroY
    module procedure findXBracketOfZeroYScalar, findXBracketOfZeroYVector
  end interface

  interface brentsMethod
    module procedure brentsMethodScalar, brentsMethodVector
  end interface

!  type :: limitBoundType
!    logical :: exists
!    real(single) :: value
!  end type limitBoundType
!  type :: limitType
!    type(limitBoundType) :: lo, hi
!  end type limitType

contains 

  !---------------------------------------------------------------------
  subroutine newtonsMethod(startX, y, dy, solutionX, status)
  !---------------------------------------------------------------------
  ! This subroutine solves the equation y(x) = 0 by Newton's Method. The
  ! user must construct two functions in the calling module, one of which
  ! returns y and the other dy/dx for a given x. These functions (and their
  ! single argument x) must be of type real. The user should pass their
  ! names in the call to newtonsMethod.

    implicit none

    real(single), intent(in) :: startX
    real(single), intent(out) :: solutionX
    integer, intent(out) :: status

!    interface
!      real(single) function y(x)
!        real(single), intent(in) :: x
!      end function y
!      real(single) function dy(x)
!        real(single), intent(in) :: x
!      end function dy
!    end interface
interface
  real function y(x)
    real, intent(in) :: x
  end function y
  real function dy(x)
    real, intent(in) :: x
  end function dy
end interface

    real(single) :: yValue, yDeriv, x, oldXstep, xStep, myPrecision

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default

    myPrecision = 10.0 ** (-precision(x))

    yValue = y(startX)
    yDeriv = dy(startX)
    if (yDeriv .eq. 0.0) then
      status = -1
  return
    end if
    Xstep = -Yvalue / yDeriv
    if (abs(startX) .gt. 0.0) then
      if (abs(Xstep / startX) .le. myPrecision) then
        solutionX = startX + Xstep
  return
      end if
    end if
    x = startX + Xstep
    oldXstep = Xstep

    do
      yValue = y(x)
      yDeriv = dy(x)
      if (yDeriv .eq. 0.0) then
        status = -1
  return
      end if

      Xstep = -yValue / yDeriv

      if (abs(x) .gt. 0.0) then
        if (abs(Xstep / x) .le. myPrecision) then
          solutionX = x + Xstep
  return
        end if
      end if

      if (abs(Xstep) .gt. abs(oldXstep)) then
        status = -2
  return
      end if

      x = x + Xstep
      oldXstep = Xstep
    end do

  end subroutine newtonsMethod

  !---------------------------------------------------------------------
  subroutine hybridNewtonsMethod(startX0, startX1, y, dy, solutionX&
  , maxNumIterations, convergenceCriterion, status)
  !---------------------------------------------------------------------
  ! Adapted from the hybrid bisection/Newton-Raphson method in Numerical
  ! Methods.
  !
  ! This subroutine solves the equation y(x) = 0 by Newton's Method. The
  ! user must construct two functions in the calling module, one of which
  ! returns y and the other dy/dx for a given x. These functions (and their
  ! single argument x) must be of type real. The user should pass their
  ! names in the call to hybridNewtonsMethod.
  !
  ! The starting values startX0 and startX1 should bracket the zero point
  ! but it is not necessary that they be in any order.

    implicit none

    real(single), intent(in)            :: startX0, startX1
    real(single), intent(out)           :: solutionX
    integer,      intent(in),  optional :: maxNumIterations
    real(single), intent(in),  optional :: convergenceCriterion
    integer,      intent(out), optional :: status

!    interface
!      real(single) function y(x)
!        real(single), intent(in) :: x
!      end function y
!      real(single) function dy(x)
!        real(single), intent(in) :: x
!      end function dy
!    end interface
interface
  real function y(x)
use types
    real, intent(in) :: x
  end function y
  real function dy(x)
    real, intent(in) :: x
  end function dy
end interface

    real(single) ::&
      yValue, yDeriv, localConvergenceCriterion, startY0, startY1,&
      xLo, xHi, dx, dxOld, oldSolutionX
    integer :: localStatus, localMaxNumIter, j

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    solutionX = 0.0 ! Don't like to leave it with no value, even if the routine
    ! can't calculate it.

    if (present(maxNumIterations)) then
      localMaxNumIter = maxNumIterations
    else
      localMaxNumIter = 100
!*** can do better than this.
    end if

    startY0 = y(startX0)
    startY1 = y(startX1)

    if (startY0 == 0.0) then
      solutionX = startX0
  return
    end if
    if (startY1 == 0.0) then
      solutionX = startX1
  return
    end if

    if (startX0 == startX1) then
      if (present(status)) status = -1 ! no space between bounds.
  return
    end if

    if (startY0 > 0.0 .and. startY1 > 0.0&
    .or. startY0 < 0.0 .and. startY1 < 0.0) then
      if (present(status)) status = -2 ! startX0 and
      ! startX1 don't bracket the root.
  return
    end if

    ! Orient bounds such that y(xLo) < 0:
    if (startY0 < 0.0) then
      xLo = startX0
      xHi = startX1
    else
      xLo = startX1
      xHi = startX0
    end if

    solutionX = 0.5 * (xLo + xHi)
    dxOld = xHi - xLo
    dx = dxOld

    if (present(convergenceCriterion)) then
      localConvergenceCriterion = abs(convergenceCriterion)
    else
      localConvergenceCriterion = 0.5 * abs(xLo + xHi)&
      * (10.0 ** (1-precision(dx)))
    end if

    yValue = y(solutionX)
    yDeriv = dy(solutionX)
    do j = 1, localMaxNumIter
      dxOld = dx

      if (((solutionX - xHi) * yDeriv - yValue)&
      * ((solutionX - xLo) * yDeriv - yValue) > 0.0&
      .or. abs(2.0 * yValue) > abs(dxOld * yDeriv)) then
        dx = 0.5 * (xHi - xLo)
        solutionX = xLo + dx
    if (solutionX == xLo) exit

      else
        dx = yValue / yDeriv
        oldSolutionX = solutionX
        solutionX = solutionX - dx
    if (solutionX == oldSolutionX) exit
      end if

    if (abs(dx) < localConvergenceCriterion) exit

      yValue = y(solutionX)
      yDeriv = dy(solutionX)

      if (yValue < 0.0) then
        xLo = solutionX
      else
        xHi = solutionX
      end if
    end do
    if (j >= localMaxNumIter) then
      localStatus = ibset(localStatus, 0)
    end if

    if (present(status)) status = localStatus

  end subroutine hybridNewtonsMethod

  !---------------------------------------------------------------------
  subroutine riddersMethod(startX0, startX1, y, solutionX&
  , maxNumIterations, convergenceCriterion, dx, status)
  !---------------------------------------------------------------------
  ! This subroutine has been adapted from an algorithm described in
  ! Numerical Methods for Fortran. I changed it slightly in that the
  ! convergence test compares the difference between the present and
  ! last values of xNew, rather than the difference between xHi and xLo
  ! as in the original algorithm, to the value of convergenceCriterion.
  !
  ! This subroutine solves the equation y(x) = 0 by Ridders' Method. The
  ! user must construct the function y(x) in the calling module. This
  ! function (and its single argument x) must be of type real. The user
  ! should pass the function name in the call to riddersMethod.
  !
  ! Note that y(startX0) and y(startX1) must be of opposite sign.

    implicit none

    real(single), intent(in)            :: startX0, startX1
    real(single), intent(out)           :: solutionX
    integer,      intent(in),  optional :: maxNumIterations
    real(single), intent(in),  optional :: convergenceCriterion
    real(single), intent(out), optional :: dx
    integer,      intent(out), optional :: status

    interface
      function y(x)
        use types

        real(single), intent(in) :: x
        real(single) :: y
      end function y
    end interface

    integer ::&
      localStatus, j
    real(single) ::&
      startY0, startY1, xLo, yLo, xMid, yMid, xHi, yHi, xOld, xNew, yNew,&
      root, expQ, fracY, localConvergenceCriterion

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default
    if (present(status)) status = localStatus
    solutionX = 0.0 ! Don't like to leave it with no value, even if the routine
    ! can't calculate it.

    dx = 0.0 ! default

    startY0 = y(startX0)
    startY1 = y(startX1)

    if (startY0 == 0) then
      solutionX = startX0
  return
    end if
    if (startY1 == 0) then
      solutionX = startX1
  return
    end if

    if (startX0 == startX1) then
      if (present(status)) status = -1 ! no space between startX0 and startX1.
  return
    end if

    if (startY0 > 0.0 .and. startY1 > 0.0&
    .or. startY0 < 0.0 .and. startY1 < 0.0) then
      if (present(status)) status = -2
  return
    end if

    ! Orient bounds such that y(xLo) < 0:
    if (startY0 < 0.0) then
      xLo = startX0
      yLo = startY0
      xHi = startX1
      yHi = startY1
    else
      xLo = startX1
      yLo = startY1
      xHi = startX0
      yHi = startY0
    end if

    dx = xHi - xLo

    xOld = xLo ! could have also chosen xHi; at this point it is
    ! indeterminate which of these two values is the 'latest estimate'.

    if (present(convergenceCriterion)) then
      localConvergenceCriterion = abs(convergenceCriterion)
    else
      localConvergenceCriterion = 0.5 * abs(startX0 + startX1)&
      * (10.0 ** (1 - precision(dx)))
    end if

    j = 1
    do
      xMid = 0.5 * (xLo + xHi)
      yMid = y(xMid)
      root = sqrt(yMid**2 - yLo * yHi)

      if (root == 0) then ! The argument of the square root can
      ! only equal zero if one at least of yLo and yHi equals zero (and
      ! if yMid is also equal to zero). The routine is supposed to screen out
      ! such cases before this point is reached. This status value indicates
      ! a bug.
        if (present(status)) status = -3
  return
      end if

      ! Solve quadratic equation to get expQ:

      expQ = (yMid + root) / yHi ! That yHi > 0 is checked implicitly
      ! elsewhere in the loop.

      ! Apply false-position method to y(x)*exp(2Q(x-xLo)/(xHi-xLo)):

      fracY = -yLo / (yHi * expQ * expQ - yLo)
      xNew = fracVal(xLo, xHi, fracY)
      yNew = y(xNew)

      if (yNew == 0.0) then
        solutionX = xNew
    exit
      end if

      ! Now decide new x bounds:

      if (yNew < 0.0) then
        xLo = xNew
        yLo = yNew
      else
        xHi = xNew
        yHi = yNew
      end if

      ! Check convergence:

      dx = xNew - xOld
      if (abs(dx) < localConvergenceCriterion) then
        solutionX = xNew
    exit
      end if

      j = j + 1
      if (present(maxNumIterations)) then
        if (j > maxNumIterations) then
          localStatus = ibset(localStatus, 0)
          solutionX = xNew
    exit
        end if
      end if

      xOld = xNew
    end do

    if (present(status)) status = localStatus

  end subroutine riddersMethod

  !---------------------------------------------------------------------
  subroutine simplexMinimization(funk, ftol, simplex, y, numFunctionEvals&
  , maxNumIterationsArg, status)
  !---------------------------------------------------------------------
  ! This implements a multidimensional minimization of the function funk(x), where x(1:numDimensions) is a vector in numDimensions dimensions, by the downhill simplex method of Nelder and Mead. The matrix p(1:N+1,1:N), where N is the number of dimensions, contains the simplex. Each of the N+1 rows of p is an N-dimensional vector: these represent the vertices of the simplex. The calling routine must supply starting values of p, which is therefore of type inout. Also input is the vector y(1:N+1), the components of which must be pre-initialized to the values of funk evaluated at the N+1 vertices (rows) of p; and ftol the fractional convergence tolerance to be achieved in the function value. On output, p and y will have been reset to numDimensions+1 new points, all within ftol of a minimum function value, and iter gives the number of function evaluations taken.
  !
  ! Adapted from Numerical Recipes in Fortran
!**** authors, press, year

    implicit none

    interface
      function funk(x)
        use types

        real(single), intent(in) :: x(:)
        real(single)             :: funk
      end function funk
    end interface

    real(single),   intent(in)            :: ftol
    real(single),   intent(inout)         :: simplex(:,:)
    real(single),   intent(out)           :: y(size(simplex, 1))
    integer(int32), intent(out)           :: numFunctionEvals
    integer,        intent(in),  optional :: maxNumIterationsArg
    integer,        intent(out), optional :: status

    integer :: numDimensions, n, ilo, ihi, inhi, i, maxNumIterations
    real(single) :: rtol, ytry, ysave
    real(single), parameter :: TINY = 1.0e-10
!    integer, parameter :: maxNumIterations=5000

real(single) :: pathLength, simplexSize, stepLength, realNumDims,&
  realNumPoints, var,&
! diff(size(simplex, 2)),&
  currentCentre(size(simplex, 2)), lastCentre(size(simplex, 2))
integer :: dimNum, ptNum
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    numDimensions = size(simplex, 2)

if (testFlag) then
  realNumDims   = real(numDimensions)
  realNumPoints = real(numDimensions+1)
  pathLength = 0.0
  do dimNum = 1, numDimensions
    lastCentre(dimNum) = sum(simplex(:,dimNum)) / realNumPoints
  end do
end if

!*** default values for y?

    if (size(simplex, 1) .ne. numDimensions + 1) then
      if (present(status)) status = -1
  return
    end if

    if (numDimensions < 2) then
      if (present(status)) status = -2
  return
    end if

    if (present(maxNumIterationsArg)) then
      maxNumIterations = maxNumIterationsArg
    else
      maxNumIterations = 5000
    end if

    ! Calculate initial y:
    do i = 1, numDimensions + 1
      y(i) = funk(simplex(i, :)) ! not counted in numFunctionEvals.
    end do

    numFunctionEvals = 0

    do ! start of iteration loop

      ! Determine which point of the simplex is the highest (ie worst), the
      ! next highest, and the lowest (ie best):

      call calcBestWorst(y, ilo, inhi, ihi)

      ! Check for convergence:

      rtol = 2.0 * abs(y(ihi) - y(ilo)) / (abs(y(ihi)) + abs(y(ilo)) + TINY)

      if (rtol < ftol) then ! if returning, put best point and value in slot 1:
        call swap(y(1), y(ilo)) ! in ssc_misc

        do n = 1, numDimensions
          call swap(simplex(1, n), simplex(ilo, n)) ! in ssc_misc
        end do

        if (present(status)) status = 0
  return
      end if

      if (numFunctionEvals >= maxNumIterations) then
!*** do anything with data?
        if (present(status)) status = ibset(status, 0)
  return
      end if

      ! Reflect the position of the point with the highest y value
      ! through the opposite face of the simplex and calculate y there:

      call moveHiPointTrial(simplex, y, funk, ihi, -1.0, ytry)
      numFunctionEvals = numFunctionEvals + 1

      if (ytry <= y(ilo)) then ! gives a result better than the best point,
      ! so try an additional extrapolation by a factor 2:

        call moveHiPointTrial(simplex, y, funk, ihi, 2.0, ytry)
        numFunctionEvals = numFunctionEvals + 1

      else if (ytry >= y(inhi)) then ! the reflected point is worse than
      ! the second-highest, so look for an intermediate lower point, ie do
      ! a 1-dimensional contraction:

        ysave = y(ihi)
        call moveHiPointTrial(simplex, y, funk, ihi, 0.5, ytry)
        numFunctionEvals = numFunctionEvals + 1

        if (ytry >= ysave) then ! can't seem to get rid of that high point.
        ! Better contract around the lowest (best) point:

          do i = 1, numDimensions + 1
          if (i == ilo) cycle
            simplex(i, :) = 0.5 * (simplex(i, :) + simplex(ilo, :)) ! vector op
            y(i) = funk(simplex(i, :))
          end do
          numFunctionEvals = numFunctionEvals + numDimensions ! Keep track
          ! of function iterations.
        end if
      end if

if (testFlag) then
  do dimNum = 1, numDimensions
    currentCentre(dimNum) = sum(simplex(:,dimNum)) / realNumPoints
  end do
  stepLength = sqrt(sum((currentCentre-lastCentre)*(currentCentre-lastCentre)))
  pathLength = pathLength + stepLength
  var = 0.0
  do ptNum = 1, numDimensions+1
    var = var + sum((simplex(ptNum,:)-currentCentre)&
      *(simplex(ptNum,:)-currentCentre))
  end do
  var = var  / realNumDims / realNumPoints
  simplexSize = sqrt(var)

  lastCentre = currentCentre

  print *, numFunctionEvals, sum(y)/realNumPoints, stepLength&
    , pathLength, simplexSize
end if

    end do

  end subroutine simplexMinimization

  !---------------------------------------------------------------------
  subroutine calcBestWorst(y, ilo, inhi, ihi)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: y(:)
    integer,      intent(out) :: ilo, ihi, inhi

!    logical(bool), allocatable :: mask(:)
    logical, allocatable :: mask(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(mask(size(y)))

    ilo  = minLoc1d(y)		! in array_utils
    mask = .true.
    mask(ilo) = .false.
    ihi  = maxLoc1d(y, mask)	! in array_utils
    mask(ihi) = .false.
    inhi = maxLoc1d(y, mask)	! in array_utils

  end subroutine calcBestWorst

  !---------------------------------------------------------------------
  subroutine moveHiPointTrial(simplex, y, funk, ihi, fac, ytry)
  !---------------------------------------------------------------------
  ! Moves the point with the highest function value away from the mean point by the factor fac, tries it the function at this new location, and replaces the high point if the new point is better.

    implicit none

    integer,      intent(in)    :: ihi
    real(single), intent(in)    :: fac
    real(single), intent(inout) :: simplex(:,:), y(size(simplex,1))
    real(single), intent(out)   :: ytry

    interface
      function funk(x)
        use types

        real(single), intent(in) :: x(:)
        real(single)             :: funk
      end function funk
    end interface

    real(single), allocatable :: ptry(:), meanPoint(:)
    integer :: numDimensions, numPoints, dimNum

    !-------------------------------------------------------------------
    ! Start

    numDimensions = size(simplex,2)
    allocate(ptry(numDimensions), meanPoint(numDimensions))

    numPoints = numDimensions + 1
    do dimNum = 1, numDimensions
      meanPoint(dimNum) = sum(simplex(:, dimNum))
    end do
    meanPoint = meanPoint / real(numPoints)

    ptry = ((1.0 - fac) * numPoints * meanPoint&
      + (numPoints * fac - 1.0) * simplex(ihi, :))&
      / real(numDimensions) ! vector op

    ! This expression is derived as follows. Suppose we have a simplex of N+1 points p_i in a space of dimension N. The position of the mean point 'a' is given by a = sum(i=1,N+1)(p_i). Now select one of the simplex points, say point p_j. (The interest in the present routine is to select that which currently gives the highest value of the function funk.) 'Moving the point through the opposite face of the simplex' means to slide it like a bead on a wire along a straight line connecting its initial position and that of the mean point. To specify where we want it to end up, it is most convenient to specify the ratio fac, which is given by
    !
    !	       |p_j_new - a_new|
    !	fac = -------------------,
    !	       |p_j_old - a_old|
    !
    ! or in other words
    !
    !	p_j_new - a_new = fac * p_j_old - fac * a_old.			(1)
    !
    ! To arrive at the formula for p_j_new we should also consider that the new average can be expressed in another way as:
    !
    !	                 p_j_old     p_j_new
    !	a_new = a_old - --------- + ---------.				(2)
    !	                   N+1         N+1
    !
    ! Inserting (2) into (1) gives the desired expression.

    ytry = funk(ptry)
    if (ytry < y(ihi)) then
      y(ihi) = ytry
!      psum = psum - p(ihi, :) + ptry	! vector op
      simplex(ihi, :) = ptry	! vector op
    end if

  end subroutine moveHiPointTrial

  !---------------------------------------------------------------------
  subroutine brentsMethod_old(bestXAtEntry, loBracketX, hiBracketX, solutionX&
  , yFn, maxNumIterations, deltaY, status)
  !---------------------------------------------------------------------
  ! Brent's Method finds a minimum of the function yFn, basically by
  ! means of parabolic interpolation. The three input x values must satisfy
  ! the following conditions:
  !	(i) They must be distinct (ie, no two the same).
  !	(ii) They must be in order either loBracketX<bestX<hiBracketX
  !	     or loBracketX>bestX>hiBracketX.
  !	(iii) yFn(bestX) must be smaller than either yFn(loBracketX)
  !	      or yFn(hiBracketX).
  !
  ! To find out how to employ the optional input deltaY, read the comment
  ! after the first calculation of xTolerance.
  !
  ! This routine was adapted from that at the end of chapter 5, Brent R P,
  ! "Algorithms for Minimization without Derivatives", Prentice-Hall, 1973.
  ! The version in chapter 10 of Press W H et al, "Numerical Recipes in
  ! Fortran", 2nd Ed, was also consulted.
  !---------------------------------------------------------------------

    real(single), intent(in)            :: bestXAtEntry
    real(single), intent(inout)         :: loBracketX,&
                                           hiBracketX
    real(single), intent(out)           :: solutionX
    integer,      intent(in),  optional :: maxNumIterations
    real(single), intent(in),  optional :: deltaY
    integer,      intent(out), optional :: status

    interface
      function yFn(x)
        use types

        real(single), intent(in) :: x
        real(single) :: yFn
      end function yFn
    end interface

    integer :: localMaxNumIter, numIter
    real(single) ::&
      localDeltaY, midBracketX, bestX, secondBestX, parabolicFitX,&
      prevSecondBestX, bestY, secondBestY, prevSecondBestY, step,&
      previousStep, twoStepsAgo, yAtLoX, yAtHiX, curvature,&
      xTolSquared, xTolerance, xTolTimes2, mostRecentX, mostRecentY,&
      yAtFittedX
    logical :: parabolicFitOk

    !-------------------------------------------------------------------
    ! Start

    if (present(status)) status = 0
    solutionX = 0.0

    if (loBracketX == bestXAtEntry .or. bestXAtEntry == hiBracketX) then
    ! This is not allowed. The routine is only guaranteed to converge if
    ! the value of the function YFn() at bestXAtEntry is less than its
    ! value at either end of the interval. In a moment we'll test for this
    ! explicitly, but a break of this rule is implied if the claimed best
    ! x value is located at one of the boundaries.
      if (present(status)) status = -1
  return
    end if

    if ((bestXAtEntry - loBracketX) / (hiBracketX - bestXAtEntry) < 0.0)&
    then
    ! The starting X values aren't in either increasing or decreasing
    ! order. They must be one or the other for the routine to work.
      if (present(status)) status = -2
  return
    end if

    YAtLoX = yFn(loBracketX)
    bestY  = yFn(bestXAtEntry)
    YAtHiX = yFn(hiBracketX)

    if (bestY >= YAtLoX .or. bestY >= YAtHiX) then
    ! This is not allowed. The routine is only guaranteed to converge if
    ! the value of the function YFn() at bestXAtEntry is less than its
    ! value at either end of the interval.
      if (present(status)) status = -3
  return
    end if

    if (hiBracketX < loBracketX) then
      call swap(loBracketX, hiBracketX)
      call swap(YAtLoX, YAtHiX)
    end if

    if (present(maxNumIterations)) then
      localMaxNumIter = maxNumIterations
    else
      localMaxNumIter = 100
    end if

    if (present(deltaY)) then
      localDeltaY = abs(deltaY)
    else
      localDeltaY = 0.0
    end if

    bestX = bestXAtEntry

    if (YAtLoX < YAtHiX) then
      secondBestX     = loBracketX
      secondBestY     = YAtLoX
      prevSecondBestX = hiBracketX
      prevSecondBestY = YAtHiX
    else
      secondBestX     = hiBracketX
      secondBestY     = YAtHiX
      prevSecondBestX = loBracketX
      prevSecondBestY = YAtLoX
    end if

    ! Perform an initial parabolic interpolation in order to estimate
    ! the x tolerance. Note that a +ve-curvature parabolic fit _must_
    ! succeed if the criteria for the initial triplet of (x,y) values
    ! are fulfilled.

    call fitParabola(bestX, secondBestX, prevSecondBestX&
      , bestY, secondBestY, prevSecondBestY, parabolicFitOk&
      , curvature, parabolicFitX)

    if (.not. parabolicFitOk) then
      if (present(status)) status = -4
  return
    end if

    if (curvature <= 0.0) then
      if (present(status)) status = -5
  return
    end if

    if (parabolicFitX <= loBracketX .or. parabolicFitX >= hiBracketX) then
      if (present(status)) status = -6
  return
    end if

    previousStep = 0.0
    step = parabolicFitX - bestX
    yAtFittedX = yFn(parabolicFitX)

    !                            a            b           u
    call doXAndYHouseKeeping(loBracketX, hiBracketX, parabolicFitX&
    !     x         w              v
      , bestX, secondBestX, prevSecondBestX, yAtFittedX, bestY&
      , secondBestY, prevSecondBestY)

    xTolSquared = (epsilon(1.0) * abs(bestY) + localDeltaY + tiny(1.0))&
      / curvature
    xTolerance = sqrt(xTolSquared)
    ! This formula arises as follows. There is always some uncertainty
    ! DY in the value of the function y(x). This is related to the
    ! uncertainty DX in x near the minimum x0 by
    !
    !	      y''(x0)
    !	DY = --------- DX * DX + (order > 2 terms).
    !	         2
    !
    ! If we neglect the higher-order terms then we can invert this to
    ! give
    !
    !	         ( 2 * DY  )
    !	DX = sqrt(---------).
    !	         ( y''(x0) )
    !
    ! y''(x0) / 2 is just the curvature returned by the parabolic fit.
    ! We analyse DY into three terms: (i) the floating-point machine
    ! precision epsilon*y(x0); (ii) an optional intent(in) component
    ! deltaY; (iii) and lastly, a small number, to prevent DX going to
    ! zero in the case that y(x0) == 0 and no deltaY was supplied.

    xTolTimes2 = 2.0 * xTolerance

    numIter = 1
    do ! until convergence
      midBracketX = 0.5 * (loBracketX + hiBracketX)

      ! Test for convergence:      
      if (abs(bestX - midBracketX)&
      <= (xTolTimes2 - 0.5 * (hiBracketX - loBracketX)))&
    exit

      ! The purpose of the next block is mainly to generate new values of
      ! step and previousStep. The tolerances may also get recalculated.

      if (abs(previousStep) <= xTolerance) then ! should always be true at
      ! first iteration.
        call goldenSection(bestX, midBracketX, loBracketX, hiBracketX, step&
          , previousStep)
        ! This sub doesn't read step and previousStep, but only writes them.

      else ! construct a trial parabolic fit
        call fitParabola(bestX, secondBestX, prevSecondBestX, bestY&
          , secondBestY, prevSecondBestY, parabolicFitOk, curvature&
          , parabolicFitX)

        twoStepsAgo  = previousStep
        previousStep = step
        step         = parabolicFitX - bestX ! This last is meaningless
        ! if either (.not. parabolicFitOk .or. curvature >= 0.0), but
        ! it is harmless, since parabolicFitX is always returned with
        ! a defined value. If the golden section routine is used, this
        ! sets its own value of step without reading the preexisting
        ! value.

        if (.not. parabolicFitOk .or. curvature <= 0.0&
        .or. parabolicFitX <= loBracketX .or. parabolicFitX >= hiBracketX&
        .or. abs(step) >= abs(twoStepsAgo / 2.0))&
        then
          call goldenSection(bestX, midBracketX, loBracketX, hiBracketX, step&
            , previousStep)
          ! This sub doesn't read step and previousStep, but only writes them.

        else ! at this point, basically, you can use the parabolic fit.
        ! We'll check however to see if the new X is too close to the
        ! bracket boundaries. If so, then we'll replace 'step' with a
        ! nominal value.

          if ( parabolicFitX < loBracketX + xTolerance&
          .or. parabolicFitX > hiBracketX - xTolerance) then
            if (bestX < midBracketX) then
              step =  xTolerance
            else
              step = -xTolerance
            end if
          end if

          xTolSquared = (epsilon(1.0) * abs(bestY) + localDeltaY&
            + tiny(1.0)) / curvature
          xTolerance = sqrt(xTolSquared)
          xTolTimes2 = 2.0 * xTolerance
        end if ! .not. parabolicFitOk
      end if ! abs(previousStep) <= xTolerance

      if (abs(step) < xTolerance) then
        if (step > 0.0) then
          step =  xTolerance
        else
          step = -xTolerance
        end if
      end if

      mostRecentX = bestX + step
      mostRecentY = yFn(mostRecentX) ! this is the one function
      ! evaluation per iteration.

      call doXAndYHouseKeeping(loBracketX, hiBracketX, mostRecentX&
        , bestX, secondBestX, prevSecondBestX, mostRecentY, bestY&
        , secondBestY, prevSecondBestY)
      
      numIter = numIter + 1
      if (numIter > localMaxNumIter) then
        if (present(status)) status = -7
  return
      end if
    end do

    solutionX = bestX

  end subroutine brentsMethod_old

  !---------------------------------------------------------------------
  subroutine brentsMethodScalar(bestXAtEntry, loBracketX, hiBracketX, solutionX&
  , yFn, maxNumIterations, deltaY, status)
  !---------------------------------------------------------------------

    real(single), intent(in)            :: bestXAtEntry
    real(single), intent(inout)         :: loBracketX,&
                                           hiBracketX
    real(single), intent(out)           :: solutionX
    integer,      intent(in),  optional :: maxNumIterations
    real(single), intent(in),  optional :: deltaY
    integer,      intent(out), optional :: status

    interface
      function yFn(x)
        use types

        real(single), intent(in) :: x(:)
        real(single) :: yFn
      end function yFn
    end interface

    real(single) :: dummyX0(1), dummyDir(1)

    !-------------------------------------------------------------------
    ! Start

    dummyX0  = 0.0
    dummyDir = 1.0
    call brentsMethodVector(dummyX0, dummyDir, bestXAtEntry, loBracketX&
      , hiBracketX, solutionX, yFn, maxNumIterations, deltaY, status)

  end subroutine brentsMethodScalar

  !---------------------------------------------------------------------
  subroutine brentsMethodVector(x0, dir, bestXAtEntry, loBracketX, hiBracketX&
  , solutionX, yFn, maxNumIterations, deltaY, status)
  !---------------------------------------------------------------------
  ! Brent's Method finds a minimum of the function yFn, basically by
  ! means of parabolic interpolation. The three input x values must satisfy
  ! the following conditions:
  !	(i) They must be distinct (ie, no two the same).
  !	(ii) They must be in order either loBracketX<bestX<hiBracketX
  !	     or loBracketX>bestX>hiBracketX.
  !	(iii) yFn(bestX) must be smaller than either yFn(loBracketX)
  !	      or yFn(hiBracketX).
  ! These conditions can be assured by running findXBracketOfMinY before
  ! calling brentsMethod.
  !
  ! To find out how to employ the optional input deltaY, read the comment
  ! after the first calculation of xTolerance.
  !
  ! This routine was adapted from that at the end of chapter 5, Brent R P,
  ! "Algorithms for Minimization without Derivatives", Prentice-Hall, 1973.
  ! The version in chapter 10 of Press W H et al, "Numerical Recipes in
  ! Fortran", 2nd Ed, was also consulted.
  !---------------------------------------------------------------------

    real(single), intent(in)            :: x0(:), dir(size(x0)), bestXAtEntry
    real(single), intent(inout)         :: loBracketX,&
                                           hiBracketX
    real(single), intent(out)           :: solutionX
    integer,      intent(in),  optional :: maxNumIterations
    real(single), intent(in),  optional :: deltaY
    integer,      intent(out), optional :: status

    interface
      function yFn(x)
        use types

        real(single), intent(in) :: x(:)
        real(single) :: yFn
      end function yFn
    end interface

    integer :: localMaxNumIter, numIter
    real(single) ::&
      localDeltaY, midBracketX, bestX, secondBestX, parabolicFitX,&
      prevSecondBestX, bestY, secondBestY, prevSecondBestY, step,&
      previousStep, twoStepsAgo, yAtLoX, yAtHiX, curvature,&
      xTolSquared, xTolerance, xTolTimes2, mostRecentX, mostRecentY,&
      yAtFittedX
    logical :: parabolicFitOk

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '          aaa brentsMethodVector'
if (testFlag) print *, '          - maxval(x0):', maxval(x0)
if (testFlag) print *, '          - maxval(dir):', maxval(dir)
if (testFlag) print *, '          - bestXAtEntry, loBracketX, hiBracketX:', bestXAtEntry, loBracketX, hiBracketX

    if (present(status)) status = 0
    solutionX = 0.0

    if (size(x0) <= 0) then
      if (present(status)) status = -8
  return
    end if

    if (all(dir == 0.0)) then
      if (present(status)) status = -9
  return
    end if

    if (loBracketX == bestXAtEntry .or. bestXAtEntry == hiBracketX) then
    ! This is not allowed. The routine is only guaranteed to converge if
    ! the value of the function YFn() at bestXAtEntry is less than its
    ! value at either end of the interval. In a moment we'll test for this
    ! explicitly, but a break of this rule is implied if the claimed best
    ! x value is located at one of the boundaries.
      if (present(status)) status = -1
  return
    end if

    if ((bestXAtEntry - loBracketX) / (hiBracketX - bestXAtEntry) < 0.0)&
    then
    ! The starting X values aren't in either increasing or decreasing
    ! order. They must be one or the other for the routine to work.
      if (present(status)) status = -2
  return
    end if
if (testFlag) print *, '          bbb0 brentsMethodVector'

    YAtLoX = yFn(x0 + dir * loBracketX)
if (testFlag) print *, '          bbb1 brentsMethodVector'
    bestY  = yFn(x0 + dir * bestXAtEntry)
if (testFlag) print *, '          bbb2 brentsMethodVector'
    YAtHiX = yFn(x0 + dir * hiBracketX)
if (testFlag) print *, '          ccc brentsMethodVector'

    if (bestY >= YAtLoX .or. bestY >= YAtHiX) then
    ! This is not allowed. The routine is only guaranteed to converge if
    ! the value of the function YFn() at bestXAtEntry is less than its
    ! value at either end of the interval.
      if (present(status)) status = -3
  return
    end if

    if (hiBracketX < loBracketX) then
      call swap(loBracketX, hiBracketX)
      call swap(YAtLoX, YAtHiX)
    end if

    if (present(maxNumIterations)) then
      localMaxNumIter = maxNumIterations
    else
      localMaxNumIter = 100
    end if

    if (present(deltaY)) then
      localDeltaY = abs(deltaY) + tiny(1.0)
    else
      localDeltaY = tiny(1.0)
    end if

    bestX = bestXAtEntry

    if (YAtLoX < YAtHiX) then
      secondBestX     = loBracketX
      secondBestY     = YAtLoX
      prevSecondBestX = hiBracketX
      prevSecondBestY = YAtHiX
    else
      secondBestX     = hiBracketX
      secondBestY     = YAtHiX
      prevSecondBestX = loBracketX
      prevSecondBestY = YAtLoX
    end if

    ! Perform an initial parabolic interpolation in order to estimate
    ! the x tolerance. Note that a +ve-curvature parabolic fit _must_
    ! succeed if the criteria for the initial triplet of (x,y) values
    ! are fulfilled.
if (testFlag) print *, '          ddd brentsMethodVector'

    call fitParabola(bestX, secondBestX, prevSecondBestX&
      , bestY, secondBestY, prevSecondBestY, parabolicFitOk&
      , curvature, parabolicFitX)
!if (testFlag) print *, '          eee0'

    if (.not. parabolicFitOk) then
      if (present(status)) status = -4
  return
    end if
!if (testFlag) print *, '          eee1'

    if (curvature <= 0.0) then
      if (present(status)) status = -5
  return
    end if
!if (testFlag) print *, '          eee2'

    if (parabolicFitX <= loBracketX .or. parabolicFitX >= hiBracketX) then
      if (present(status)) status = -6
  return
    end if
!if (testFlag) print *, '          eee3'

    previousStep = 0.0
!if (testFlag) print *, '          eee4'
    step = parabolicFitX - bestX
if (testFlag) print *, '          eee brentsMethodVector'
    yAtFittedX = yFn(x0 + dir * parabolicFitX)
if (testFlag) print *, '          fff brentsMethodVector'

    !                            a            b           u
    call doXAndYHouseKeeping(loBracketX, hiBracketX, parabolicFitX&
    !     x         w              v
      , bestX, secondBestX, prevSecondBestX, yAtFittedX, bestY&
      , secondBestY, prevSecondBestY)

    xTolSquared = (epsilon(1.0) * abs(bestY) + localDeltaY)&
      / curvature
    xTolerance = sqrt(xTolSquared)
    ! This formula arises as follows. There is always some uncertainty
    ! DY in the value of the function y(x). This is related to the
    ! uncertainty DX in x near the minimum x0 by
    !
    !	      y''(x0)
    !	DY = --------- DX^2 + (order>2 terms).
    !	         2
    !
    ! If we neglect the higher-order terms then we can invert this to
    ! give
    !
    !	         ( 2 * DY  )
    !	DX = sqrt(---------).
    !	         ( y''(x0) )
    !
    ! y''(x0) / 2 is just the curvature returned by the parabolic fit.
    ! We analyse DY into three terms: (i) the floating-point machine
    ! precision epsilon*y(x0); (ii) an optional intent(in) component
    ! deltaY; (iii) and lastly, a small number, to prevent DX going to
    ! zero in the case that y(x0) == 0 and no deltaY was supplied.

    xTolTimes2 = 2.0 * xTolerance

if (testFlag) print *, '          ggg brentsMethodVector'
    numIter = 1
    do ! until convergence
if (testFlag) print *, ''
if (testFlag) print *, '            000 numIter=', numIter, '/', localMaxNumIter
      midBracketX = 0.5 * (loBracketX + hiBracketX)

      ! Test for convergence:      
      if (abs(bestX - midBracketX)&
      <= (xTolTimes2 - 0.5 * (hiBracketX - loBracketX)))&
    exit

      ! The purpose of the next block is mainly to generate new values of
      ! step and previousStep. The tolerances may also get recalculated.

      if (abs(previousStep) <= xTolerance) then ! should always be true at
      ! first iteration.
        call goldenSection(bestX, midBracketX, loBracketX, hiBracketX, step&
          , previousStep)
        ! This sub doesn't read step and previousStep, but only writes them.

      else ! construct a trial parabolic fit
        call fitParabola(bestX, secondBestX, prevSecondBestX, bestY&
          , secondBestY, prevSecondBestY, parabolicFitOk, curvature&
          , parabolicFitX)

        twoStepsAgo  = previousStep
        previousStep = step
        step         = parabolicFitX - bestX ! This last is meaningless
        ! if either (.not. parabolicFitOk .or. curvature >= 0.0), but
        ! it is harmless, since parabolicFitX is always returned with
        ! a defined value. If the golden section routine is used, this
        ! sets its own value of step without reading the preexisting
        ! value.

        if (.not. parabolicFitOk .or. curvature <= 0.0&
        .or. parabolicFitX <= loBracketX .or. parabolicFitX >= hiBracketX&
        .or. abs(step) >= abs(twoStepsAgo / 2.0))&
        then
          call goldenSection(bestX, midBracketX, loBracketX, hiBracketX, step&
            , previousStep)
          ! This sub doesn't read step and previousStep, but only writes them.

        else ! at this point, basically, you can use the parabolic fit.
        ! We'll check however to see if the new X is too close to the
        ! bracket boundaries. If so, then we'll replace 'step' with a
        ! nominal value.

          if ( parabolicFitX < loBracketX + xTolerance&
          .or. parabolicFitX > hiBracketX - xTolerance) then
            if (bestX < midBracketX) then
              step =  xTolerance
            else
              step = -xTolerance
            end if
          end if

          xTolSquared = (epsilon(1.0) * abs(bestY) + localDeltaY&
            + tiny(1.0)) / curvature
          xTolerance = sqrt(xTolSquared)
          xTolTimes2 = 2.0 * xTolerance
        end if ! .not. parabolicFitOk
      end if ! abs(previousStep) <= xTolerance

      if (abs(step) < xTolerance) then
        if (step > 0.0) then
          step =  xTolerance
        else
          step = -xTolerance
        end if
      end if

      mostRecentX = bestX + step
      mostRecentY = yFn(x0 + dir * mostRecentX) ! this is the one function
      ! evaluation per iteration.

      call doXAndYHouseKeeping(loBracketX, hiBracketX, mostRecentX&
        , bestX, secondBestX, prevSecondBestX, mostRecentY, bestY&
        , secondBestY, prevSecondBestY)
      
      numIter = numIter + 1
      if (numIter > localMaxNumIter) then
        if (present(status)) status = -7
  return
      end if
    end do
if (testFlag) print *, '          hhh brentsMethodVector'

    solutionX = bestX
if (testFlag) print *, '          zzz brentsMethodVector'

  end subroutine brentsMethodVector

  !---------------------------------------------------------------------
  subroutine fitParabola(x0, x1, x2, y0, y1, y2, fitIsOk&
  , curvature, bestFitX)
  !---------------------------------------------------------------------
  ! x0, x1, x2 must be distinct values, but they don't have to occur in
  ! any order.

!***** put this in linear.f90??

    implicit none

    real(single), intent(in)  :: x1,&
                                 x0,&
                                 x2,&
                                 y1,&
                                 y0,&
                                 y2
    logical,      intent(out) :: fitIsOk
    real(single), intent(out) :: curvature,&
                                 bestFitX

    real(single)  :: num, den, xdiff

    !-------------------------------------------------------------------
    ! Start

    curvature = 0.0 ! default
    bestFitX  = 0.0 ! default
    fitIsOk = .true. ! default

    num = (y1 - y0) * (x2 - x1) * (x2 + x1)&
      -   (y2 - y1) * (x1 - x0) * (x1 + x0)

    den = (y1 - y0) * (x2 - x1) - (y2 - y1) * (x1 - x0)
    xdiff = (x0 - x2) * (x1 - x0) * (x2 - x1)

    if (xdiff == 0.0) then
      fitIsOk = .false.
  return
    end if
    ! Note that xdiff is guaranteed not to equal 0 so long as the x points
    ! are all different values.

    curvature = den / xdiff
    if (curvature .ne. 0.0) bestFitX = num / 2.0 / den

  end subroutine fitParabola

  !---------------------------------------------------------------------
  subroutine doXAndYHouseKeeping(loBracketX, hiBracketX, mostRecentX, bestX&
  , secondBestX, prevSecondBestX, mostRecentY, bestY, secondBestY&
  , prevSecondBestY)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)     :: mostRecentX,&
                                    mostRecentY
    real(single), intent(inout)  :: loBracketX,&
                                    hiBracketX,&
                                    prevSecondBestX,&
                                    secondBestX,&
                                    bestX,&
                                    prevSecondBestY,&
                                    secondBestY,&
                                    bestY

    !-------------------------------------------------------------------
    ! Start

    if (mostRecentY <= bestY) then
      if (mostRecentX >= bestX) then
        loBracketX = bestX
      else
        hiBracketX = bestX
      end if
      prevSecondBestX = secondBestX
      prevSecondBestY = secondBestY
      secondBestX = bestX
      secondBestY = bestY
      bestX = mostRecentX
      bestY = mostRecentY
        
    else ! mostRecentY > bestY
      if (mostRecentX < bestX) then
        loBracketX = mostRecentX
      else
        hiBracketX = mostRecentX
      end if

      if (mostRecentY <= secondBestY .or. secondBestX == bestX) then
        prevSecondBestX = secondBestX
        prevSecondBestY = secondBestY
        secondBestX = mostRecentX
        secondBestY = mostRecentY
      else if (mostRecentY <= prevSecondBestY&
      .or. prevSecondBestX == bestX .or. prevSecondBestX == secondBestX) then
        prevSecondBestX = mostRecentX
        prevSecondBestY = mostRecentY
      end if
    end if

  end subroutine doXAndYHouseKeeping

  !---------------------------------------------------------------------
  subroutine goldenSection(bestX, midBracketX, loBracketX, hiBracketX, step&
  , previousStep)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: bestX, midBracketX, loBracketX, hiBracketX
    real(single), intent(out) :: step, previousStep
    
    !-------------------------------------------------------------------
    ! Start

    if (bestX >= midBracketX) then ! do golden section step, which
    ! we take into the larger of the two segments.
      previousStep = loBracketX - bestX
    else
      previousStep = hiBracketX - bestX
    end if

    step = (1.0 - GoldenRatio) * previousStep ! take the golden section step.
    ! The constant GoldenRatio is in ssclib/constants.

  end subroutine goldenSection

  !---------------------------------------------------------------------
  subroutine powellsMethod(x, dirs, bestY, numIters, yFn&
  , minFracDeltaY, maxAllowedDeltaX, maxNumIterations, axisIsPinned&
  , xLimits, status, xLimitType, limitingAxisI)
  !---------------------------------------------------------------------
  ! This subroutine finds the minimum bestY=yFn(p) of the function yFn
  ! in a space x(). Vector x() should be loaded with starting values. The
  ! first index of dirs() is the dimension number, the second is the
  ! vector number.

    implicit none

    real(single), intent(inout)         :: x(:), dirs(size(x),size(x))
    real(single), intent(out)           :: bestY
    integer,      intent(out)           :: numIters
    real(single), intent(in),  optional :: minFracDeltaY
    real(single), intent(in),  optional :: maxAllowedDeltaX
    integer,      intent(in),  optional :: maxNumIterations
    logical,      intent(in),  optional :: axisIsPinned(size(x))
    type(RangeT), intent(in),  optional :: xLimits(size(x))
    integer,      intent(out), optional :: status
    integer,      intent(out), optional :: xLimitType, limitingAxisI

    interface
      function yFn(x)
        use types

        real(single), intent(in) :: x(:)
        real(single) :: yFn
      end function yFn
    end interface

    integer, parameter :: numReservedStatusValues = 3

    integer ::&
      i, directionOfLargestDecrease, localMaxNumIter, numDims, foreignStatus
    real(single) ::&
      largestDeltaY, y0, previousBestY, yAtExtrap, localMinFracDeltaY,&
      averageAbsY, remainderDeltaY, criterion2
    real(single), allocatable :: xOld(:), xExtrap(:), dirTemp(:)

    !-------------------------------------------------------------------
    ! Start

    if (present(status)) status = 0

    numDims = size(x)

    do i = 1, numDims
      if (all(dirs(:,i) == 0.0)) then
        if (present(status)) status = -1
  return
      end if
    end do

    if (present(minFracDeltaY)) then
      if (minFracDeltaY < 0.0) then
        if (present(status)) status = ibset(status, 0)
        localMinFracDeltaY = 0.0
      else
        localMinFracDeltaY = minFracDeltaY
      end if
    else
      localMinFracDeltaY = 0.0
    end if

    if (present(maxNumIterations)) then
      localMaxNumIter = maxNumIterations
    else
      localMaxNumIter = 100
    end if

    if (present(axisIsPinned)) then
      if (all(axisIsPinned)) then
        status = -2
  return
      end if

      do i = 1, numDims
        where(axisIsPinned) dirs(:,i) = 0.0
!*** I _think_ this step (and the 'cycle') are all that is necessary to achieve minimization while avoiding any change in x in the pinned directions.
      end do
    end if

    allocate(xOld(numDims), xExtrap(numDims), dirTemp(numDims))

    bestY = yFn(x)
    ! save the initial point
    xOld = x

    do numIters = 1, localMaxNumIter
      y0 = bestY
      largestDeltaY = 0.0 ! will be the biggest function decrease

      do i = 1, numDims ! In each iteration, loop over all directions in
      ! the set.

        if (present(axisIsPinned)) then
          if (axisIsPinned(i))&
      cycle
        end if

        ! Copy the direction.
        dirTemp = dirs(:,i)

        previousBestY = bestY
        call findBestYThisDirection(yFn, localMaxNumIter, x, dirTemp, bestY&
          , localMinFracDeltaY * abs(bestY), foreignStatus, maxAllowedDeltaX&
          , xLimits, xLimitType, limitingAxisI)
        ! foreignStatus 0 to -17

        if (foreignStatus .ne. 0) then
          if (present(status)) status = foreignStatus - numReservedStatusValues
          ! Thus -9 indicates that x has reached one of xLimits.
  return
        end if

        if (previousBestY - bestY > largestDeltaY) then
        ! Record it if it is the largest decrease so far.
          largestDeltaY = previousBestY - bestY
          directionOfLargestDecrease = i
        end if
      end do

      averageAbsY = (abs(y0) + abs(bestY)) / 2.0
  if ((y0 - bestY) <= (localMinFracDeltaY + epsilon(1.0)) * averageAbsY) return

      ! Construct the extrapolated point and the average direction moved.
      ! Save the old starting point.
      xExtrap = 2.0 * x - xOld
      dirTemp = x - xOld
      xOld = x

      yAtExtrap = yFn(xExtrap)
    if (yAtExtrap >= y0) cycle ! One reason not to use a new direction.
      remainderDeltaY = (y0 - bestY) - largestDeltaY
      criterion2 = 2.0 * (y0 - 2.0 * bestY + yAtExtrap) * remainderDeltaY&
        * remainderDeltaY - largestDeltaY * (y0 - yAtExtrap) * (y0 - yAtExtrap)
    if (criterion2 >= 0.0) cycle ! Other reason not to use the new direction.

      ! Move to minimum of new direction:
      call findBestYThisDirection(yFn, localMaxNumIter, x, dirTemp, bestY&
        , localMinFracDeltaY * abs(bestY), foreignStatus, maxAllowedDeltaX&
        , xLimits, xLimitType, limitingAxisI)
      ! foreignStatus 0 to -17

      if (foreignStatus .ne. 0) then
        if (present(status)) status = foreignStatus - numReservedStatusValues
        ! Thus -9 indicates that x has reached one of xLimits.
  return
      end if

      ! Save new direction:
      dirs(:, directionOfLargestDecrease) = dirs(:, numDims)
      dirs(:, numDims)   = dirTemp
    end do

    if (present(status)) status = -3 ! because max iterations exceeded.
    !*** set the output minimum x

  end subroutine powellsMethod

  !---------------------------------------------------------------------
  subroutine findBestYThisDirection(yFn, maxNumIterations&
  , r, dir, bestY, minDeltaY, status, maxAllowedDeltaX, rLimits&
  , xLimitType, limitingAxisI)
  !---------------------------------------------------------------------

    implicit none

    integer,      intent(in)             :: maxNumIterations
    real(single), intent(inout)          :: r(:), dir(size(r))
    real(single), intent(out)            :: bestY
    integer,      intent(out)            :: status
    real(single), intent(in)             :: minDeltaY
    real(single), intent(in),   optional :: maxAllowedDeltaX
    type(RangeT), intent(in),   optional :: rLimits(size(r))
    integer,      intent(out),  optional :: xLimitType, limitingAxisI

    interface
      function yFn(r)
        use types

        real(single), intent(in) :: r(:)
        real(single) :: yFn
      end function yFn
    end interface

    real(single) :: xLo, xMid, xHi, yAtXLo, yAtXMid, yAtXHi, stepToMin
    integer :: foreignStatus
    integer :: numReservedStatusValsFXB = 8

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '        aaa findBestYThisDirection'

    status = 0
    
    xLo  = 0.0
    xMid = 1.0
if (testFlag) print *, '        bbb findBestYThisDirection'
    call findXBracketOfMinYVector(r, dir, xLo, xMid, xHi, yAtXLo, yAtXMid&
      , yAtXHi, yFn, foreignStatus, maxAllowedDeltaX, rLimits, xLimitType&
      , limitingAxisI) ! foreignStatus 0 to -8

if (testFlag) print *, '        ccc findBestYThisDirection'

    if (foreignStatus .ne. 0) then
      r = r + dir * xHi ! this is only of any use if status = -6 (=> rLimits exceeded).
      status = foreignStatus
  return
    end if
if (testFlag) print *, '        ddd findBestYThisDirection'

    call brentsMethodVector(r, dir, xMid, xLo, xHi, stepToMin, yFn&
      , maxNumIterations, minDeltaY, foreignStatus) ! foreignStatus 0 to -9
if (testFlag) print *, '        eee findBestYThisDirection'

    if (foreignStatus .ne. 0) then
      status = foreignStatus - numReservedStatusValsFXB
  return
    end if

    dir = stepToMin * dir
    r = r + dir

    bestY = yFn(r)
if (testFlag) print *, '        zzz findBestYThisDirection'

  end subroutine findBestYThisDirection

  !---------------------------------------------------------------------
  subroutine findXBracketOfMinYScalar(xLo, xMid, xHi, yAtXLo, yAtXMid&
  , yAtXHi, yFn, status, maxAllowedDeltaX, xLimits, xLimitType)
  !---------------------------------------------------------------------

    real(single), intent(inout)         :: xLo, xMid
    real(single), intent(out)           :: xHi, yAtXLo, yAtXMid, yAtXHi
    integer,      intent(out)           :: status
    real(single), intent(in),  optional :: maxAllowedDeltaX
    type(RangeT), intent(in),  optional :: xLimits
    integer,      intent(out), optional :: xLimitType

    interface
      function yFn(x)
        use types

        real(single), intent(in) :: x(:)
        real(single) :: yFn
      end function yFn
    end interface

    !-------------------------------------------------------------------
    ! Start

    call findXBracketOfMinYVector((/0.0/), (/1.0/), xLo, xMid, xHi, yAtXLo&
      , yAtXMid, yAtXHi, yFn, status, maxAllowedDeltaX, (/xLimits/)&
      , xLimitType)

  end subroutine findXBracketOfMinYScalar

  !---------------------------------------------------------------------
  subroutine findXBracketOfMinYVector(r0, dir, xLo, xMid, xHi, yAtXLo&
  , yAtXMid, yAtXHi, yFn, status, maxAllowedDeltaX, rLimits, xLimitTypeArg&
  , limitingAxisI)
  !---------------------------------------------------------------------
  ! Returns 3 x values (aligned along direction dir from x0) xLo, xMid, xHi
  ! which have the property yFn(xMid) < either yFn(xLo) or yFn(xHi). Thus
  ! one _knows_ that the minimum _must_ be somewhere within the interval
  ! xLo - xHi, and can 'hunt it down like a scared rabbit' as Press et al
  ! have it.
  !
  ! Most negative status value = -8.

    implicit none

    real(single), intent(in)            :: r0(:), dir(size(r0))
    real(single), intent(inout)         :: xLo, xMid
    real(single), intent(out)           :: xHi, yAtXLo, yAtXMid, yAtXHi
    integer,      intent(out)           :: status
    real(single), intent(in),  optional :: maxAllowedDeltaX
    type(RangeT), intent(in),  optional :: rLimits(size(r0))
    integer,      intent(out), optional :: xLimitTypeArg, limitingAxisI

    interface
      function yFn(r)
        use types

        real(single), intent(in) :: r(:)
        real(single) :: yFn
      end function yFn
    end interface

    real(single) ::&
      maxAllowedX, minAllowedX, trialXHi, trialYHi, curvature, parabolicFitX,&
      yAtXParab
    logical :: fitIsOk, noFreeXRange
    integer :: foreignStatus, xLimitType, i, loLimitAxisI, hiLimitAxisI
    integer, parameter :: numReservedStatusValues = 6
    type(RangeT) :: xLimits

    !-------------------------------------------------------------------
    ! Start

    status = 0

    xHi     = 0.0
    yAtXLo  = 0.0
    yAtXMid = 0.0
    yAtXHi  = 0.0
    if (present(xLimitTypeArg)) xLimitTypeArg = UNDEFINED
    if (present(limitingAxisI)) limitingAxisI = 0
    ! Just so as not to leave them undefined in case of a bad status return
    ! (in fact these defaults are absolutely necessary - these are the values
    ! returned when a parab fit is good, for example).

    if (size(r0) <= 0) then
      status = -1
  return
    end if

    if (all(dir == 0.0)) then
      status = -2
  return
    end if

    if (present(maxAllowedDeltaX)) then
      if (maxAllowedDeltaX <= 0.0) then
        status = -3
  return
      end if
    end if

    if (present(rLimits)) then
      do i = 1, size(r0)
        if (.not.rangesAreWellFormed((/rLimits(i)/))) then
        ! rangesAreWellFormed is in ssclib/dss_ranges_aux.
          status = -4
  return
        end if
      end do
    end if

    call calcXLimits(r0, dir, xLimits, loLimitAxisI, hiLimitAxisI&
      , noFreeXRange, rLimits)

    if (noFreeXRange) then
      status = -5
  return
    end if

    call forceStartingXsWithinLimits(r0, dir, yFn, xLimits, xLo, xMid)

    maxAllowedX = huge(1.0) / 10.0
    minAllowedX = -maxAllowedX

    yAtXLo  = yFn(r0 + dir * xLo)
    yAtXMid = yFn(r0 + dir * xMid)
    if (yAtXMid > yAtXLo) then
      call swap(xLo, xMid) ! swap is in ssclib/ssc_misc.
      call swap(yAtXLo, yAtXMid)
    end if

    xHi = xMid + GoldenRatio * (xMid - xLo) ! first guess for xHi.
    ! GoldenRatio is defined in ssclib/constants.

    call forceXWithinLimits(xLimits, loLimitAxisI, hiLimitAxisI&
      , xHi, xLimitType, limitingAxisI)
    ! This adjusts xHi such that it is not outside xLimits. in a moment
    ! we will test to see whether a minimum has been achieved; if not,
    ! then the value of xLimitType is tested and, if not UNDEFINED, that
    ! means that the routine has run up against one of the limits without
    ! being able to bracket a minimum.

    call checkThatXIsInRange(xHi, xLo, maxAllowedDeltaX&
      , minAllowedX, maxAllowedX, foreignStatus)
    if (foreignStatus < 0) then
      status = foreignStatus - numReservedStatusValues
  return
    end if

    yAtXHi = yFn(r0 + dir * xHi)
    ! Test for minimum:
    if (yAtXMid < yAtXHi)&
  return

    if (xLimitType .ne. UNDEFINED) then
      status = -6
      if (present(xLimitTypeArg)) xLimitTypeArg = xLimitType
  return
    end if

    do ! keep returning here until the minimum is found.
      call fitParabola(xMid, xLo, xHi, yAtXMid, yAtXLo, yAtXHi&
        , fitIsOk, curvature, parabolicFitX)

      if (.not. fitIsOk .or. curvature <= 0.0) then
        trialXHi = xHi + GoldenRatio * (xHi - xMid)

        call forceXWithinLimits(xLimits, loLimitAxisI, hiLimitAxisI&
          , trialXHi, xLimitType, limitingAxisI)
        ! This adjusts trialXHi such that it is not outside xLimits.
        ! There is a test to see whether a minimum has been achieved at the
        ! end of the loop: if not, then the value of xLimitType is tested
        ! and, if not UNDEFINED, that means that the routine has run up
        ! against one of the limits without being able to bracket a minimum.

        call checkThatXIsInRange(trialXHi, xLo, maxAllowedDeltaX&
          , minAllowedX, maxAllowedX, foreignStatus)
        if (foreignStatus < 0) then
          status = foreignStatus - numReservedStatusValues
  return
        end if

        trialYHi = yFn(r0 + dir * trialXHi) ! will be tested at end of loop

      else ! parabolic fit seems to be fine. Check first where the fitted X
      ! value finds itself:

        if ((xMid - parabolicFitX) * (xMid - xLo) >= 0.0) then
        ! either parabolicFitX = xMid or it is on the xLo side of it. This
        ! ought not to be possible at this point. The status return
        ! -99 indicates a bug.
          status = -99
  return
        end if

        if ((xMid - parabolicFitX) * (parabolicFitX - xHi) > 0.0) then
        ! parabolicFitX is between xMid and xHi. Try the function value:

          yAtXParab = yFn(r0 + dir * parabolicFitX)
          if (yAtXParab < yAtXHi) then ! minimum must be between
          ! present xMid and xHi.
            xLo  = xMid
            xMid = parabolicFitX
            yAtXLo  = yAtXMid
            yAtXMid = yAtXParab
  return
          else if (yAtXParab > yAtXMid) then ! minimum must be between
          ! present xLo and parabolicFitX.
            xHi    = parabolicFitX
            yAtXHi = yAtXParab
  return
          end if

          ! No minimum yet found. Do the default step:
          trialXHi = xHi + GoldenRatio * (xHi - xMid)

          call forceXWithinLimits(xLimits, loLimitAxisI, hiLimitAxisI&
            , trialXHi, xLimitType, limitingAxisI)
          ! This adjusts trialXHi such that it is not outside xLimits.
          ! There is a test to see whether a minimum has been achieved at
          ! the end of the loop: if not, then the value of xLimitType is
          ! tested and, if not UNDEFINED, that means that the routine has
          ! run up against one of the limits without being able to bracket
          ! a minimum.

          call checkThatXIsInRange(trialXHi, xLo, maxAllowedDeltaX&
            , minAllowedX, maxAllowedX, foreignStatus)
          if (foreignStatus < 0) then
            status = foreignStatus - numReservedStatusValues
  return
          end if

          trialYHi = yFn(r0 + dir * trialXHi) ! will be tested at end of
          ! loop.

        else ! parabolicFitX is either = xHi or on the far side of it. Check
        ! that it is not too far:

          call forceXWithinLimits(xLimits, loLimitAxisI, hiLimitAxisI&
            , parabolicFitX, xLimitType, limitingAxisI)
          ! This adjusts parabolicFitX such that it is not outside xLimits.
          ! There is a test to see whether a minimum has been achieved at
          ! the end of the loop: if not, then the value of xLimitType is
          ! tested and, if not UNDEFINED, that means that the routine has
          ! run up against one of the limits without being able to bracket
          ! a minimum.

          call checkThatXIsInRange(parabolicFitX, xLo, maxAllowedDeltaX&
            , minAllowedX, maxAllowedX, foreignStatus)
          if (foreignStatus < 0) then
            status = foreignStatus - numReservedStatusValues
  return
          end if

          yAtXParab = yFn(r0 + dir * parabolicFitX)
          if (yAtXParab < yAtXHi) then ! damn, no minimum found yet. Do
          ! some interim housekeeping and squeeze in another step:

            xMid = xHi
            xHi  = parabolicFitX
            trialXHi = xHi + GoldenRatio * (xHi - xMid)

            call forceXWithinLimits(xLimits, loLimitAxisI, hiLimitAxisI&
              , trialXHi, xLimitType, limitingAxisI)
            ! This adjusts trialXHi such that it is not outside xLimits.
            ! There is a test to see whether a minimum has been achieved at
            ! the end of the loop: if not, then the value of xLimitType is
            ! tested and, if not UNDEFINED, that means that the routine has
            ! run up against one of the limits without being able to bracket
            ! a minimum.

            call checkThatXIsInRange(trialXHi, xLo, maxAllowedDeltaX&
              , minAllowedX, maxAllowedX, foreignStatus)
            if (foreignStatus < 0) then
              status = foreignStatus - numReservedStatusValues
  return
            end if

            yAtXMid = yAtXHi
            yAtXHi  = yAtXParab
            trialYHi = yFn(r0 + dir * trialXHi) ! will be tested at end of
            ! loop.

          else ! we've found a minimum and, after housekeeping, the test
          ! at the start of the loop should detect it.
            trialXHi = parabolicFitX
            trialYHi = yAtXParab
          end if
        end if ! end of parabolicFitX-location decision tree.
      end if ! end of block which generates trialXHi and trialYHi.

      ! Housekeeping:
      xLo  = xMid
      xMid = xHi
      xHi  = trialXHi
      yAtXLo  = yAtXMid
      yAtXMid = yAtXHi
      yAtXHi  = trialYHi

      ! Now test for minimum:
      if (yAtXMid < yAtXHi)&
  return

      ! No minimum found - bugger. Now comes the day of reckoning if xHi
      ! ran up against a limit:
      if (xLimitType .ne. UNDEFINED) then
        status = -6
        if (present(xLimitTypeArg)) xLimitTypeArg = xLimitType
  return
      end if
    end do

  end subroutine findXBracketOfMinYVector

  !---------------------------------------------------------------------
  subroutine calcXLimits(r0, dir, xLimits, loLimitAxisI, hiLimitAxisI&
  , noFreeXRange, rLimits)
  !---------------------------------------------------------------------
  ! r0 + x * dir defines a straight line in a space of dimension N = size(r0). Given a set of limiting ranges, 1 range for each of the N axes, this subroutine calculates the range of permitted x values.
  !
  ! NOTE! If rLimits is present, this subroutine assumes that all rLimits ranges are 'well formed' (ie, pass ssclib/dss_ranges_aux.f90:rangesAreWellFormed); if not true, beware: you will get a goat fuck.

    implicit none

    real(single), intent(in)            :: r0(:), dir(size(r0))
    type(RangeT), intent(out)           :: xLimits
    integer,      intent(out)           :: loLimitAxisI, hiLimitAxisI
    logical,      intent(out)           :: noFreeXRange
    type(RangeT), intent(in),  optional :: rLimits(size(r0))

    integer :: axisI, loStatus, hiStatus
    type(RangeT) :: trialXRange, minXRange

    !-------------------------------------------------------------------
    ! Start

    ! defaults:
    call initializeRangeT(xLimits) ! in ssclib/dss_ranges_aux.
    xLimits%lower%type = UNDEFINED
    xLimits%upper%type = UNDEFINED
    loLimitAxisI = 0
    hiLimitAxisI = 0
    noFreeXRange = .false.

  if (.not.present(rLimits)) return

    do axisI = 1, size(r0)
      if (dir(axisI) == 0.0) then ! check to see if r0 is within the
      ! limiting range on this axis. If so, then the range of x is
      ! infinite (because r0(axisI) + x * dir(axisI) = r0(axisI) for
      ! any x); if not, then there is no range at all.

        if (valueIsWithinRange(r0(axisI), rLimits(axisI))) then
        ! in ssclib/dss_ranges_aux.
    cycle
        else
          noFreeXRange = .true.
  return
        end if
      end if

      trialXRange = calcLimitXRange(r0(axisI), dir(axisI), rLimits(axisI))

      if (loLimitAxisI == 0) then ! on no previous axis has a limit been
      ! exceeded. Note that loLimitAxisI and hiLimitAxisI are set at the
      ! same time, so there's no point in testing them both.

        loLimitAxisI = axisI
        hiLimitAxisI = axisI
        minXRange = trialXRange

      else
        call compareRangeBounds(trialXRange, minXRange, loStatus, hiStatus)
        ! in ssclib/dss_ranges_aux.

        if (hiStatus == BOUNDALTB) then
          hiLimitAxisI = axisI
          minXRange%upper = trialXRange%upper
        end if

        if (loStatus == BOUNDAGTB) then
          loLimitAxisI = axisI
          minXRange%lower = trialXRange%lower
        end if

        if (.not.rangesAreWellFormed((/minXRange/))) then
        ! upper must be < lower. This call is in ssclib/dss_ranges_aux.
          noFreeXRange = .true.
  return
        end if
      end if ! first limit encountered.
    end do ! loop over axes.

  end subroutine calcXLimits

  !---------------------------------------------------------------------
  subroutine forceStartingXsWithinLimits(r0, dir, yFn, xLimits, xLo, xMid)
  !---------------------------------------------------------------------
  ! yFn is a function defined on a space of dimension N = size(r0). The subroutine findXBracketOfMinYVector attempts to find three x values which bracket a minimum value of yFn along the line r0 + x * dir. Two starting values, xLo and xMid, are supplied to the subroutine. The present subroutine ensures that these starting values are within the limiting range of x values (if these limits exist).
  !
  ! There are 3 basic limit scenarios: no limits, 1 limit (upper or lower), both upper and lower limits. The outcome of the subroutine in each of the three is as follows:
  !
  !	(i) no limits: no change in xLo, xMid.
  !
  !	(ii) 1 limit only: given dx = abs(xMid - xLo), the pair is moved if necessary (keeping the same spacing) such that the x with the smallest yFn value is placed at least 1*dx away from the limit.
  !
  !	(iii) limits both sides. If the gap between the limits > 2*dx, then as for case (ii); othewise, dx is set to 1/2 the gap size, then as for case (ii).
  !
  ! NOTE! It is assumed that xLo .ne. xMid; it is not however necessary for xLo to be < xMid (or vice versa).

    implicit none

    real(single), intent(in)    :: r0(:), dir(size(r0))
    type(RangeT), intent(in)    :: xLimits
    real(single), intent(inout) :: xLo, xMid

    interface
      function yFn(r)
        use types

        real(single), intent(in) :: r(:)
        real(single) :: yFn
      end function yFn
    end interface

    real(single) :: dx, xHi, yLo, yMid, yHi, hiLimitValue, loLimitValue

    !-------------------------------------------------------------------
    ! Start

    if (xLimits%lower%type==UNDEFINED .and. xLimits%upper%type==UNDEFINED)&
  return ! case (i).

    if (xLo > xMid) call swap(xLo, xMid) ! this is temporarily convenient.
    dx = xMid - xLo

    if (xLimits%lower%type==UNDEFINED) then ! case (ii), upper limit.
      if (xLimits%upper%type==INCLUSIVE) then
        hiLimitValue = xLimits%upper%value
      else ! xLimits%upper%type==EXCLUSIVE; UNDEFINED is not here an option.
        hiLimitValue = nearest(xLimits%upper%value, -1.0)
      end if

      if (xMid > hiLimitValue) then
        xLo  = hiLimitValue - dx
        xMid = hiLimitValue
        yLo  = yFn(r0 + dir * xLo)
        yMid = yFn(r0 + dir * xMid)

        if (yLo < yMid)&
  return

        xLo  = hiLimitValue - 2.0 * dx
        xMid = hiLimitValue -       dx
  return

      else ! xMid <= hiLimitValue
        yLo  = yFn(r0 + dir * xLo)
        yMid = yFn(r0 + dir * xMid)

        if (yLo < yMid)&
  return

        if (xMid <= hiLimitValue - dx)&
  return

        xLo  = hiLimitValue - 2.0 * dx
        xMid = hiLimitValue -       dx
  return
      end if

    else if (xLimits%upper%type==UNDEFINED) then ! case (ii), lower limit.
      if (xLimits%lower%type==INCLUSIVE) then
        loLimitValue = xLimits%lower%value
      else ! xLimits%lower%type==EXCLUSIVE
        loLimitValue = nearest(xLimits%lower%value, 1.0)
      end if

      if (xLo < loLimitValue) then
        xLo  = loLimitValue
        xMid = loLimitValue + dx
        yLo  = yFn(r0 + dir * xLo)
        yMid = yFn(r0 + dir * xMid)

        if (yLo > yMid)&
  return

        xLo  = loLimitValue +       dx
        xMid = loLimitValue + 2.0 * dx
  return

      else ! xLo >= loLimitValue
        yLo  = yFn(r0 + dir * xLo)
        yMid = yFn(r0 + dir * xMid)

        if (yLo > yMid)&
  return

        if (xLo >= loLimitValue + dx)&
  return

        xLo  = loLimitValue +       dx
        xMid = loLimitValue + 2.0 * dx
  return
      end if

    else ! both limits defined. Case (iii).
      if (xLimits%upper%type==INCLUSIVE) then
        hiLimitValue = xLimits%upper%value
      else ! xLimits%upper%type==EXCLUSIVE; UNDEFINED is not here an option.
        hiLimitValue = nearest(xLimits%upper%value, -1.0)
      end if

      if (xLimits%lower%type==INCLUSIVE) then
        loLimitValue = xLimits%lower%value
      else ! xLimits%lower%type==EXCLUSIVE
        loLimitValue = nearest(xLimits%lower%value, 1.0)
      end if

      if (hiLimitValue - loLimitValue < dx) then
        xLo  = loLimitValue
        xMid = (loLimitValue + hiLimitValue) / 2.0
        xHi  = hiLimitValue
        yLo  = yFn(r0 + dir * xLo)
        yMid = yFn(r0 + dir * xMid)
        yHi  = yFn(r0 + dir * xHi)

        if (yLo < yMid) then ! assume yHi > yMid - not necessarily true! But we should realize that this whole routine is a bit of a lick and a promise. It isn't easy to make a really rigorous separation between cases where the minimum really is out of range and cases where it just seems like it.
          xLo = xMid
          xMid = xHi
        end if
  return
      end if

      ! otherwise, similar to case (ii).

      if (xMid > hiLimitValue) then
        xLo  = hiLimitValue - dx
        xMid = hiLimitValue
        yLo  = yFn(r0 + dir * xLo)
        yMid = yFn(r0 + dir * xMid)

        if (yLo < yMid)&
  return

        xLo  = hiLimitValue - 2.0 * dx
        xMid = hiLimitValue -       dx
  return

      else if (xLo < loLimitValue) then
        xLo  = loLimitValue
        xMid = loLimitValue + dx
        yLo  = yFn(r0 + dir * xLo)
        yMid = yFn(r0 + dir * xMid)

        if (yLo > yMid)&
  return

        xLo  = loLimitValue +       dx
        xMid = loLimitValue + 2.0 * dx
  return

      else ! xMid <= hiLimitValue .and. xLo >= loLimitValue
        yLo  = yFn(r0 + dir * xLo)
        yMid = yFn(r0 + dir * xMid)

        if (yLo < yMid)&
  return

        if (xMid <= hiLimitValue - dx .and. xLo >= loLimitValue + dx)&
  return

        if (hiLimitValue - xMid < xLo - loLimitValue) then
          xLo  = hiLimitValue - 2.0 * dx
          xMid = hiLimitValue -       dx
        else
          xLo  = loLimitValue +       dx
          xMid = loLimitValue + 2.0 * dx
        end if
  return
      end if
    end if

  end subroutine forceStartingXsWithinLimits

  !---------------------------------------------------------------------
  subroutine forceXWithinLimits(xLimits, loLimitAxisI, hiLimitAxisI&
  , xTrial, xLimitType, limitingAxisI)
  !---------------------------------------------------------------------
  ! This subroutine tests xTrial to see if it is within the range xLimits.
  ! If not, xTrial is given the value of the closest value of x which is
  ! within the range.

    implicit none

    type(RangeT), intent(in)            :: xLimits
    integer,      intent(in)            :: loLimitAxisI, hiLimitAxisI
    real(single), intent(inout)         :: xTrial
    integer,      intent(out)           :: xLimitType
    integer,      intent(out), optional :: limitingAxisI

    real(single) :: loLimitValue, hiLimitValue

    !-------------------------------------------------------------------
    ! Start

    ! defaults:
    xLimitType = UNDEFINED
    if (present(limitingAxisI)) limitingAxisI = 0

    if (valueIsWithinRange(xTrial, xLimits))&
  return
    ! in ssclib/dss_ranges_aux.

    if (xLimits%lower%type==UNDEFINED) then

      ! xLimits%upper%type cannot also be UNDEFINED, otherwise xTrial
      ! would have been within the range.

      if (xLimits%upper%type==INCLUSIVE) then
        xTrial = xLimits%upper%value
      else ! xLimits%upper%type==EXCLUSIVE; UNDEFINED is not here an option.
        xTrial = nearest(xLimits%upper%value, -1.0)
      end if

      xLimitType = xLimits%upper%type
      if (present(limitingAxisI)) limitingAxisI = hiLimitAxisI

    else if (xLimits%upper%type==UNDEFINED) then

      ! xLimits%lower%type cannot also be UNDEFINED, otherwise xTrial
      ! would have been within the range.

      if (xLimits%lower%type==INCLUSIVE) then
        xTrial = xLimits%lower%value
      else ! xLimits%lower%type==EXCLUSIVE
        xTrial = nearest(xLimits%lower%value, 1.0)
      end if

      xLimitType = xLimits%lower%type
      if (present(limitingAxisI)) limitingAxisI = loLimitAxisI

    else ! both limits are defined.
      if (xLimits%upper%type==INCLUSIVE) then
        hiLimitValue = xLimits%upper%value
      else ! xLimits%upper%type==EXCLUSIVE; UNDEFINED is not here an option.
        hiLimitValue = nearest(xLimits%upper%value, -1.0)
      end if

      if (xLimits%lower%type==INCLUSIVE) then
        loLimitValue = xLimits%lower%value
      else ! xLimits%lower%type==EXCLUSIVE
        loLimitValue = nearest(xLimits%lower%value, 1.0)
      end if

      if (abs(loLimitValue - xTrial) < abs(xTrial - hiLimitValue)) then
        xTrial = loLimitValue
        xLimitType = xLimits%lower%type
        if (present(limitingAxisI)) limitingAxisI = loLimitAxisI
      else
        xTrial = hiLimitValue
        xLimitType = xLimits%upper%type
        if (present(limitingAxisI)) limitingAxisI = hiLimitAxisI
      end if
    end if

  end subroutine forceXWithinLimits

  !---------------------------------------------------------------------
  subroutine checkThatXIsInRange(trialXHi, xLo, maxAllowedDeltaX&
  , minAllowedX, maxAllowedX, status)
  !---------------------------------------------------------------------
  ! Most negative status value = -2.

    implicit none

    real(single), intent(inout)        :: trialXHi
    real(single), intent(in)           :: xLo, minAllowedX, maxAllowedX
    real(single), intent(in), optional :: maxAllowedDeltaX
    integer,      intent(out)          :: status

    !-------------------------------------------------------------------
    ! Start

    status = 0

    if (present(maxAllowedDeltaX)) then
      if (maxAllowedDeltaX <= 0.0) then
        status = -1
  return
      end if

      if (abs(trialXHi - xLo) > maxAllowedDeltaX) then
        trialXHi = xLo + sign(maxAllowedDeltaX, trialXHi - xLo)
        status = ibset(status, 0)
      end if

    else
      if (maxAllowedX <= minAllowedX) then
        status = -2
  return
      end if

      if (trialXHi >= maxAllowedX) then
        trialXHi = maxAllowedX
        status = ibset(status, 1)
      else if (trialXHi <= minAllowedX) then
        trialXHi = minAllowedX
        status = ibset(status, 2)
      end if
    end if

  end subroutine checkThatXIsInRange

  !---------------------------------------------------------------------
  logical function xIsWithinLimits(x0, dir, x, xLimits)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: x0(:), dir(size(x0))
    real(single), intent(inout)        :: x
    type(RangeT), intent(in), optional :: xLimits(size(x0))

    real(single), allocatable :: x1(:)
    logical :: limitSomeAxisAlreadyExceeded
    integer :: i
    type(RangeT) :: trialXRange, minXRange

    !-------------------------------------------------------------------
    ! Start

    xIsWithinLimits = .true. ! default
  if (.not.present(xLimits)) return

    allocate(x1(size(x0)))
    x1 = x0 + dir * x
    limitSomeAxisAlreadyExceeded = .false. ! default
    do i = 1, size(x0)
      if (.not.valueIsWithinRange(x1(i), xLimits(i))) then
      ! valueIsWithinRange is in ssclib/dss_ranges_aux.
        trialXRange = calcLimitXRange(x0(i), dir(i), xLimits(i))
        if (limitSomeAxisAlreadyExceeded) then
          call andRanges(trialXRange, minXRange) ! in ssclib/dss_ranges_aux.
        else
          limitSomeAxisAlreadyExceeded = .true.
          minXRange = trialXRange
        end if
      end if
    end do

    if (limitSomeAxisAlreadyExceeded) then
      xIsWithinLimits = .false.
      if (x >= 0.0) then
        x = minXRange%upper%value
      else
        x = minXRange%lower%value
      end if
    end if

  end function xIsWithinLimits

  !---------------------------------------------------------------------
  function calcLimitXRange(x0, mul, xLimits) result(aLimits)
  !---------------------------------------------------------------------
  ! NOTE that this function assumes that xLimits is 'well-formed' and
  ! that mul!=0.

    implicit none

    real(single), intent(in)           :: x0, mul
    type(RangeT), intent(in), optional :: xLimits
    type(RangeT)                       :: aLimits

    !-------------------------------------------------------------------
    ! Start

    call initializeRangeT(aLimits) ! in ssclib/dss_ranges_aux.

    select case(xLimits%lower%type)
    case(UNDEFINED)
      aLimits%lower%type = UNDEFINED
    case(INCLUSIVE)
      aLimits%lower%type = INCLUSIVE
      aLimits%lower%value = (xLimits%lower%value - x0) / mul
    case(EXCLUSIVE)
      aLimits%lower%type = EXCLUSIVE
      aLimits%lower%value = (xLimits%lower%value - x0) / mul
    case default
      aLimits%lower%type = UNDEFINED
    end select

    select case(xLimits%upper%type)
    case(UNDEFINED)
      aLimits%upper%type = UNDEFINED
    case(INCLUSIVE)
      aLimits%upper%type = INCLUSIVE
      aLimits%upper%value = (xLimits%upper%value - x0) / mul
    case(EXCLUSIVE)
      aLimits%upper%type = EXCLUSIVE
      aLimits%upper%value = (xLimits%upper%value - x0) / mul
    case default
      aLimits%upper%type = UNDEFINED
    end select

    ! Note: if mul < 0, then aLimits is not well formed: ie,
    ! aLimits%lower%value will be > aLimits%upper%value. This is now
    ! corrected:

    if (mul < 0.0) call swapRangeBounds(aLimits) ! in dss_ranges_aux.

  end function calcLimitXRange

  !---------------------------------------------------------------------
  subroutine findXBracketOfZeroYScalar(maxNumIterations, xLo, xHi&
  , yAtXLo, yAtXHi, yFn, status)!, maxAllowedDeltaX)
  !---------------------------------------------------------------------

    integer,      intent(in)           :: maxNumIterations
    real(single), intent(inout)        :: xLo
    real(single), intent(out)          :: xHi, yAtXLo, yAtXHi
    integer,      intent(out)          :: status
!    real(single), intent(in), optional :: maxAllowedDeltaX

    interface
      function yFn(x)
        use types

        real(single), intent(in) :: x(:)
        real(single) :: yFn
      end function yFn
    end interface

    real(single) :: dummyX0(1), dummyDir(1)

    !-------------------------------------------------------------------
    ! Start

    dummyX0  = 0.0
    dummyDir = 1.0
    call findXBracketOfZeroYVector(dummyX0, dummyDir, maxNumIterations&
      , xLo, xHi, yAtXLo, yAtXHi, yFn, status)!, maxAllowedDeltaX)

  end subroutine findXBracketOfZeroYScalar

  !---------------------------------------------------------------------
  subroutine findXBracketOfZeroYVector(x0, dir, maxNumIterations, xLo, xHi&
  , yAtXLo, yAtXHi, yFn, status)!, maxAllowedDeltaX)
  !---------------------------------------------------------------------
  ! Adapted from the subroutine zbrac in chapter 9 of Numerical Recipes
  ! for Fortran, 2nd Ed., W H Press et al.

    real(single), intent(in)           :: x0(:), dir(size(x0))
    integer,      intent(in)           :: maxNumIterations
    real(single), intent(inout)        :: xLo
    real(single), intent(out)          :: xHi, yAtXLo, yAtXHi
    integer,      intent(out)          :: status
!    real(single), intent(in), optional :: maxAllowedDeltaX
!***** maxAllowedDeltaX not used yet!

    interface
      function yFn(x)
        use types

        real(single), intent(in) :: x(:)
        real(single) :: yFn
      end function yFn
    end interface

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    status = -1 ! default

    xHi = xLo + 1.0
    yAtXLo = yFn(x0 + dir * xLo)
    yAtXHi = yFn(x0 + dir * xHi)
    do i = 1, maxNumIterations
      if (yAtXLo * yAtXHi < 0.0) then
        status = 0
    exit
      end if

      if (abs(yAtXLo) < abs(yAtXHi)) then
        xLo = xLo + GoldenRatio * (xLo - xHi)
        yAtXLo = yFn(x0 + dir * xLo)
      else
        xHi = xHi + GoldenRatio * (xHi - xLo)
        yAtXHi = yFn(x0 + dir * xHi)
      end if
    end do

  end subroutine findXBracketOfZeroYVector

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module minimizations
