!
! Module: save_image
!
! Authors:  Ian Stewart - SSC/LUX
!        :  Masaaki Sakano - SSC/LUX
!          
! Description: Contains routines for saving old and new images.
! 
! Comments: TODO:
! 
! $Id: save_image.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module save_image

  use errorhandling
  use dal
  use caloaldefs	! for just consistency checking of INSTRUME
  use cal		! for just consistency checking of INSTRUME
  use dss

  use array_utils	! for type(aryEdgesInfoT)
  use dss_aux
  use wcs_aux
!use dal_aux

  implicit none 

  private 

  public replaceImage
  public saveImage
  public saveDetImage
  public getDetImageEdgesInfo

  interface replaceImage
    module procedure replaceImageSingle, replaceImageDouble, replaceImageBool
    module procedure replaceImageInt8,   replaceImageInt16,  replaceImageInt32
  end interface

  interface saveImage
    module procedure saveImageSingle, saveImageDouble, saveImageBool
    module procedure saveImageInt8,   saveImageInt16,  saveImageInt32
    module procedure saveImageLogical
  end interface

  interface saveDetImage
    module procedure saveDetImageSingle, saveDetImageDouble
    module procedure saveDetImageInt32,  saveDetImageInt16, saveDetImageInt8
  end interface

  interface getDetImageEdgesInfo
    module procedure getDetImageEdgesInfoSingle, getDetImageEdgesInfoDouble
    module procedure getDetImageEdgesInfoInt32,  getDetImageEdgesInfoInt16
    module procedure getDetImageEdgesInfoInt8
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine replaceImageSingle(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, replaces the
  ! 0th array with an array containing the new image, then adjusts wcs
  ! keywords if present to conform to the start pixels xStartArg and
  ! yStartArg of the new image.

    implicit none

    real(single),   intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    real(single), pointer ::&
      arrayDataReal32(:,:) => null()

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call replaceImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), REAL32&
      , set, newImageArray)

    arrayDataReal32 => real32Array2Data(newImageArray)
    arrayDataReal32 = newImage
    call release(set)

  end subroutine replaceImageSingle

  !---------------------------------------------------------------------
  subroutine replaceImageDouble(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, replaces the
  ! 0th array with an array containing the new image, then adjusts wcs
  ! keywords if present to conform to the start pixels xStartArg and
  ! yStartArg of the new image.

!***** add optional newPixelSize arguments.


    implicit none

    real(double),   intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    real(double), pointer ::&
      arrayDataReal64(:,:) => null()

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call replaceImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), REAL64&
      , set, newImageArray)

    arrayDataReal64 => real64Array2Data(newImageArray)
    arrayDataReal64 = newImage
    call release(set)

  end subroutine replaceImageDouble

  !---------------------------------------------------------------------
  subroutine replaceImageInt8(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, replaces the
  ! 0th array with an array containing the new image, then adjusts wcs
  ! keywords if present to conform to the start pixels xStartArg and
  ! yStartArg of the new image.

!***** add optional newPixelSize arguments.


    implicit none

    integer(int8),  intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    integer(int8), pointer ::&
      arrayDataInt8(:,:) => null()

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call replaceImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER8&
      , set, newImageArray)

    arrayDataInt8 => int8Array2Data(newImageArray)
    arrayDataInt8 = newImage
    call release(set)

  end subroutine replaceImageInt8

  !---------------------------------------------------------------------
  subroutine replaceImageInt16(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, replaces the
  ! 0th array with an array containing the new image, then adjusts wcs
  ! keywords if present to conform to the start pixels xStartArg and
  ! yStartArg of the new image.

!***** add optional newPixelSize arguments.


    implicit none

    integer(int16), intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    integer(int16), pointer ::&
      arrayDataInt16(:,:) => null()

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call replaceImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER16&
      , set, newImageArray)

    arrayDataInt16 => int16Array2Data(newImageArray)
    arrayDataInt16 = newImage
    call release(set)

  end subroutine replaceImageInt16

  !---------------------------------------------------------------------
  subroutine replaceImageInt32(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, replaces the
  ! 0th array with an array containing the new image, then adjusts wcs
  ! keywords if present to conform to the start pixels xStartArg and
  ! yStartArg of the new image.

!***** add optional newPixelSize arguments.


    implicit none

    integer(int32), intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    integer(int32), pointer ::&
      arrayDataInt32(:,:) => null()

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call replaceImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER32&
      , set, newImageArray)

    arrayDataInt32 => int32Array2Data(newImageArray)
    arrayDataInt32 = newImage
    call release(set)

  end subroutine replaceImageInt32

  !---------------------------------------------------------------------
  subroutine replaceImageBool(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, replaces the
  ! 0th array with an array containing the new image, then adjusts wcs
  ! keywords if present to conform to the start pixels xStartArg and
  ! yStartArg of the new image.
  !
  ! The output array is actually int8 rather than boolean.

!***** add optional newPixelSize arguments.



    implicit none

    logical(bool),  intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    integer(int8), pointer ::&
      arrayDataInt8(:,:) => null()

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call replaceImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), BOOLEAN&
      , set, newImageArray)

    arrayDataInt8 => int8Array2Data(newImageArray)
    arrayDataInt8 = 0
    where(newImage) arrayDataInt8 = 1

    call release(set)

  end subroutine replaceImageBool

  !---------------------------------------------------------------------
  subroutine replaceImageCommon(oldImageSetName, newImageSetName&
  , tempImageSetName, xStart, yStart, xSize, ySize, newDataType&
  , newSet, newArray)
  !---------------------------------------------------------------------
!*** The previous version seemed to be writing the new array to the primary HDU and tasks that use the dal could not open it. Let's try another way:
  ! NOTE: make sure to  call release(newSet)  after use.


!***** add optional newPixelSize arguments.

!*** allow for input of name or block number for the block to replace??


    implicit none

    character(*),   intent(in)  :: oldImageSetName,&
                                   newImageSetName,&
                                   tempImageSetName
    integer(int16), intent(in)  :: xStart,&
                                   yStart
    integer,        intent(in)  :: xSize,&
                                   ySize
    integer(int32), intent(in)  :: newDataType
    type(DataSetT), intent(out) :: newSet ! just so it can be released
    type(ArrayT),   intent(out) :: newArray

    type(DataSetT) :: refSet
    type(ArrayT) :: refArray
!    integer :: foreignStatus
    type(WcsT) :: setWcs, wcs
    logical :: wcsPresent = .false. ! default
    integer(int32), pointer ::&
      newDimVector(:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (oldImageSetName .eq. newImageSetName) then
      newSet = dataSet(oldImageSetName, MODIFY)
      refSet = clone(oldImageSetName, tempImageSetName, TEMP)

    else
      refSet = dataSet(oldImageSetName, READ)
      newSet = clone(oldImageSetName, newImageSetName, MODIFY)
    end if

    call deleteBlock(newSet, 0)
!*** allow for input of name or block number??

call release(newSet)
newSet = dataSet(newImageSetName, MODIFY)

    allocate(newDimVector(0:1))
    newDimVector = (/xSize, ySize/)

    newArray=addArray(newSet, "PRIMARY", newDataType, newDimVector)
    refArray = array(refSet, 0, READ)
    call setAttributes(attributable(newArray), attributable(refArray))
    call release(refSet)
    deallocate(newDimVector)

    ! Regularize and adjust the wcs:

!    call getWcsInfo(newSet, setWcs, foreignStatus)
!    if (foreignStatus == 0) then ! set contains the wcs keywords. Check
!    ! to see if the array block also has a set:
!
!      call getWcsInfo(newArray, wcs, foreignStatus)
!      if (foreignStatus == 0) then ! array also contains the wcs keywords.
!        call warning('setAlsoHasWcs','')
!
!      else ! array doesn't contains any wcs keywords. Copy over those from
!      ! the set:
!        call setAttribute(newArray, 'CRPIX1', setWcs%X%refPixels)
!        call setAttribute(newArray, 'CRPIX2', setWcs%Y%refPixels)
!        call setAttribute(newArray, 'CDELT1', setWcs%X%pixelSizeWorld)
!        call setAttribute(newArray, 'CDELT2', setWcs%Y%pixelSizeWorld)
!        call setAttribute(newArray, 'CRVAL1', setWcs%X%refWorld)
!        call setAttribute(newArray, 'CRVAL2', setWcs%Y%refWorld)
!        call setAttribute(newArray, 'CTYPE1', setWcs%X%ctype)
!        call setAttribute(newArray, 'CTYPE2', setWcs%Y%ctype)
!
!        call warning('setAlsoHasWcs','')
!      end if
!
!      wcsPresent = .true.
!
!    else ! set header doesn't contain any wcs keywords. Check the array:
!      call getWcsInfo(newArray, wcs, foreignStatus)
!      if (foreignStatus .ne. 0) wcsPresent = .true.
!    end if

if (hasWcs(newSet)) then ! set contains the wcs keywords.

  call readWcs(newSet, setWcs)
  ! Now check to see if the array block also has a set:

  if (hasWcs(newArray)) then ! array also contains the wcs keywords.
    call warning('setAlsoHasWcs','')

  else ! array doesn't contains any wcs keywords. Copy over those from
  ! the set:
!    call setAttribute(newArray, 'CRPIX1', setWcs%X%refPixels)
!    call setAttribute(newArray, 'CRPIX2', setWcs%Y%refPixels)
!    call setAttribute(newArray, 'CDELT1', setWcs%X%pixelSizeWorld)
!    call setAttribute(newArray, 'CDELT2', setWcs%Y%pixelSizeWorld)
!    call setAttribute(newArray, 'CRVAL1', setWcs%X%refWorld)
!    call setAttribute(newArray, 'CRVAL2', setWcs%Y%refWorld)
!    call setAttribute(newArray, 'CTYPE1', setWcs%X%ctype)
!    call setAttribute(newArray, 'CTYPE2', setWcs%Y%ctype)

call writeWcs(newArray, setWcs) ! ssclib/wcs_aux

    call warning('onlySetHadWcs','')
  end if

  wcsPresent = .true.

else ! set header doesn't contain any wcs keywords. Check the array:
  if (hasWcs(newArray)) wcsPresent = .true.
end if

    if (wcsPresent) then
      if (xStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX1', wcs%X%refPixels - (xStart - 1))
      end if
      if (yStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX2', wcs%Y%refPixels - (yStart - 1))
      end if
!*** all this assumes that the pixel sizes stay the same!
    end if

  end subroutine replaceImageCommon

  !---------------------------------------------------------------------
  subroutine saveImageSingle(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.

    implicit none

    real(single),   intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    real(single), pointer ::&
      arrayDataReal32(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), REAL32&
      , newSet, newArray, wcs)

    arrayDataReal32 => real32Array2Data(newArray)
    arrayDataReal32 = newImage

    call release(newSet)

  end subroutine saveImageSingle

  !---------------------------------------------------------------------
  subroutine saveImageDouble(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.

    implicit none

    real(double),   intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    real(double), pointer ::&
      arrayDataReal64(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), REAL64&
      , newSet, newArray, wcs)

    arrayDataReal64 => real64Array2Data(newArray)
    arrayDataReal64 = newImage

    call release(newSet)

  end subroutine saveImageDouble

  !---------------------------------------------------------------------
  subroutine saveImageBool(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.
  !
  ! The output array is actually int8 rather than boolean.

    implicit none

    logical(bool),  intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int8), pointer ::&
      arrayDataInt8(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER8&
      , newSet, newArray, wcs)
    ! Can't read or write an array of type BOOLEAN yet.

    arrayDataInt8 => int8Array2Data(newArray)
    arrayDataInt8 = 0
    where(newImage) arrayDataInt8 = 1

    call release(newSet)

  end subroutine saveImageBool

  !---------------------------------------------------------------------
  subroutine saveImageLogical(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.
  !
  ! The output array is actually int8 rather than boolean.

!*** replace the bool routine in ssclib/save_image by this!

    implicit none

    logical,  intent(in)                 :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int8), pointer ::&
      arrayDataInt8(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER8&
      , newSet, newArray, wcs)
    ! Can't read or write an array of type BOOLEAN yet.

    arrayDataInt8 => int8Array2Data(newArray)
    arrayDataInt8 = 0
    where(newImage) arrayDataInt8 = 1

    call release(newSet)

  end subroutine saveImageLogical

  !---------------------------------------------------------------------
  subroutine saveImageInt8(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.

    implicit none

    integer(int8),  intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int8), pointer ::&
      arrayDataInt8(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER8&
      , newSet, newArray, wcs)

    arrayDataInt8 => int8Array2Data(newArray)
    arrayDataInt8 = newImage

    call release(newSet)

  end subroutine saveImageInt8

  !---------------------------------------------------------------------
  subroutine saveImageInt16(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.

    implicit none

    integer(int16),  intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int16), pointer ::&
      arrayDataInt16(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER16&
      , newSet, newArray, wcs)

    arrayDataInt16 => int16Array2Data(newArray)
    arrayDataInt16 = newImage

    call release(newSet)

  end subroutine saveImageInt16

  !---------------------------------------------------------------------
  subroutine saveImageInt32(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.

    implicit none

    integer(int32), intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int32), pointer ::&
      arrayDataInt32(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER32&
      , newSet, newArray, wcs)

    arrayDataInt32 => int32Array2Data(newArray)
    arrayDataInt32 = newImage

    call release(newSet)

  end subroutine saveImageInt32

  !---------------------------------------------------------------------
  subroutine saveImageCommon(oldImageSetName, newImageSetName&
  , tempImageSetName, xStart, yStart, xSize, ySize, newDataType&
  , newSet, newArray, wcsArg)
  !---------------------------------------------------------------------

!***** add optional newPixelSize arguments.

!*** allow for input of name or block number of the ref block??


    implicit none

    character(*),   intent(in)  :: oldImageSetName,&
                                   newImageSetName,&
                                   tempImageSetName
    integer(int16), intent(in)  :: xStart,&
                                   yStart
    integer,        intent(in)  :: xSize,&
                                   ySize
    integer(int32), intent(in)  :: newDataType
    type(DataSetT), intent(out) :: newSet ! just so it can be released
    type(ArrayT),   intent(out) :: newArray
    type(WcsT),     intent(in), optional :: wcsArg

    type(DataSetT) :: refSet
    type(ArrayT) :: refArray
!    integer :: foreignStatus
    type(WcsT) :: setWcs, wcs
    logical :: wcsPresent = .false. ! default
    integer(int32), pointer ::&
      newDimVector(:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (oldImageSetName .eq. newImageSetName) then
      refSet = clone(oldImageSetName, tempImageSetName, TEMP)

    else
      refSet = dataSet(oldImageSetName, READ)
    end if
    newSet = dataSet(newImageSetName, CREATE)
    call setAttributes(attributable(newSet), attributable(refSet))

    allocate(newDimVector(0:1))
    newDimVector = (/xSize, ySize/)
    newArray = addArray(newSet, "PRIMARY", newDataType, newDimVector)
    deallocate(newDimVector)
    refArray = array(refSet, 0, READ)
    call setAttributes(attributable(newArray), attributable(refArray))

    if (hasDss(newArray)) then
      call dssDelete(block(newArray)) ! Gets rid of DSS attributes
      ! carried over from the ref set.
    end if

    call release(refSet)

    ! Regularize and adjust the wcs:

    if (present(wcsArg)) then
      call writeWcs(newArray, wcsArg) ! in ssclib/wcs_aux
      wcsPresent = .true.
    else
      if (hasWcs(newSet)) then ! set contains the wcs keywords.

        call readWcs(newSet, setWcs)
        ! Now check to see if the array block also has a set:

        if (hasWcs(newArray)) then ! array also contains the wcs keywords.
          call warning('setAlsoHasWcs','')

        else ! array doesn't contains any wcs keywords. Copy over those from
        ! the set:
          call deleteWcs(newSet) ! in ssclib/wcs_aux
          call writeWcs(newArray, setWcs) ! in ssclib/wcs_aux

          call warning('onlySetHadWcs','')
        end if

        wcsPresent = .true.

      else ! set header doesn't contain any wcs keywords. Check the array:
        if (hasWcs(newArray)) wcsPresent = .true.
      end if
    end if

    if (wcsPresent) then
      if (xStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX1', wcs%X%refPixels - (xStart - 1))
      end if
      if (yStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX2', wcs%Y%refPixels - (yStart - 1))
      end if
!*** all this assumes that the pixel sizes stay the same!
    end if

  end subroutine saveImageCommon


  !---------------------------------------------------------------------
  subroutine saveDetImageDouble(detImage, detImageSetName&
     , detImageEdgesInfo, templateFitsSetName, wcsExtended, detWcs &
     , strTelescop, strInstrum)
  !---------------------------------------------------------------------
  ! This subroutine writes a new image in the DETX/DETY coordinates.
  ! 
  ! wcsExtended can be given instead of, or in addition to, detImageEdgesInfo 
  ! in order to directly control the coordinate information in the header attributes. 
  ! Make sure wcsExtended%withPhysical==.true. if you want to add
  ! the PHYSICAL coordinates information in the output file.
  ! If both wcsExtended and detWcs are given, WCS information is overwritten
  ! according to detWcs at the end.  Obviously detWcs can not include
  ! any PHYSICAL coordinate information.
  ! 
  ! If templateFitsSetName is given, all the primary header attributes
  ! except for those for DSS and WCS are copied to the output file.
  ! TELESCOP attribute can be directly specified via strTelescop;
  ! otherwise, unless templateFitsSetName is given and has the attribute,
  ! the default 'XMM' is written in the output file.
  ! INSTRUME attribute can be directly specified via strInstrum (string),
  ! such as (EMOS1|EMOS1|EPN); in default this routine does nothing about it.

    implicit none

    real(double),   intent(in)           :: detImage(:,:)
    character(*),   intent(in)           :: detImageSetName
    type(aryEdgesInfoT), intent(in), optional :: detImageEdgesInfo	! defined in ssclib/array_utils
    character(*),   intent(in), optional :: templateFitsSetName, strTelescop, strInstrum
    type(WcsAxesExtendedT), intent(in), optional :: wcsExtended	! defined in wcs_aux
    type(WcsT),     intent(in), optional :: detWcs		! defined in wcs_aux

    type(aryEdgesInfoT) :: localDetImageEdgesInfo	! defined in ssclib/array_utils
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    real(double), pointer ::&
      arrayDataReal64(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (present(detImageEdgesInfo)) then
      localDetImageEdgesInfo = detImageEdgesInfo
    else
      localDetImageEdgesInfo = getDetImageEdgesInfo(detImage)
    end if

    call saveDetImageCommon(detImageSetName&
     , localDetImageEdgesInfo, REAL64, newSet, newArray &
     , templateFitsSetName=templateFitsSetName &
     , wcsExtended=wcsExtended, detWcs=detWcs &
     , strTelescop=strTelescop, strInstrum=strInstrum)

    arrayDataReal64 => real64Array2Data(newArray)
    arrayDataReal64 = detImage

    call release(newSet)

  end subroutine saveDetImageDouble


  !---------------------------------------------------------------------
  subroutine saveDetImageSingle(detImage, detImageSetName&
     , detImageEdgesInfo, templateFitsSetName, wcsExtended, detWcs &
     , strTelescop, strInstrum)
  !---------------------------------------------------------------------
  ! This subroutine writes a new image in the DETX/DETY coordinates.
  ! See the comments in saveDetImageDouble() for detail.

    implicit none

    real(single),   intent(in)           :: detImage(:,:)
    character(*),   intent(in)           :: detImageSetName
    type(aryEdgesInfoT), intent(in), optional :: detImageEdgesInfo	! defined in ssclib/array_utils
    character(*),   intent(in), optional :: templateFitsSetName, strTelescop, strInstrum
    type(WcsAxesExtendedT), intent(in), optional :: wcsExtended	! defined in wcs_aux
    type(WcsT),     intent(in), optional :: detWcs		! defined in wcs_aux

    type(aryEdgesInfoT) :: localDetImageEdgesInfo	! defined in ssclib/array_utils
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    real(single), pointer ::&
      arrayDataReal32(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (present(detImageEdgesInfo)) then
      localDetImageEdgesInfo = detImageEdgesInfo
    else
      localDetImageEdgesInfo = getDetImageEdgesInfo(detImage)
    end if

!print *,'DEBUG:common starting'
    call saveDetImageCommon(detImageSetName&
     , localDetImageEdgesInfo, REAL32, newSet, newArray &
     , templateFitsSetName=templateFitsSetName &
     , wcsExtended=wcsExtended, detWcs=detWcs &
     , strTelescop=strTelescop, strInstrum=strInstrum)

    arrayDataReal32 => real32Array2Data(newArray)
    arrayDataReal32 = detImage
!print *,'DEBUG:common ends.'

    call release(newSet)
!print *,'DEBUG:release(newSet) ends.'

  end subroutine saveDetImageSingle


  !---------------------------------------------------------------------
  subroutine saveDetImageInt32(detImage, detImageSetName&
     , detImageEdgesInfo, templateFitsSetName, wcsExtended, detWcs &
     , strTelescop, strInstrum)
  !---------------------------------------------------------------------
  ! This subroutine writes a new image in the DETX/DETY coordinates.
  ! See the comments in saveDetImageDouble() for detail.

    implicit none

    integer(int32), intent(in)           :: detImage(:,:)
    character(*),   intent(in)           :: detImageSetName
    type(aryEdgesInfoT), intent(in), optional :: detImageEdgesInfo	! defined in ssclib/array_utils
    character(*),   intent(in), optional :: templateFitsSetName, strTelescop, strInstrum
    type(WcsAxesExtendedT), intent(in), optional :: wcsExtended	! defined in wcs_aux
    type(WcsT),     intent(in), optional :: detWcs		! defined in wcs_aux

    type(aryEdgesInfoT) :: localDetImageEdgesInfo	! defined in ssclib/array_utils
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int32), pointer ::&
      arrayDataInt32(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (present(detImageEdgesInfo)) then
      localDetImageEdgesInfo = detImageEdgesInfo
    else
      localDetImageEdgesInfo = getDetImageEdgesInfo(detImage)
    end if

    call saveDetImageCommon(detImageSetName&
     , localDetImageEdgesInfo, INTEGER32, newSet, newArray &
     , templateFitsSetName=templateFitsSetName &
     , wcsExtended=wcsExtended, detWcs=detWcs &
     , strTelescop=strTelescop, strInstrum=strInstrum)

    arrayDataInt32 => int32Array2Data(newArray)
    arrayDataInt32 = detImage

    call release(newSet)

  end subroutine saveDetImageInt32


  !---------------------------------------------------------------------
  subroutine saveDetImageInt16(detImage, detImageSetName&
     , detImageEdgesInfo, templateFitsSetName, wcsExtended, detWcs &
     , strTelescop, strInstrum)
  !---------------------------------------------------------------------
  ! This subroutine writes a new image in the DETX/DETY coordinates.
  ! See the comments in saveDetImageDouble() for detail.

    implicit none

    integer(int16), intent(in)           :: detImage(:,:)
    character(*),   intent(in)           :: detImageSetName
    type(aryEdgesInfoT), intent(in), optional :: detImageEdgesInfo	! defined in ssclib/array_utils
    character(*),   intent(in), optional :: templateFitsSetName, strTelescop, strInstrum
    type(WcsAxesExtendedT), intent(in), optional :: wcsExtended	! defined in wcs_aux
    type(WcsT),     intent(in), optional :: detWcs		! defined in wcs_aux

    type(aryEdgesInfoT) :: localDetImageEdgesInfo	! defined in ssclib/array_utils
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int16), pointer ::&
      arrayDataInt16(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (present(detImageEdgesInfo)) then
      localDetImageEdgesInfo = detImageEdgesInfo
    else
      localDetImageEdgesInfo = getDetImageEdgesInfo(detImage)
    end if

    call saveDetImageCommon(detImageSetName&
     , localDetImageEdgesInfo, INTEGER16, newSet, newArray &
     , templateFitsSetName=templateFitsSetName &
     , wcsExtended=wcsExtended, detWcs=detWcs &
     , strTelescop=strTelescop, strInstrum=strInstrum)

    arrayDataInt16 => int16Array2Data(newArray)
    arrayDataInt16 = detImage

    call release(newSet)

  end subroutine saveDetImageInt16


  !---------------------------------------------------------------------
  subroutine saveDetImageInt8(detImage, detImageSetName&
     , detImageEdgesInfo, templateFitsSetName, wcsExtended, detWcs &
     , strTelescop, strInstrum)
  !---------------------------------------------------------------------
  ! This subroutine writes a new image in the DETX/DETY coordinates.
  ! See the comments in saveDetImageDouble() for detail.

    implicit none

    integer(int8), intent(in)           :: detImage(:,:)
    character(*),   intent(in)           :: detImageSetName
    type(aryEdgesInfoT), intent(in), optional :: detImageEdgesInfo	! defined in ssclib/array_utils
    character(*),   intent(in), optional :: templateFitsSetName, strTelescop, strInstrum
    type(WcsAxesExtendedT), intent(in), optional :: wcsExtended	! defined in wcs_aux
    type(WcsT),     intent(in), optional :: detWcs		! defined in wcs_aux

    type(aryEdgesInfoT) :: localDetImageEdgesInfo	! defined in ssclib/array_utils
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int8), pointer ::&
      arrayDataInt8(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    if (present(detImageEdgesInfo)) then
      localDetImageEdgesInfo = detImageEdgesInfo
    else
      localDetImageEdgesInfo = getDetImageEdgesInfo(detImage)
    end if

    call saveDetImageCommon(detImageSetName&
     , localDetImageEdgesInfo, INTEGER8, newSet, newArray &
     , templateFitsSetName=templateFitsSetName &
     , wcsExtended=wcsExtended, detWcs=detWcs &
     , strTelescop=strTelescop, strInstrum=strInstrum)

    arrayDataInt8 => int8Array2Data(newArray)
    arrayDataInt8 = detImage

    call release(newSet)

  end subroutine saveDetImageInt8


  !---------------------------------------------------------------------
  subroutine saveDetImageCommon(detImageSetName&
     , detImageEdgesInfo, newDataType, newSet, newArray &
     , templateFitsSetName, wcsExtended, detWcs &
     , strTelescop, strInstrum)
  !---------------------------------------------------------------------
  ! If wcsExtended is given, detImageEdgesInfo is ignored apart from 
  ! %arySize information.
  ! If both wcsExtended and detWcs are given, WCS information is overwritten
  ! according to detWcs .
  ! If strTelescop is a null string, and the input file does not have
  ! the attribute TELESCOP, the default value 'XMM' is written.
  ! If strInstrum is a null string, it is not written.

    implicit none

    character(*),   intent(in)      :: detImageSetName
    type(aryEdgesInfoT), intent(in) :: detImageEdgesInfo	! defined in ssclib/array_utils
    integer(int32), intent(in)  :: newDataType	! passed to addArray (defined in dal), eg., INTEGER32.
    type(DataSetT), intent(out) :: newSet	! just so it can be released in the calling routine.
    type(ArrayT),   intent(out) :: newArray
    character(*),   intent(in), optional :: templateFitsSetName, strTelescop, strInstrum
    type(WcsAxesExtendedT), intent(in), optional :: wcsExtended	! defined in wcs_aux
    type(WcsT),     intent(in), optional :: detWcs		! defined in wcs_aux

    type(DataSetT) :: refSet
    ! type(ArrayT) :: refArray
    integer(int32), pointer ::&
      newDimVector(:) => null()
    integer(int32) :: instrumentId
    character(5)  :: instrumentStrInCal	! 5 from caloalutils/src/caloaldefs.f90
    character(80) :: localStrTelescop

logical, parameter :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *,'...Start saveDetImageCommon()...'
if (testFlag) call printAryEdgesInfo(detImageEdgesInfo, 'detImageEdgesInfo')	! defined in array_utils

    if (detImageEdgesInfo%aryDimension < 1) then
      call printAryEdgesInfo(detImageEdgesInfo, 'detImageEdgesInfo')	! defined in array_utils
      call error('zeroAryDimension' &
         , 'ssclib:save_image:saveDetImageCommon() Contact the code developer.')
    end if

    newSet = dataSet(detImageSetName, CREATE)
    allocate(newDimVector(0:1))
    newDimVector = detImageEdgesInfo%arySize	! Array to Array operation.
    newArray = addArray(newSet, "PRIMARY", newDataType, newDimVector)
    deallocate(newDimVector)

if (testFlag) print *,'...Start templateFitsSetName...'

    if (present(templateFitsSetName)) then
      refSet = dataSet(templateFitsSetName, READ)
      call setAttributes(attributable(newSet), attributable(refSet))	! Copy the attributes in the header.
      ! refArray = array(refSet, 0, READ)	! This does not work if the input file is an event file etc.  Plus, this is not needed even if the input file is an image file.
      ! call setAttributes(attributable(newArray), attributable(refArray))

      if (hasDss(newArray)) then
        call dssDelete(block(newArray)) ! Gets rid of DSS attributes
        ! carried over from the ref set.
      end if

      call release(refSet)
    end if

! if (testFlag) print *,'...Start strTelescop...'

    localStrTelescop = ''
    if (present(strTelescop)) localStrTelescop = trim(strTelescop)
    if (localStrTelescop .ne. '') then
      call setAttribute(newSet, 'TELESCOP', trim(localStrTelescop), 'Telescope (mission) name')
    else if (.not. hasAttribute(newSet,   'TELESCOP') .and. &
             .not. hasAttribute(newArray, 'TELESCOP') ) then
      call message(AppMsg, Verbose, 'The attribute TELESCOP is not defined, hence is written as "XMM".')
      call setAttribute(newSet, 'TELESCOP', 'XMM', 'Telescope (mission) name')	! Default.
    end if

! if (testFlag) print *,'...Start strInstrum...'

    if (present(strInstrum)) then
      if (trim(strInstrum) .ne. '') then
        call setAttribute(newSet, 'INSTRUME', trim(strInstrum), 'Instrument name')
        call CAL_getState(instrument=instrumentId)
        instrumentStrInCal = instId(instrumentId)
        if (trim(strInstrum) .ne. trim(instrumentStrInCal)) then
          write(errStr, '(a)') 'Given string for Instrument (' &
             //trim(strInstrum)//') to saveDetImage is inconsistent with the CAL setting (' &
             //trim(instrumentStrInCal)//').'
          call warning('inconsistentInstrument')
        end if
      end if
    end if

! if (testFlag) print *,'...Start wcsExtended...'

    if (present(wcsExtended)) then
      call addDetCoordsAttr(newArray, wcsExtended=wcsExtended)	! Add the header attributes.
    else
      call addDetCoordsAttr(newArray, detImageEdgesInfo=detImageEdgesInfo)	! Add the header attributes.
    end if

! if (testFlag) print *,'...Start detWcs...'

    ! Overwrite the wcs, if given:

    if (present(detWcs)) then
      call writeWcs(newArray, detWcs) ! in ssclib/wcs_aux
    end if

  end subroutine saveDetImageCommon


  !---------------------------------------------------------------------
  subroutine addDetCoordsAttr(outArray, detImageEdgesInfo, wcsExtended)
  !---------------------------------------------------------------------
  ! Add the coordinate-related attributes for a DETX/Y image FITS file.
  ! Either of detImageEdgesInfo or wcsExtended must be given.
  ! If both are given, the latter (wcsExtended) is ignored.
  ! If you want to give the PHYSICAL coordinates with wcsExtended,
  ! make sure  wcsExtended%withPhysical==.true.

    implicit none

    type(ArrayT),   intent(in) :: outArray
    type(aryEdgesInfoT),    intent(in), optional :: detImageEdgesInfo	! defined in ssclib/array_utils
    type(WcsAxesExtendedT), intent(in), optional :: wcsExtended	! defined in ssclib/wcs_aux

    type(WcsAxesExtendedT) :: wcsE

    !-------------------------------------------------------------------
    ! Start

    if (present(detImageEdgesInfo)) then
      if (detImageEdgesInfo%aryDimension .ne. 2) call error('bug001InaddDetCoordsAttr', '')

      wcsE%withPhysical = .true.

      wcsE%CRPIXL(:) = 1
      wcsE%CDELTL(:) = (detImageEdgesInfo%uEdge - detImageEdgesInfo%lEdge) / detImageEdgesInfo%arySize	! Array operation.
      wcsE%CRVALL(:) = detImageEdgesInfo%lEdge + wcsE%CDELTL/2	! Array operation.  
      wcsE%CTYPEL(1) = 'DETX'
      wcsE%CTYPEL(2) = 'DETY'
      wcsE%CTYPE(1) = 'DETX_ANG'
      wcsE%CTYPE(2) = 'DETY_ANG'
      wcsE%CUNIT(:) = 'deg'
      wcsE%CDELT(:) = 0.0d0
      wcsE%CRPIX(:) = 0.0d0
      wcsE%CRVAL(:) = 0.0d0
      wcsE%LTV1   = 0.0d0
      wcsE%LTV2   = 0.0d0
      wcsE%LTM1_1 = 0.0d0
      wcsE%LTM1_2 = 0.0d0
      wcsE%LTM2_1 = 0.0d0
      wcsE%LTM2_2 = 0.0d0
      wcsE%CD1_1 = 0.0d0
      wcsE%CD1_2 = 0.0d0
      wcsE%CD2_1 = 0.0d0
      wcsE%CD2_2 = 0.0d0

    else if (present(wcsExtended)) then
      wcsE = wcsExtended

    else
      call error('notEnoughArgsInaddDetCoordsAttr','Contact the code developer (problem in ssclib/save_image).')
    end if

    call writeWcs(outArray, wcsE)	! defined in wcs_aux
    ! call dumpWcsExtended(wcsE, verbosityArg=0_int32, varName='wcsExtended_inSub')

  end subroutine addDetCoordsAttr


  !---------------------------------------------------------------------
  function getDetImageEdgesInfoDouble(detImage, outBinSizeXY, outOffsetXY) result(detImageEdgesInfo)
  !---------------------------------------------------------------------
  ! This returns the nominal detImageEdgesInfo for a given 2-dim array;
  ! only this function needs is a pair of sizes of array, but for the sake
  ! of practicality it accepts the (image) array.
  ! In default, the (0,0) in PHYSICAL is located at
  ! the centre of the array, and a bin width is 80.0 for both X/Y.
  ! If outOffsetXY=(a,b) is given, the (a,b) in PHYSICAL is located at 
  ! the centre of the array.

    implicit none

    real(double),  intent(in) :: detImage(:,:)
    real(single),  intent(in), dimension(2), optional :: outBinSizeXY, outOffsetXY	! 80.0 and 0.0 in default
    type(aryEdgesInfoT) :: detImageEdgesInfo	! defined in ssclib/array_utils

    real(single), dimension(2) :: localOutBinSizeXY, localOutOffsetXY	! 80.0 and 0.0 in default
    real(double) :: a, b
    real(double), dimension(2) :: lEdge, uEdge

    !-------------------------------------------------------------------
    ! Start

    if (present(outBinSizeXY)) then
      localOutBinSizeXY = outBinSizeXY
    else
      localOutBinSizeXY = 80.0	! Default
    end if
    if (present(outOffsetXY)) then
      localOutOffsetXY = outOffsetXY
    else
      localOutOffsetXY = 0.0	! Default
    end if

    a = size(detImage,1) / 2.0 * localOutBinSizeXY(1)
    b = size(detImage,2) / 2.0 * localOutBinSizeXY(2)
    lEdge = (/ -a+localOutOffsetXY(1), -b+localOutOffsetXY(2) /)
    uEdge = (/  a+localOutOffsetXY(1),  b+localOutOffsetXY(2) /)
    detImageEdgesInfo = getAryEdgesInfo(detImage, lEdge, uEdge)	! defined in ssclib/array_utils

  end function getDetImageEdgesInfoDouble


  !---------------------------------------------------------------------
  function getDetImageEdgesInfoSingle(detImage, outBinSizeXY, outOffsetXY) result(detImageEdgesInfo)
  !---------------------------------------------------------------------
  ! See the description in getDetImageEdgesInfoDouble() for detail.

    implicit none

    real(single),  intent(in) :: detImage(:,:)
    real(single),  intent(in), dimension(2), optional :: outBinSizeXY, outOffsetXY	! 80.0 and 0.0 in default
    type(aryEdgesInfoT) :: detImageEdgesInfo	! defined in ssclib/array_utils

    real(single), dimension(2) :: localOutBinSizeXY, localOutOffsetXY	! 80.0 and 0.0 in default
    real(double) :: a, b
    real(single), dimension(2) :: lEdge, uEdge

    !-------------------------------------------------------------------
    ! Start

    if (present(outBinSizeXY)) then
      localOutBinSizeXY = outBinSizeXY
    else
      localOutBinSizeXY = 80.0	! Default
    end if
    if (present(outOffsetXY)) then
      localOutOffsetXY = outOffsetXY
    else
      localOutOffsetXY = 0.0	! Default
    end if

    a = size(detImage,1) / 2.0 * localOutBinSizeXY(1)
    b = size(detImage,2) / 2.0 * localOutBinSizeXY(2)
    lEdge = (/ -a+localOutOffsetXY(1), -b+localOutOffsetXY(2) /)
    uEdge = (/  a+localOutOffsetXY(1),  b+localOutOffsetXY(2) /)
    detImageEdgesInfo = getAryEdgesInfo(detImage, lEdge, uEdge)	! defined in ssclib/array_utils

  end function getDetImageEdgesInfoSingle


  !---------------------------------------------------------------------
  function getDetImageEdgesInfoInt32(detImage, outBinSizeXY, outOffsetXY) result(detImageEdgesInfo)
  !---------------------------------------------------------------------
  ! See the description in getDetImageEdgesInfoDouble() for detail.

    implicit none

    integer(int32), intent(in) :: detImage(:,:)
    real(single),   intent(in), dimension(2), optional :: outBinSizeXY, outOffsetXY	! 80.0 and 0.0 in default
    type(aryEdgesInfoT) :: detImageEdgesInfo	! defined in ssclib/array_utils

    !-------------------------------------------------------------------
    ! Start

    detImageEdgesInfo = getDetImageEdgesInfoSingle(real(detImage, single), outBinSizeXY, outOffsetXY)

  end function getDetImageEdgesInfoInt32


  !---------------------------------------------------------------------
  function getDetImageEdgesInfoInt16(detImage, outBinSizeXY, outOffsetXY) result(detImageEdgesInfo)
  !---------------------------------------------------------------------
  ! See the description in getDetImageEdgesInfoDouble() for detail.

    implicit none

    integer(int16), intent(in) :: detImage(:,:)
    real(single),   intent(in), dimension(2), optional :: outBinSizeXY, outOffsetXY	! 80.0 and 0.0 in default
    type(aryEdgesInfoT) :: detImageEdgesInfo	! defined in ssclib/array_utils

    !-------------------------------------------------------------------
    ! Start

    detImageEdgesInfo = getDetImageEdgesInfoSingle(real(detImage, single), outBinSizeXY, outOffsetXY)

  end function getDetImageEdgesInfoInt16


  !---------------------------------------------------------------------
  function getDetImageEdgesInfoInt8(detImage, outBinSizeXY, outOffsetXY) result(detImageEdgesInfo)
  !---------------------------------------------------------------------
  ! See the description in getDetImageEdgesInfoDouble() for detail.

    implicit none

    integer(int8), intent(in) :: detImage(:,:)
    real(single),   intent(in), dimension(2), optional :: outBinSizeXY, outOffsetXY	! 80.0 and 0.0 in default
    type(aryEdgesInfoT) :: detImageEdgesInfo	! defined in ssclib/array_utils

    !-------------------------------------------------------------------
    ! Start

    detImageEdgesInfo = getDetImageEdgesInfoSingle(real(detImage, single), outBinSizeXY, outOffsetXY)

  end function getDetImageEdgesInfoInt8


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module save_image
