!
! Module: array_utils
!
! Authors:  Ian Stewart - SSC/LUX
!        :  Masaaki Sakano - SSC/LUX
!          
! Description: Contains routines to process arrays in various simple ways. The criteria for a subroutine or function to be placed here rather than elsewhere are:
!	(i) One of the input or output arguments must be an array, of any dimension >= 1, of any data type;
!	(ii) The operation performed on the array should either be a logical or simple arithmetic one.
! 
! Comments: TODO:
! 
! $Id: array_utils.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module array_utils

  use types		! For single, double, int16 etc type definitions.
  use dal		! For INTEGER_NULL
  use errorHandling

  use geometric_types	! for Point2dT type definition.
  use ssc_misc		! for swap.
  use sort_mod		! for findUniqueValues.

  implicit none 

  private 

  ! Take indices, return indices:
  public getAxesFromIndices
  public getIndicesFromAxes
  public calcIndicesFromIndices

  ! Take array, return indices:
  public maskArrayLimits
  public largestTrueIndex
  public smallestTrueIndex
  public getBinNumber
  public maxLoc1d
  public minLoc1d

  ! Take and return array:
  public findUniqueValues
  public vectorCross
  public normalizeVector
  public findEdges
  public selectArraySubset
  public cyclicConvolution
!  public scaleBalance
  public rotateArray
  public addPatch
  public makeGaussianConvolver
  public invertMask

  ! Take array, return scalars:
!  public linearInterpolate
!************* OBSOLETE! Use module interpolate.f90 now instead. IMS 2006-09-14.
  public trapezoidSum
  public getGridDiffs
  public interpolateGridDiff
!  public centreHistoMassMedian
  public median
  public valueAtHistoFraction
!  public preciseSum
!** can't see that this is any better than sum().
  public angleBetween
  public sumInt32
  public sumInt32Dbl
  public getUbound

  ! Misceloneous
  public printAryEdgesInfo
  public getAryEdgesInfo
  public calcAryStatInfoMask	! Subroutine
  public getAryStatInfo		! Function
  public convertAryStatIntToInt32
  public printAryStatInfo
  public getAnnularMaskAry


  type, public :: AryEdgesInfoT
    integer :: aryDimension = -1	! Uninitialised.
    integer,      allocatable :: arySize(:), lboundIndex(:)
    real(double), allocatable :: lEdge(:), uEdge(:)
  end type AryEdgesInfoT

  type, public :: aryStatInfoFlagT
     integer        :: status = 0	! Normal
     logical        :: isValLowerUsed = .false., isValUpperUsed = .false. &
        , isAreaLowerUsed = .false., isAreaUpperUsed = .false., isMaskUsed = .false.
  end type aryStatInfoFlagT

  !! type(aryStatInfo???T) ---------------------------------------------
  ! %totalsum              => double/single/int32
  ! %realsum               => double
  ! %mean, sigma           => double (but single for SingleT)
  ! %totalentry, validentry         => int32
  ! %minval, maxval        => double/single/int32/int16/int8
  ! %minindices, maxindices         => int32
  ! %vallower, valupper    => double (but single for SingleT)
  ! %minareaindices, maxareaindices => int32
  ! %flag%status                => Integer
  ! %flag%isValLowerUsed, isValUpperUsed, isAreaLowerUsed, isAreaUpperUsed, isMaskUsed => Logical
  !---------------------------------------------------------------------
  type, public :: aryStatInfoDoubleT
     real(double)   :: totalsum, mean, sigma
     real(double)   :: realsum
     integer(int32) :: totalentry, validentry
     real(double)   :: minval, maxval
     integer(int32), dimension(:), allocatable :: minindices, maxindices
     real(double)   :: vallower, valupper
     integer(int32), dimension(:), allocatable :: minareaindices, maxareaindices
     type(aryStatInfoFlagT) :: flag
  end type aryStatInfoDoubleT

  type, public :: aryStatInfoSingleT
     real(single)   :: totalsum, mean, sigma
     real(double)   :: realsum
     integer(int32) :: totalentry, validentry
     real(single)   :: minval, maxval
     integer(int32), dimension(:), allocatable :: minindices, maxindices
     real(single)   :: vallower, valupper
     integer(int32), dimension(:), allocatable :: minareaindices, maxareaindices
     type(aryStatInfoFlagT) :: flag
  end type aryStatInfoSingleT

  type, public :: aryStatInfoInt32T
     integer(int32) :: totalsum
     real(double)   :: realsum
     real(double)   :: mean, sigma
     integer(int32) :: totalentry, validentry
     integer(int32) :: minval, maxval
     integer(int32), dimension(:), allocatable :: minindices, maxindices
     real(double)   :: vallower, valupper
     integer(int32), dimension(:), allocatable :: minareaindices, maxareaindices
     type(aryStatInfoFlagT) :: flag
  end type aryStatInfoInt32T

  type, public :: aryStatInfoInt16T
     integer(int32) :: totalsum
     real(double)   :: realsum
     real(double)   :: mean, sigma
     integer(int32) :: totalentry, validentry
     integer(int16) :: minval, maxval
     integer(int32), dimension(:), allocatable :: minindices, maxindices
     real(double)   :: vallower, valupper
     integer(int32), dimension(:), allocatable :: minareaindices, maxareaindices
     type(aryStatInfoFlagT) :: flag
  end type aryStatInfoInt16T

  type, public :: aryStatInfoInt8T
     integer(int32) :: totalsum
     real(double)   :: realsum
     real(double)   :: mean, sigma
     integer(int32) :: totalentry, validentry
     integer(int8)  :: minval, maxval
     integer(int32), dimension(:), allocatable :: minindices, maxindices
     real(double)   :: vallower, valupper
     integer(int32), dimension(:), allocatable :: minareaindices, maxareaindices
     type(aryStatInfoFlagT) :: flag
  end type aryStatInfoInt8T

  interface getAxesFromIndices
    module procedure getAxesFromIndicesDouble, getAxesFromIndicesSingle
    module procedure getAxesFromIndicesEdgesDouble, getAxesFromIndicesEdgesSingle
  end interface

  interface getIndicesFromAxes
    module procedure getIndicesFromAxesSingle, getIndicesFromAxesDouble
    module procedure getIndicesFromAxesEdgesDouble, getIndicesFromAxesEdgesSingle
  end interface

  interface calcIndicesFromIndices
    module procedure calcIndicesFromIndicesDblDbl
    module procedure calcIndicesFromIndicesDblI32, calcIndicesFromIndicesDblI16
    module procedure calcIndicesFromIndicesSglSgl
    module procedure calcIndicesFromIndicesSglI32, calcIndicesFromIndicesSglI16
    module procedure calcIndicesFromIndicesI32I32
    module procedure calcIndicesFromIndicesI16I16
  end interface

  interface maskArrayLimits
    module procedure maskArrayLimitsLogical, maskArrayLimitsBool
  end interface

  interface largestTrueIndex
    module procedure largestTrueIndexBool, largestTrueIndexDefault
  end interface
  interface smallestTrueIndex
    module procedure smallestTrueIndexBool, smallestTrueIndexDefault
  end interface
  ! NOTE!!! At present 'logical' and 'logical(bool)' are different. If 'types' ever defines them the same, then this overloading will have to be deleted.

!  interface getBinNumber
!    module procedure getBinNumberSingle, getBinNumberDouble
!  end interface
  interface getBinNumber
    module procedure getBinNumberSingle_new_hunt, getBinNumberDouble_new_hunt
    module procedure getBinNumberSingle_new_locate
    module procedure getBinNumberDouble_new_locate
  end interface

  interface maxLoc1d
    module procedure maxLoc1dSingle, maxLoc1dDouble
  end interface
  interface minLoc1d
    module procedure minLoc1dSingle, minLoc1dDouble
  end interface

  interface findUniqueValues
    module procedure findUniqueValuesInt8, findUniqueValuesInt16
    module procedure findUniqueValuesInt32
  end interface

  interface vectorCross
    module procedure vectorCrossSingle, vectorCrossDouble
  end interface

  interface normalizeVector
    module procedure normalizeVectorSingle, normalizeVectorDouble
  end interface

  interface selectArraySubset
    module procedure selectArraySubsetInt8
    module procedure selectArraySubsetInt16
    module procedure selectArraySubsetInt32
    module procedure selectArraySubsetSingle
    module procedure selectArraySubsetDouble
!    module procedure selectArraySubsetBool
    module procedure selectArraySubsetLogical
  end interface

  interface rotateArray
    module procedure rotateArraySingle
  end interface
  interface addPatch
    module procedure addPatchPtrSingle
    module procedure addPatchSingle
  end interface

  interface invertMask
    module procedure invertMask2dLogical, invertMask2dBool
    module procedure invertMask2dInt8
    module procedure invertMask3dLogical, invertMask3dBool
    module procedure invertMask3dInt8
  end interface

  interface trapezoidSum
    module procedure trapezoidSumSingle, trapezoidSumDouble
  end interface

!  interface centreHistoMassMedian
!    module procedure cohmMedianInt8Vector, cohmMedianInt8Array
!    module procedure cohmMedianInt16Vector, cohmMedianInt16Array
!    module procedure cohmMedianInt32Vector, cohmMedianInt32Array
!  end interface

  interface median
    module procedure medianSingleVector, medianSingleArray
    module procedure medianDoubleVector, medianDoubleArray
  end interface

  interface valueAtHistoFraction
    module procedure valueAtHistoFracSingleArr, valueAtHistoFracDoubleArr
    module procedure valueAtHistoFracSingleVec, valueAtHistoFracDoubleVec
  end interface

!  interface preciseSum
!    module procedure preciseSumVectorSingle, preciseSumVectorDouble
!    module procedure preciseSumArraySingle, preciseSumArrayDouble
!  end interface

  interface angleBetween
    module procedure angleBetweenSingle, angleBetweenDouble
  end interface

  interface sumInt32
    module procedure sumInt32Int81d, sumInt32Int161d, sumInt32Int321d
    module procedure sumInt32Int82d, sumInt32Int162d, sumInt32Int322d
    module procedure sumInt32Int83d, sumInt32Int163d, sumInt32Int323d
    module procedure sumInt32Int84d, sumInt32Int164d, sumInt32Int324d
  end interface

  interface sumInt32Dbl
    module procedure sumInt32DblInt81d, sumInt32DblInt161d, sumInt32DblInt321d
    module procedure sumInt32DblInt82d, sumInt32DblInt162d, sumInt32DblInt322d
    module procedure sumInt32DblInt83d, sumInt32DblInt163d, sumInt32DblInt323d
    module procedure sumInt32DblInt84d, sumInt32DblInt164d, sumInt32DblInt324d
  end interface

  interface getUbound
    module procedure getUboundFromEdgesInfo, getUboundFromScalars
  end interface

  interface getAryEdgesInfo
    module procedure getAryEdgesInfoDbl1d, getAryEdgesInfoDbl2d, getAryEdgesInfoDbl3d 
    module procedure getAryEdgesInfoSgl1d, getAryEdgesInfoSgl2d, getAryEdgesInfoSgl3d 
    module procedure getAryEdgesInfoInt321d, getAryEdgesInfoInt322d, getAryEdgesInfoInt323d
    module procedure getAryEdgesInfoInt161d, getAryEdgesInfoInt162d, getAryEdgesInfoInt163d
    module procedure getAryEdgesInfoInt81d, getAryEdgesInfoInt82d, getAryEdgesInfoInt83d
  end interface

  interface calcAryStatInfoMask
    ! See the descriptions above the definition of type(aryStatInfo???T)
    module procedure calcAryStatInfoMaskDouble1d, calcAryStatInfoMaskSingle1d
    module procedure calcAryStatInfoMaskDouble2d, calcAryStatInfoMaskSingle2d
    module procedure calcAryStatInfoMaskInt322d, calcAryStatInfoMaskInt162d
    module procedure calcAryStatInfoMaskInt321d, calcAryStatInfoMaskInt161d
    module procedure calcAryStatInfoMaskInt82d
    module procedure calcAryStatInfoMaskInt81d
  end interface

  interface getAryStatInfo
    ! See the descriptions above the definition of type(aryStatInfo???T)
    module procedure getAryStatInfoDouble1d, getAryStatInfoSingle1d
    module procedure getAryStatInfoDouble2d, getAryStatInfoSingle2d
    module procedure getAryStatInfoInt322d, getAryStatInfoInt162d
    module procedure getAryStatInfoInt321d, getAryStatInfoInt161d
    module procedure getAryStatInfoInt82d
    module procedure getAryStatInfoInt81d
  end interface

  interface convertAryStatIntToInt32
    module procedure convertAryStatInt16ToInt32, convertAryStatInt8ToInt32
  end interface

  interface printAryStatInfo
    module procedure printAryStatInfoDouble, printAryStatInfoSingle
    module procedure printAryStatInfoInt32, printAryStatInfoInt16
    module procedure printAryStatInfoInt8
  end interface

  ! interface getAnnularMask
  !   module procedure getAnnularMaskAry
  !   module procedure getAnnularMaskDoubleName, getAnnularMaskDoubleSet
  !   ! module procedure getAnnularMaskSingleName, getAnnularMaskSingleSet
  !   ! module procedure getAnnularMaskInt32Name, getAnnularMaskInt32Set
  !   ! module procedure getAnnularMaskInt16Name, getAnnularMaskInt16Set
  !   ! module procedure getAnnularMaskInt8Name, getAnnularMaskInt8Set
  ! end interface

  character(6), parameter :: HPASI = "(A032,", HPASIALL = "(A032)"	! Global in this file.
  integer, parameter :: NCHAR4GETSTR4ASI = 1024

contains 

  !---------------------------------------------------------------------
  function getAxesFromIndicesDouble(indices, iLbound, iUbound, lEdge, uEdge) result(axes)
  !---------------------------------------------------------------------
  ! Return the axes (Double) for the input (i,j) for an array.
  !  indices(:) are the coordinates in unit of the index of the array of interest.
  !  i(L|U)bound(:) are the array of (l|u)bound of the array of interest.
  !  (l|u)Edge(:) are the array of the lower/upper bounds in unit of
  !  physically meaningful values of the array of interest; e.g.,
  !  lEdge=(0.5,0.5) uEdge=(256.5,256.5) etc.
  !
  ! If uEdge is not given, it is assumed that the width of axes is
  ! the same as the size of the array (= abs(iUbound-iLbound) for each axis).
  !
  ! Examples:
  !  Case 1: The axes for the indices (i,j)=(3,3) in the array
  !   (1:5,1:5) with the edge (0.5:5.5,0.5:5.5) is (3.0,3.0).
  !  Case 2: The axes for the indices (i,j)=(3,3) in the array
  !   (1:5,1:5) with the edge (2.5:7.5,2.5:7.5) is (5.0,5.0).
  !  Case 3: The axes for the indices (i,j)=(3,3) in the array
  !   (1:5,1:5) with the edge (0.0:10,0.0:10) is (5,5).
  !  Case 4: The axes for the indices (i,j)=(3,3) in the array
  !   (1:5,1:5) with the edge (-10:0.0,-10:0.0) is (-5,-5).
  !  Case 5: The axes for the indices (i,j)=(5,5) in the array
  !   (3:7,3:7) with the edge (0.0:10,0.0:10) is (5,5).

    implicit none

    real(double), intent(in) :: indices(:)
    integer,      intent(in)         :: iLbound(size(indices)), iUbound(size(indices))
    real(double), intent(in)           :: lEdge(size(indices))
    real(double), intent(in), optional :: uEdge(size(indices))
    real(double) :: axes(size(indices))	! result

    real(double) :: localUEdge(size(indices))
    real(double) :: unitAxe
    integer :: i, nDim

    !-------------------------------------------------------------------
    ! Start

    if (present(uEdge)) then
      localUEdge = uEdge
    else
      localUEdge = lEdge + iUbound-iLbound + 1	! Array operation
    end if
    nDim = size(indices)

    do i=1, nDim
      unitAxe = (localUEdge(i)-lEdge(i))/(iUbound(i)-iLbound(i)+1)	! Increment of the axe per pixel (=index)
      axes(i) = lEdge(i) - unitAxe/2.0 + unitAxe*(indices(i)-iLbound(i)+1)
    end do

  end function getAxesFromIndicesDouble


  !---------------------------------------------------------------------
  function getAxesFromIndicesSingle(indices, iLbound, iUbound, lEdge, uEdge) result(axes)
  !---------------------------------------------------------------------
  ! Return the axes (Single) for the input (i,j) for an array.
  ! See getAxesFromIndicesDouble() for detail.

    implicit none

    real(single), intent(in) :: indices(:)
    integer,      intent(in)         :: iLbound(size(indices)), iUbound(size(indices))
    real(single), intent(in)           :: lEdge(size(indices))
    real(single), intent(in), optional :: uEdge(size(indices))
    real(single) :: axes(size(indices))	! result

    real(double) :: indicesDbl(size(indices))
    real(double) :: lEdgeDbl(size(indices))

    !-------------------------------------------------------------------
    ! Start

    indicesDbl = real(indices, double)
    lEdgeDbl   = real(lEdge, double)
    if (present(uEdge)) then
      axes = getAxesFromIndicesDouble(indicesDbl, iLbound, iUbound, lEdgeDbl &
         , real(uEdge, double))
    else
      axes = getAxesFromIndicesDouble(indicesDbl, iLbound, iUbound, lEdgeDbl)
    end if

  end function getAxesFromIndicesSingle


  !---------------------------------------------------------------------
  function getAxesFromIndicesEdgesDouble(indices, aryEdgesInfo) result(axes)
  !---------------------------------------------------------------------
  ! Return the axes (Double) for the input (i,j) for an array.
  ! See getAxesFromIndicesDouble() for detail.

    implicit none

    real(double), intent(in) :: indices(:)
    type(AryEdgesInfoT), intent(in) :: aryEdgesInfo
    real(double) :: axes(size(indices))	! result

    !-------------------------------------------------------------------
    ! Start

    axes = getAxesFromIndicesDouble(indices &
       , aryEdgesInfo%lboundIndex, getUbound(aryEdgesInfo) &
       , aryEdgesInfo%lEdge &
       , aryEdgesInfo%uEdge)

  end function getAxesFromIndicesEdgesDouble


  !---------------------------------------------------------------------
  function getAxesFromIndicesEdgesSingle(indices, aryEdgesInfo) result(axes)
  !---------------------------------------------------------------------
  ! Return the axes (Single) for the input (i,j) for an array.
  ! See getAxesFromIndicesDouble() for detail.

    implicit none

    real(single), intent(in) :: indices(:)
    type(AryEdgesInfoT), intent(in) :: aryEdgesInfo
    real(single) :: axes(size(indices))	! result

    !-------------------------------------------------------------------
    ! Start

    axes = getAxesFromIndicesSingle(indices &
       ,      aryEdgesInfo%lboundIndex, getUbound(aryEdgesInfo) &
       , real(aryEdgesInfo%lEdge, single) &
       , real(aryEdgesInfo%uEdge, single))

  end function getAxesFromIndicesEdgesSingle


  !---------------------------------------------------------------------
  function getIndicesFromAxesDouble(axes, iLbound, iUbound, lEdge, uEdge) result(indices)
  !---------------------------------------------------------------------
  ! Return the indices(i,j) (Double) for the input axes(x,y) for an array.
  ! Indices, though the returned values here are Real, mean
  ! the index for the given array, therefore for an array
  !   ary(int(returned_i), int(returned_j))
  ! will give something significant in the Fortran code.
  ! Axes are arbitrary and give the frame, which may mean something physical.
  !
  ! If uEdge is not given, it is assumed that the width of axes is
  ! the same as the size of the array (= abs(iUbound-iLbound) for each axis).
  !
  ! See the comments in getAxesFromIndicesDouble() for the further detail
  ! re the meaning of those arguments.
  !
  ! Examples:
  !  Case 1: The indices for the axes (x,y)=(3.0,3.0) in the array
  !   (1:5,1:5) with the edge (0.5:5.5,0.5:5.5) is (3,3).
  !  Case 2: The indices for the axes (x,y)=(5.0,5.0) in the array
  !   (1:5,1:5) with the edge (2.5:7.5,2.5:7.5) is (3,3).
  !  Case 3: The indices for the axes (x,y)=(5.0,5.0) in the array
  !   (1:5,1:5) with the edge (0.0:10,0.0:10) is (3,3).
  !  Case 4: The indices for the axes (x,y)=(-5,-5) in the array
  !   (1:5,1:5) with the edge (-10:0.0,-10:0.0) is (3,3).
  !  Case 5: The indices for the axes (x,y)=(5.0,5.0) in the array
  !   (3:7,3:7) with the edge (0.0:10,0.0:10) is (5,5).

    implicit none

    real(double), intent(in) :: axes(:)
    integer, intent(in)              :: iLbound(size(axes)), iUbound(size(axes))
    real(double), intent(in)           :: lEdge(size(axes))
    real(double), intent(in), optional :: uEdge(size(axes))
    real(double) :: indices(size(axes))	! result

    real(double) :: localUEdge(size(axes))
    real(double) :: unitIndex
    integer :: i, nDim

    !-------------------------------------------------------------------
    ! Start

    if (present(uEdge)) then
      localUEdge = uEdge
    else
      localUEdge = lEdge + iUbound-iLbound + 1	! Array operation
    end if
    nDim = size(axes)

    do i=1, nDim
      unitIndex = real(iUbound(i)-iLbound(i)+1, double)/(localUEdge(i)-lEdge(i))	! Increment of the index per unit axis value
      indices(i) = iLbound(i) - 0.5 + unitIndex*(axes(i)-lEdge(i))
    end do

  end function getIndicesFromAxesDouble


  !---------------------------------------------------------------------
  function getIndicesFromAxesSingle(axes, iLbound, iUbound, lEdge, uEdge) result(indices)
  !---------------------------------------------------------------------
  ! Return the indices(i,j) (Single) for the input axes(x,y) for an array.
  ! See the description in getIndicesFromAxesDouble() for detail.

    implicit none

    real(single), intent(in) :: axes(:)
    integer, intent(in)              :: iLbound(size(axes)), iUbound(size(axes))
    real(single), intent(in)           :: lEdge(size(axes))
    real(single), intent(in), optional :: uEdge(size(axes))
    real(single) :: indices(size(axes))	! result

    real(double) :: axesDbl(size(axes))
    real(double) :: lEdgeDbl(size(axes))

    !-------------------------------------------------------------------
    ! Start

    axesDbl  = real(axes, double)
    lEdgeDbl = real(lEdge, double)
    if (present(uEdge)) then
      indices = getIndicesFromAxesDouble(axesDbl, iLbound, iUbound, lEdgeDbl &
         , real(uEdge, double))
    else
      indices = getIndicesFromAxesDouble(axesDbl, iLbound, iUbound, lEdgeDbl)
    end if

  end function getIndicesFromAxesSingle

  !---------------------------------------------------------------------
  function getIndicesFromAxesEdgesDouble(axes, aryEdgesInfo) result(indices)
  !---------------------------------------------------------------------
  ! Return the indices(i,j) (Double) for the input axes(x,y) for an array.
  ! See the description in getIndicesFromAxesDouble() for detail.

    implicit none

    real(double), intent(in) :: axes(:)
    type(AryEdgesInfoT), intent(in) :: aryEdgesInfo
    real(double) :: indices(size(axes))	! result

    !-------------------------------------------------------------------
    ! Start

    indices = getIndicesFromAxesDouble(axes &
       , aryEdgesInfo%lboundIndex, getUbound(aryEdgesInfo) &
       , aryEdgesInfo%lEdge &
       , aryEdgesInfo%uEdge)

  end function getIndicesFromAxesEdgesDouble

  !---------------------------------------------------------------------
  function getIndicesFromAxesEdgesSingle(axes, aryEdgesInfo) result(indices)
  !---------------------------------------------------------------------
  ! Return the indices(i,j) (Single) for the input axes(x,y) for an array.
  ! See the description in getIndicesFromAxesDouble() for detail.

    implicit none

    real(single), intent(in) :: axes(:)
    type(AryEdgesInfoT), intent(in) :: aryEdgesInfo
    real(single) :: indices(size(axes))	! result

    !-------------------------------------------------------------------
    ! Start

    indices = getIndicesFromAxesSingle(axes &
       ,      aryEdgesInfo%lboundIndex, getUbound(aryEdgesInfo) &
       , real(aryEdgesInfo%lEdge, single) &
       , real(aryEdgesInfo%uEdge, single))

  end function getIndicesFromAxesEdgesSingle


  !---------------------------------------------------------------------
  subroutine calcIndicesFromIndicesDblDbl(oldIndices, newIndices, oldLbound, newLbound)
  !---------------------------------------------------------------------
  ! Calculate the indices(i,j) (Double) on the new frame converted from
  ! those on the old frame.
  ! The most likely case is to get a pair of indices of an array
  ! in a subroutine/function, all the lbound-s of which must be 1,
  ! from the old pair of indices, where the lbound-s can be arbitrary.
  !
  ! Example (1-dimension):
  !   integer :: ary(5:8)
  !   ary(7) = -99	! i=7 is the 3rd element in the array.
  !   call aSub(iPoints=(/7/), lbounds=lbound(ary))
  !
  !   subroutine aSub(iPoints, lbounds)
  !     integer, intent(in) :: iPoints(:), lbounds(:)
  !     integer :: outAry(size(iPoints))
  !
  !     call getIndicesFromIndices(iPoints, outAry, lbounds)
  !     print *, outAry		! => 3
  !     ! nb., the old index=7 corresponds to index=3 in the current context.
  !   end subroutine aSub
  !

    implicit none

    real(double), intent(in)    :: oldIndices(:)
    real(double), intent(out)   :: newIndices(size(oldIndices))
    integer, intent(in)           :: oldLbound(size(oldIndices))
    integer, intent(in), optional :: newLbound(size(oldIndices))

    integer :: localNewLbound(size(oldIndices))

    !-------------------------------------------------------------------
    ! Start

    if (present(newLbound)) then
      localNewLbound = newLbound
    else
      localNewLbound = 1
    end if

    newIndices(:) = oldIndices - oldLbound + localNewLbound

  end subroutine calcIndicesFromIndicesDblDbl

  !---------------------------------------------------------------------
  subroutine calcIndicesFromIndicesDblI32(oldIndices, newIndices, oldLbound, newLbound)
  !---------------------------------------------------------------------
  ! See calcIndicesFromIndicesDblDbl for detail.

    implicit none

    real(double),   intent(in)  :: oldIndices(:)
    integer(int32), intent(out) :: newIndices(size(oldIndices))
    integer, intent(in)           :: oldLbound(size(oldIndices))
    integer, intent(in), optional :: newLbound(size(oldIndices))

    real(double) :: newIndicesDbl(size(oldIndices))

    !-------------------------------------------------------------------
    ! Start

    call calcIndicesFromIndicesDblDbl(oldIndices, newIndicesDbl, oldLbound, newLbound)
    newIndices(:) = nint(newIndicesDbl, int32)

  end subroutine calcIndicesFromIndicesDblI32

  !---------------------------------------------------------------------
  subroutine calcIndicesFromIndicesDblI16(oldIndices, newIndices, oldLbound, newLbound)
  !---------------------------------------------------------------------
  ! See calcIndicesFromIndicesDblDbl for detail.

    implicit none

    real(double),   intent(in)  :: oldIndices(:)
    integer(int16), intent(out) :: newIndices(size(oldIndices))
    integer, intent(in)           :: oldLbound(size(oldIndices))
    integer, intent(in), optional :: newLbound(size(oldIndices))

    real(double) :: newIndicesDbl(size(oldIndices))

    !-------------------------------------------------------------------
    ! Start

    call calcIndicesFromIndicesDblDbl(oldIndices, newIndicesDbl, oldLbound, newLbound)
    newIndices(:) = nint(newIndicesDbl, int16)

  end subroutine calcIndicesFromIndicesDblI16

  !---------------------------------------------------------------------
  subroutine calcIndicesFromIndicesSglSgl(oldIndices, newIndices, oldLbound, newLbound)
  !---------------------------------------------------------------------
  ! See calcIndicesFromIndicesDblDbl for detail.

    implicit none

    real(single),   intent(in)  :: oldIndices(:)
    real(single),   intent(out) :: newIndices(size(oldIndices))
    integer, intent(in)           :: oldLbound(size(oldIndices))
    integer, intent(in), optional :: newLbound(size(oldIndices))

    real(double) :: newIndicesDbl(size(oldIndices))

    !-------------------------------------------------------------------
    ! Start

    call calcIndicesFromIndicesDblDbl(real(oldIndices, double), newIndicesDbl &
       , oldLbound, newLbound)
    newIndices(:) = real(newIndicesDbl, single)

  end subroutine calcIndicesFromIndicesSglSgl

  !---------------------------------------------------------------------
  subroutine calcIndicesFromIndicesSglI32(oldIndices, newIndices, oldLbound, newLbound)
  !---------------------------------------------------------------------
  ! See calcIndicesFromIndicesDblDbl for detail.

    implicit none

    real(single),   intent(in)  :: oldIndices(:)
    integer(int32), intent(out) :: newIndices(size(oldIndices))
    integer, intent(in)           :: oldLbound(size(oldIndices))
    integer, intent(in), optional :: newLbound(size(oldIndices))

    real(double) :: newIndicesDbl(size(oldIndices))

    !-------------------------------------------------------------------
    ! Start

    call calcIndicesFromIndicesDblDbl(real(oldIndices, double), newIndicesDbl &
       , oldLbound, newLbound)
    newIndices(:) = nint(newIndicesDbl, int32)

  end subroutine calcIndicesFromIndicesSglI32

  !---------------------------------------------------------------------
  subroutine calcIndicesFromIndicesSglI16(oldIndices, newIndices, oldLbound, newLbound)
  !---------------------------------------------------------------------
  ! See calcIndicesFromIndicesDblDbl for detail.

    implicit none

    real(single),   intent(in)  :: oldIndices(:)
    integer(int16), intent(out) :: newIndices(size(oldIndices))
    integer, intent(in)           :: oldLbound(size(oldIndices))
    integer, intent(in), optional :: newLbound(size(oldIndices))

    real(double) :: newIndicesDbl(size(oldIndices))

    !-------------------------------------------------------------------
    ! Start

    call calcIndicesFromIndicesDblDbl(real(oldIndices, double), newIndicesDbl &
       , oldLbound, newLbound)
    newIndices(:) = nint(newIndicesDbl, int16)

  end subroutine calcIndicesFromIndicesSglI16

  !---------------------------------------------------------------------
  subroutine calcIndicesFromIndicesI32I32(oldIndices, newIndices, oldLbound, newLbound)
  !---------------------------------------------------------------------
  ! See calcIndicesFromIndicesDblDbl for detail.

    implicit none

    integer(int32), intent(in)  :: oldIndices(:)
    integer(int32), intent(out) :: newIndices(size(oldIndices))
    integer, intent(in)           :: oldLbound(size(oldIndices))
    integer, intent(in), optional :: newLbound(size(oldIndices))

    integer :: localNewLbound(size(oldIndices))

    !-------------------------------------------------------------------
    ! Start

    if (present(newLbound)) then
      localNewLbound = newLbound
    else
      localNewLbound = 1
    end if

    newIndices(:) = oldIndices - oldLbound + localNewLbound

  end subroutine calcIndicesFromIndicesI32I32

  !---------------------------------------------------------------------
  subroutine calcIndicesFromIndicesI16I16(oldIndices, newIndices, oldLbound, newLbound)
  !---------------------------------------------------------------------
  ! See calcIndicesFromIndicesDblDbl for detail.

    implicit none

    integer(int16), intent(in)  :: oldIndices(:)
    integer(int16), intent(out) :: newIndices(size(oldIndices))
    integer, intent(in)           :: oldLbound(size(oldIndices))
    integer, intent(in), optional :: newLbound(size(oldIndices))

    integer(int32) :: newIndicesI32(size(oldIndices))

    !-------------------------------------------------------------------
    ! Start

    call calcIndicesFromIndicesI32I32(int(oldIndices, int32), newIndicesI32 &
       , oldLbound, newLbound)
    newIndices(:) = int(newIndicesI32, int16)

  end subroutine calcIndicesFromIndicesI16I16


  !---------------------------------------------------------------------
  subroutine maskArrayLimitsLogical(mask, minXi, maxXi, minYi, maxYi, status)
  !---------------------------------------------------------------------
  ! Adapted from CGP's im_lims subroutine in implot-1.0.
  !
  ! Given a 2D logical array mask(), this subroutine returns the following:
  !
  !	minXi = the index of the first column which contains at least 1
  !	TRUE value;
  !
  !	maxXi = the index of the last column which contains at least 1
  !	TRUE value;
  !
  !	minYi = the index of the first row which contains at least 1 TRUE
  !	value;
  !
  !	maxYi = the index of the last row which contains at least 1 TRUE
  !	value.
  !
  ! Note that the index of the first row/column is 1, also that mask() is
  ! assumed to follow the position-vector format, ie the first index of
  ! mask() is the column number and the second the row number.
  !
  ! status, if given, is 0 in default.  If negative, something goes wrong.

    implicit none

!    logical(bool),  intent(in)            :: mask(:,:)
    logical,        intent(in)            :: mask(:,:)
    integer(int16), intent(out)           :: minXi, maxXi, minYi, maxYi
    integer,        intent(out), optional :: status

    ! integer(int16) :: localStatus
    integer(int16) :: xSize, ySize, xi, yi

    !-------------------------------------------------------------------
    ! Start

    ! Defaults:
    ! localStatus = 0
    minXi = 1
    maxXi = 0
    minYi = 1
    maxYi = 0
    if (present(status)) status = 0	! Default (= Normal end)
    
    if (count(mask) .le. 0) then
      if (present(status)) status = -1
  return
    end if

    xSize = size(mask, 1)
    ySize = size(mask, 2)

    if (xSize .le. 0 .or. ySize .le. 0) then
      if (present(status)) status = -2
  return
    end if

    do xi = 1, xSize
      if (any(mask(xi, :))) then
        minXi = xi
    exit
      end if
    end do

    do xi = xSize, 1, -1
      if (any(mask(xi, :))) then
        maxXi = xi
    exit
      end if
    end do

    do yi = 1, ySize
      if (any(mask(:, yi))) then
        minYi = yi
    exit
      end if
    end do

    do yi = ySize, 1, -1
      if (any(mask(:, yi))) then
        maxYi = yi
    exit
      end if
    end do

  end subroutine maskArrayLimitsLogical

  !---------------------------------------------------------------------
  subroutine maskArrayLimitsBool(mask, minXi, maxXi, minYi, maxYi, status)
  !---------------------------------------------------------------------

    implicit none

    logical(bool),  intent(in)            :: mask(:,:)
    integer(int16), intent(out)           :: minXi, maxXi, minYi, maxYi
    integer,        intent(out), optional :: status

    call maskArrayLimitsLogical(logical(mask), minXi, maxXi, minYi, maxYi, status)

  end subroutine maskArrayLimitsBool

  !---------------------------------------------------------------------
  integer function smallestTrueIndexBool(maskVector)
  !---------------------------------------------------------------------
  ! The functions smallestTrueIndex() and largestTrueIndex() are related
  ! to maskArrayLimits() but are much more basic. The function
  ! smallestTrueIndex() returns the index of the first TRUE value of the
  ! input logical vector maskVector(). The first index is assumed to be 1
  ! not 0.
  !
  ! If maskVector() contains no TRUE values, size(maskVector) + 1 is returned.

    implicit none

    logical(bool), intent(in) :: maskVector(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    smallestTrueIndexBool = size(maskVector) + 1
    do i = 1, size(maskVector)
    if (maskVector(i)) then
        smallestTrueIndexBool = i
    exit
      end if
    end do

  end function smallestTrueIndexBool

  !---------------------------------------------------------------------
  integer function smallestTrueIndexDefault(maskVector)
  !---------------------------------------------------------------------
  ! The functions smallestTrueIndex() and largestTrueIndex() are related
  ! to maskArrayLimits() but are much more basic. The function
  ! smallestTrueIndex() returns the index of the first TRUE value of the
  ! input logical vector maskVector(). The first index is assumed to be 1
  ! not 0.
  !
  ! If maskVector() contains no TRUE values, size(maskVector) + 1 is returned.

    implicit none

    logical, intent(in) :: maskVector(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    smallestTrueIndexDefault = size(maskVector) + 1
    do i = 1, size(maskVector)
    if (maskVector(i)) then
        smallestTrueIndexDefault = i
    exit
      end if
    end do

  end function smallestTrueIndexDefault

  !---------------------------------------------------------------------
  integer function largestTrueIndexBool(maskVector)
  !---------------------------------------------------------------------
  ! The functions smallestTrueIndex() and largestTrueIndex() are related
  ! to maskArrayLimits() but are much more basic. The function
  ! largestTrueIndex() returns the index of the last TRUE value of the
  ! input logical vector maskVector(). The first index is assumed to be 1
  ! not 0.
  !
  ! If maskVector() contains no TRUE values, 0 is returned. NOTE that the
  ! behaviour of getBinNumber() depends on this default, so don't change
  ! it!

    implicit none

    logical(bool), intent(in) :: maskVector(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    largestTrueIndexBool = 0
    do i = size(maskVector), 1, -1
      if (maskVector(i)) then
        largestTrueIndexBool = i
    exit
      end if
    end do

  end function largestTrueIndexBool

  !---------------------------------------------------------------------
  integer function largestTrueIndexDefault(maskVector)
  !---------------------------------------------------------------------
  ! The functions smallestTrueIndex() and largestTrueIndex() are related
  ! to maskArrayLimits() but are much more basic. The function
  ! largestTrueIndex() returns the index of the last TRUE value of the
  ! input logical vector maskVector(). The first index is assumed to be 1
  ! not 0.
  !
  ! If maskVector() contains no TRUE values, 0 is returned. NOTE that the
  ! behaviour of getBinNumber() depends on this default, so don't change
  ! it!

    implicit none

    logical, intent(in) :: maskVector(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    largestTrueIndexDefault = 0
    do i = size(maskVector), 1, -1
      if (maskVector(i)) then
        largestTrueIndexDefault = i
    exit
      end if
    end do

  end function largestTrueIndexDefault

  !---------------------------------------------------------------------
  function getBinNumberSingle_new_hunt(xBinEdges, x, lastBinNum)&
  result(binNum)
  !---------------------------------------------------------------------

    implicit none

    real(single),   intent(in) :: xBinEdges(:), x
    integer(int32), intent(in) :: lastBinNum
    integer(int32)             :: binNum

    !-------------------------------------------------------------------
    ! Start

    binNum = getBinNumberDouble_new_hunt(real(xBinEdges, double)&
      , real(x, double), lastBinNum)

  end function getBinNumberSingle_new_hunt

  !---------------------------------------------------------------------
  function getBinNumberSingle_new_locate(xBinEdges, x) result(binNum)
  !---------------------------------------------------------------------

    implicit none

    real(single),   intent(in) :: xBinEdges(:), x
    integer(int32)             :: binNum

    !-------------------------------------------------------------------
    ! Start

    binNum = getBinNumberDouble_new_locate(real(xBinEdges, double)&
      , real(x, double))

  end function getBinNumberSingle_new_locate

  !---------------------------------------------------------------------
  function getBinNumberDouble_new_hunt(xBinEdges, x, lastBinNum)&
  result(binNum)
  !---------------------------------------------------------------------
  ! This function is designed to return the index or number of the bin
  ! (the first bin is binNum = 1) in which the input value x falls. The
  ! function performs as follows:
  !
  ! It compares the value of x with the bin 'edges' stored in xBinEdges.
  !
  ! If x < xBinEdges(1), binNum = 0 is returned. It is up to the calling
  ! routine to catch these instances if it is necessary to prevent this
  ! return value from occurring.
  !
  ! If xBinEdges(i) <= x < xBinEdges(i+1), binNum = i is returned.
  !
  ! If xBinEdges(size(xBinEdges)-1) <= x <= xBinEdges(size(xBinEdges)),
  ! binNum = size(xBinEdges)-1 is returned.
  !
  ! If x > xBinEdges(size(xBinEdges)), binNum = size(xBinEdges) is returned.
  ! It is up to the calling routine to catch these instances if it is
  ! necessary to prevent this return value from occurring.
  !.....................................................................

!!!!!!!!! This is a new routine adapted from subroutine hunt() in chapter 3 of Numerical Recipes in Fortran, Press et al. NOTE that, unlike hunt(), the present routine requires xBinEdges to be monotonically increasing.

    implicit none

    real(double),   intent(in) :: xBinEdges(:), x
    integer(int32), intent(in) :: lastBinNum
    integer(int32)             :: binNum

    integer :: numBins, step, binLo, binHi
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '    aaa getBinNumberDouble_new_hunt'
if (testFlag) print *, '    bbb getBinNumberDouble_new_hunt lastBinNum=', lastBinNum

    numBins = size(xBinEdges)-1
    if (lastBinNum < 1 .or. lastBinNum > numBins) then ! input guess is
    ! not useful, go straight to bisection routine.
      binNum = getBinNumberDouble_new_locate(xBinEdges, x)
if (testFlag) print *, '    zzz getBinNumberDouble_new_hunt lastBinNum no good - went straight to locate.'
if (testFlag) print *, ''
  return
    end if

    step = 1
if (testFlag) print *, '    step=',step
if (testFlag) print *, '    x, lastBinNum, xBinEdges(lastBinNum):', x, lastBinNum, xBinEdges(lastBinNum)
if (testFlag) print *, '    x >= xBinEdges(lastBinNum)?'
    if (x >= xBinEdges(lastBinNum)) then ! hunt upward:
if (testFlag) print *, '    yes. hunt upward.'
      binLo = lastBinNum
if (testFlag) print *, '    binLo=',binLo
      do ! double size of step at each iteration.
        binHi = binLo + step
if (testFlag) print *, '    binHi=',binHi

if (testFlag) print *, '    binHi > numBins+1?'
        if (binHi > numBins+1) then ! done hunting, because off end of
        ! table; set up for bisection.
          binHi = numBins+1
if (testFlag) print *, '    yes. done up hunting, because off end of table; set up for bisection.'
if (testFlag) print *, '    binHi=',binHi
      exit
        end if
if (testFlag) print *, '    no.'

if (testFlag) print *, '    x < xBinEdges(binHi)?'
!        if (x < xBinEdges(binHi)) then ! done hunting; value bracketed.
        if (x < xBinEdges(binHi) .or. (binHi==numBins+1 .and. x==xBinEdges(binHi))) then ! done hunting; value bracketed.
if (testFlag) print *, '    yes. done up hunting; value bracketed.'
      exit
        end if

        ! If got to here, not done hunting.
        binLo = binHi
        step = step + step
if (testFlag) print *, '    no. not done up hunting. step=', step
      end do ! end of step-doubling loop.

    else ! hunt downward:
      binHi = lastBinNum
      do ! double size of step at each iteration.
        binLo = binHi - step

        if (binLo < 1) then ! done hunting, because off end of
        ! table; set up for bisection.
          binLo = 1
if (testFlag) print *, '    done down hunting, because off end of table; set up for bisection.'
      exit
        end if

        if (x >= xBinEdges(binLo)) then ! done hunting; value bracketed.
if (testFlag) print *, '    done down hunting; value bracketed.'
      exit
        end if

        ! If got to here, not done hunting.
        binHi = binLo
        step = step + step
if (testFlag) print *, '    not done down hunting. step=', step
      end do ! end of step-doubling loop.
    end if ! end decide hunt up or down.

    binNum = getBinNumberDouble_new_locate(xBinEdges(binLo:binHi), x)
    if (binNum > 0 .and. binNum <= numBins)&
      binNum = binNum + binLo - 1

if (testFlag) print *, '    binNum=', binNum
if (testFlag) print *, '    zzz getBinNumberDouble_new_hunt'
if (testFlag) print *, ''
  end function getBinNumberDouble_new_hunt

  !---------------------------------------------------------------------
  function getBinNumberDouble_new_locate(xBinEdges, x) result(binNum)
  !---------------------------------------------------------------------
  ! This function is designed to return the index or number of the bin
  ! (the first bin is binNum = 1) in which the input value x falls. The
  ! function performs as follows:
  !
  ! It compares the value of x with the bin 'edges' stored in xBinEdges.
  !
  ! If x < xBinEdges(1), binNum = 0 is returned. It is up to the calling
  ! routine to catch these instances if it is necessary to prevent this
  ! return value from occurring.
  !
  ! If xBinEdges(i) <= x < xBinEdges(i+1), binNum = i is returned.
  !
  ! If xBinEdges(size(xBinEdges)-1) <= x <= xBinEdges(size(xBinEdges)),
  ! binNum = size(xBinEdges)-1 is returned.
  !
  ! If x > xBinEdges(size(xBinEdges)), binNum = size(xBinEdges) is returned.
  ! It is up to the calling routine to catch these instances if it is
  ! necessary to prevent this return value from occurring.
  !.....................................................................

!!!!!!!!! This is a new routine adapted from subroutine locate() in chapter 3 of Numerical Recipes in Fortran, Press et al. NOTE that, unlike locate(), the present routine requires xBinEdges to be monotonically increasing.

    implicit none

    real(double),   intent(in) :: xBinEdges(:), x
    integer(int32)             :: binNum

    integer :: numEdges, binLo, binHi, binMid
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '      aaa getBinNumberDouble_new_locate'
    numEdges = size(xBinEdges)
if (testFlag) print *, '      numEdges, x, xBinEdges(numEdges):', numEdges, x, xBinEdges(numEdges)
if (testFlag) print *, '      x==xBinEdges(numEdges)?', (x==xBinEdges(numEdges))
    if (numEdges < 2) then
      binNum = 0
  return
    end if

    binLo = 0
    binHi = numEdges+1
    do while(binHi - binLo > 1)
      binMid = (binLo + binHi) / 2
      if (x >= xBinEdges(binMid)) then
        binLo = binMid
      else
        binHi = binMid
      end if
    end do

    if (x == xBinEdges(1)) then
      binNum = 1
    else if (x == xBinEdges(numEdges)) then
      binNum = numEdges - 1
if (testFlag) print *, '      XXXXXX binNum:', binNum
    else
      binNum = binLo
    end if
if (testFlag) print *, '      zzz getBinNumberDouble_new_locate'
!if (testFlag) print *, ''

  end function getBinNumberDouble_new_locate

  !---------------------------------------------------------------------
  function getBinNumberSingle_060519(xBinEdges, x) result(binNum)
  !---------------------------------------------------------------------
!*** replace the present array_utils sub by this. IMS 2006-05-19.

  ! This function is designed to return the index or number of the bin
  ! (the first bin is binNum = 1) in which the input value x falls. The
  ! function performs as follows:
  !
  ! It compares the value of x with the bin 'edges' stored in xBinEdges.
  !
  ! If x < xBinEdges(1), binNum = 0 is returned. It is up to the calling
  ! routine to catch these instances if it is necessary to prevent this
  ! return value from occurring.
  !
  ! If xBinEdges(i) <= x < xBinEdges(i+1), binNum = i is returned.
  !
  ! If xBinEdges(size(xBinEdges)-1) <= x <= xBinEdges(size(xBinEdges)),
  ! binNum = size(xBinEdges)-1 is returned.
  !
  ! If x > xBinEdges(size(xBinEdges)), binNum = size(xBinEdges) is returned.
  ! It is up to the calling routine to catch these instances if it is
  ! necessary to prevent this return value from occurring.
  !.....................................................................

    implicit none

    real(single), intent(in) :: xBinEdges(:), x
    integer(int32)           :: binNum

    !-------------------------------------------------------------------
    ! Start

    if (size(xBinEdges) < 2) then
      if (size(xBinEdges)==1) then
        if (x > xBinEdges(1)) then
          binNum = 1
        else
          binNum = 0
        end if
      else
        binNum = 0
      end if
  return
    end if

    binNum = largestTrueIndex(xBinEdges<=x)

  end function getBinNumberSingle_060519

  !---------------------------------------------------------------------
  function getBinNumberDouble_060519(xBinEdges, x) result(binNum)
  !---------------------------------------------------------------------
!*** replace the present array_utils sub by this. IMS 2006-05-19.

  ! This function is designed to return the index or number of the bin
  ! (the first bin is binNum = 1) in which the input value x falls. The
  ! function performs as follows:
  !
  ! It compares the value of x with the bin 'edges' stored in xBinEdges.
  !
  ! If x < xBinEdges(1), binNum = 0 is returned. It is up to the calling
  ! routine to catch these instances if it is necessary to prevent this
  ! return value from occurring.
  !
  ! If xBinEdges(i) <= x < xBinEdges(i+1), binNum = i is returned.
  !
  ! If xBinEdges(size(xBinEdges)-1) <= x <= xBinEdges(size(xBinEdges)),
  ! binNum = size(xBinEdges)-1 is returned.
  !
  ! If x > xBinEdges(size(xBinEdges)), binNum = size(xBinEdges) is returned.
  ! It is up to the calling routine to catch these instances if it is
  ! necessary to prevent this return value from occurring.
  !.....................................................................

    implicit none

    real(double), intent(in) :: xBinEdges(:), x
    integer(int32)           :: binNum

    !-------------------------------------------------------------------
    ! Start

    if (size(xBinEdges) < 2) then
      if (size(xBinEdges)==1) then
        if (x > xBinEdges(1)) then
          binNum = 1
        else
          binNum = 0
        end if
      else
        binNum = 0
      end if
  return
    end if

    binNum = largestTrueIndex(xBinEdges<=x)

  end function getBinNumberDouble_060519

  !---------------------------------------------------------------------
  function getBinNumberSingle_orig(xBinEdges, x) result(binNum)
  !---------------------------------------------------------------------
  ! This function is designed to return the index or number of the bin
  ! (the first bin is binNum = 1) in which the input value x falls. The
  ! function performs as follows:
  !
  ! It compares the value of x with the bin 'edges' stored in xBinEdges.
  !
  ! If x < xBinEdges(1), binNum = 0 is returned. It is up to the calling
  ! routine to catch these instances if it is necessary to prevent this
  ! return value from occurring.
  !
  ! If xBinEdges(i) <= x < xBinEdges(i+1), binNum = i is returned.
  !
  ! If xBinEdges(size(xBinEdges)-1) <= x <= xBinEdges(size(xBinEdges)),
  ! binNum = size(xBinEdges)-1 is returned.
  !
  ! If x > xBinEdges(size(xBinEdges)), binNum = size(xBinEdges) is returned.
  ! It is up to the calling routine to catch these instances if it is
  ! necessary to prevent this return value from occurring.
  !.....................................................................

    implicit none

    real(single), intent(in) :: xBinEdges(:), x
    integer(int32)           :: binNum

    !-------------------------------------------------------------------
    ! Start

    if (size(xBinEdges) < 2) then
      binNum = 0
  return
    end if

    binNum = largestTrueIndex(logical(xBinEdges <= x, bool))

  end function getBinNumberSingle_orig

  !---------------------------------------------------------------------
  function getBinNumberDouble_orig(xBinEdges, x) result(binNum)
  !---------------------------------------------------------------------
  ! This function is designed to return the index or number of the bin
  ! (the first bin is binNum = 1) in which the input value x falls. The
  ! function performs as follows:
  !
  ! It compares the value of x with the bin 'edges' stored in xBinEdges.
  !
  ! If x < xBinEdges(1), binNum = 0 is returned. It is up to the calling
  ! routine to catch these instances if it is necessary to prevent this
  ! return value from occurring.
  !
  ! If xBinEdges(i) <= x < xBinEdges(i+1), binNum = i is returned.
  !
  ! If xBinEdges(size(xBinEdges)-1) <= x <= xBinEdges(size(xBinEdges)),
  ! binNum = size(xBinEdges)-1 is returned.
  !
  ! If x > xBinEdges(size(xBinEdges)), binNum = size(xBinEdges) is returned.
  ! It is up to the calling routine to catch these instances if it is
  ! necessary to prevent this return value from occurring.
  !.....................................................................

    implicit none

    real(double), intent(in) :: xBinEdges(:), x
    integer(int32)           :: binNum

    !-------------------------------------------------------------------
    ! Start

    if (size(xBinEdges) < 2) then
      binNum = 0
  return
    end if

    binNum = largestTrueIndex(logical(xBinEdges <= x, bool))

  end function getBinNumberDouble_orig

  !---------------------------------------------------------------------
  function maxLoc1dSingle(vector, mask)
  !---------------------------------------------------------------------
  ! I wrote this because I was sick of the mucking around needed to use
  ! maxloc when the argument is a simple vector.
  !.....................................................................

    implicit none

    real(single), intent(in)           :: vector(:)
    logical,      intent(in), optional :: mask(size(vector))
    integer                            :: maxLoc1dSingle

    integer :: maxLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(mask)) then
      maxLocVec = maxloc(vector, mask)
    else
      maxLocVec = maxloc(vector)
    end if
    maxLoc1dSingle = maxLocVec(1)

  end function maxLoc1dSingle

  !---------------------------------------------------------------------
  function maxLoc1dDouble(vector, mask)
  !---------------------------------------------------------------------
  ! I wrote this because I was sick of the mucking around needed to use
  ! maxloc when the argument is a simple vector.
  !.....................................................................

    implicit none

    real(double), intent(in)           :: vector(:)
    logical,      intent(in), optional :: mask(size(vector))
    integer                            :: maxLoc1dDouble

    integer :: maxLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(mask)) then
      maxLocVec = maxloc(vector, mask)
    else
      maxLocVec = maxloc(vector)
    end if
    maxLoc1dDouble = maxLocVec(1)

  end function maxLoc1dDouble

  !---------------------------------------------------------------------
  function minLoc1dSingle(vector, mask)
  !---------------------------------------------------------------------
  ! I wrote this because I was sick of the mucking around needed to use
  ! minloc when the argument is a simple vector.
  !.....................................................................

    implicit none

    real(single), intent(in)           :: vector(:)
    logical,      intent(in), optional :: mask(size(vector))
    integer                            :: minLoc1dSingle

    integer :: minLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(mask)) then
      minLocVec = minloc(vector, mask)
    else
      minLocVec = minloc(vector)
    end if
    minLoc1dSingle = minLocVec(1)

  end function minLoc1dSingle

  !---------------------------------------------------------------------
  function minLoc1dDouble(vector, mask)
  !---------------------------------------------------------------------
  ! I wrote this because I was sick of the mucking around needed to use
  ! minloc when the argument is a simple vector.
  !.....................................................................

    implicit none

    real(double), intent(in)           :: vector(:)
    logical,      intent(in), optional :: mask(size(vector))
    integer                            :: minLoc1dDouble

    integer :: minLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(mask)) then
      minLocVec = minloc(vector, mask)
    else
      minLocVec = minloc(vector)
    end if
    minLoc1dDouble = minLocVec(1)

  end function minLoc1dDouble


  !---------------------------------------------------------------------
  function findUniqueValuesInt8(vector, mask) result(uniqueValues)
  !---------------------------------------------------------------------
  ! Note: you should call this function by pointer assignment, viz
  !
  !	myPointer => findUniqueValues(vector)
  !
  ! Remember to then deallocate myPointer after you've finished with it!

    implicit none

    integer(int8), intent(in)           :: vector(:)
    logical,       intent(in), optional :: mask(size(vector))
    integer(int8), pointer              :: uniqueValues(:)

    integer :: numValues, numLiveValues, i, numUniqueValues
    logical, allocatable :: localMask(:)
    integer(int8), allocatable :: tempVector(:), tempUniqueVals(:)

    !-------------------------------------------------------------------
    ! Start

    numValues = size(vector)
    allocate(localMask(numValues))
    if (present(mask)) then
      localMask = mask
    else
      localMask = .true.
    end if
    
    numLiveValues = count(localMask)
    if (numLiveValues <= 0) then
      allocate(uniqueValues(0))
return
    end if

    allocate(tempVector(numLiveValues), tempUniqueVals(numLiveValues))
    tempVector = pack(vector, localMask)
    call quickSort(tempVector) ! in ssclib/sort_mod.
    numUniqueValues = 1
    tempUniqueVals(numUniqueValues) = tempVector(1)
    do i = 2, numLiveValues
      if (tempVector(i) .ne. tempVector(i-1)) then
        numUniqueValues = numUniqueValues + 1
        tempUniqueVals(numUniqueValues) = tempVector(i)
      end if
    end do
    allocate(uniqueValues(numUniqueValues))
    uniqueValues = tempUniqueVals(:numUniqueValues)

  end function findUniqueValuesInt8

  !---------------------------------------------------------------------
  function findUniqueValuesInt16(vector, mask) result(uniqueValues)
  !---------------------------------------------------------------------
  ! Note: you should call this function by pointer assignment, viz
  !
  !	myPointer => findUniqueValues(vector)
  !
  ! Remember to then deallocate myPointer after you've finshed with it!

    implicit none

    integer(int16), intent(in)           :: vector(:)
    logical,        intent(in), optional :: mask(size(vector))
    integer(int16), pointer              :: uniqueValues(:)

    integer :: numValues, numLiveValues, i, numUniqueValues
    logical, allocatable :: localMask(:)
    integer(int16), allocatable :: tempVector(:), tempUniqueVals(:)

    !-------------------------------------------------------------------
    ! Start

    numValues = size(vector)
    allocate(localMask(numValues))
    if (present(mask)) then
      localMask = mask
    else
      localMask = .true.
    end if
    
    numLiveValues = count(localMask)
    if (numLiveValues <= 0) then
      allocate(uniqueValues(0))
return
    end if

    allocate(tempVector(numLiveValues), tempUniqueVals(numLiveValues))
    tempVector = pack(vector, localMask)
    call quickSort(tempVector) ! in ssclib/sort_mod.
    numUniqueValues = 1
    tempUniqueVals(numUniqueValues) = tempVector(1)
    do i = 2, numLiveValues
      if (tempVector(i) .ne. tempVector(i-1)) then
        numUniqueValues = numUniqueValues + 1
        tempUniqueVals(numUniqueValues) = tempVector(i)
      end if
    end do
    allocate(uniqueValues(numUniqueValues))
    uniqueValues = tempUniqueVals(:numUniqueValues)

  end function findUniqueValuesInt16

  !---------------------------------------------------------------------
  function findUniqueValuesInt32(vector, mask) result(uniqueValues)
  !---------------------------------------------------------------------
  ! Note: you should call this function by pointer assignment, viz
  !
  !	myPointer => findUniqueValues(vector)
  !
  ! Remember to then deallocate myPointer after you've finshed with it!

    implicit none

    integer(int32), intent(in)           :: vector(:)
    logical,        intent(in), optional :: mask(size(vector))
    integer(int32), pointer              :: uniqueValues(:)

    integer :: numValues, numLiveValues, i, numUniqueValues
    logical, allocatable :: localMask(:)
    integer(int32), allocatable :: tempVector(:), tempUniqueVals(:)

    !-------------------------------------------------------------------
    ! Start

    numValues = size(vector)
    allocate(localMask(numValues))
    if (present(mask)) then
      localMask = mask
    else
      localMask = .true.
    end if
    
    numLiveValues = count(localMask)
    if (numLiveValues <= 0) then
      allocate(uniqueValues(0))
return
    end if

    allocate(tempVector(numLiveValues), tempUniqueVals(numLiveValues))
    tempVector = pack(vector, localMask)
    call quickSort(tempVector) ! in ssclib/sort_mod.
    numUniqueValues = 1
    tempUniqueVals(numUniqueValues) = tempVector(1)
    do i = 2, numLiveValues
      if (tempVector(i) .ne. tempVector(i-1)) then
        numUniqueValues = numUniqueValues + 1
        tempUniqueVals(numUniqueValues) = tempVector(i)
      end if
    end do
    allocate(uniqueValues(numUniqueValues))
    uniqueValues = tempUniqueVals(:numUniqueValues)

  end function findUniqueValuesInt32


  !---------------------------------------------------------------------
  function vectorCrossSingle(vectorA, vectorB)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: vectorA(3), vectorB(3)
    real(single)             :: vectorCrossSingle(3)

    !-------------------------------------------------------------------
    ! Start

    vectorCrossSingle(1) = vectorA(2) * vectorB(3) - vectorA(3) * vectorB(2)
    vectorCrossSingle(2) = vectorA(3) * vectorB(1) - vectorA(1) * vectorB(3)
    vectorCrossSingle(3) = vectorA(1) * vectorB(2) - vectorA(2) * vectorB(1)

  end function vectorCrossSingle

  !---------------------------------------------------------------------
  function vectorCrossDouble(vectorA, vectorB)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: vectorA(3), vectorB(3)
    real(double)             :: vectorCrossDouble(3)

    !-------------------------------------------------------------------
    ! Start

    vectorCrossDouble(1) = vectorA(2) * vectorB(3) - vectorA(3) * vectorB(2)
    vectorCrossDouble(2) = vectorA(3) * vectorB(1) - vectorA(1) * vectorB(3)
    vectorCrossDouble(3) = vectorA(1) * vectorB(2) - vectorA(2) * vectorB(1)

  end function vectorCrossDouble

  !---------------------------------------------------------------------
  function normalizeVectorSingle(argument)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: argument(:)
    real(single) :: normalizeVectorSingle(size(argument)), vectorLength2
    integer :: i

    !-------------------------------------------------------------------
    ! Start

!*** use parallel facilities!

    vectorLength2 = 0.0
    do i = 1, size(argument)
      vectorLength2 = vectorLength2 + argument(i) * argument(i)
    end do
    if (vectorLength2 .le. 0.0) then
      normalizeVectorSingle = 0.0
    else
      normalizeVectorSingle = argument / sqrt(vectorLength2)
    end if

  end function normalizeVectorSingle

  !---------------------------------------------------------------------
  function normalizeVectorDouble(argument)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: argument(:)
    real(double)             :: normalizeVectorDouble(size(argument))

    real(double) :: vectorLength2
    integer :: i

    !-------------------------------------------------------------------
    ! Start

!*** use parallel facilities!

    vectorLength2 = 0.0d0
    do i = 1, size(argument)
      vectorLength2 = vectorLength2 + argument(i) * argument(i)
    end do
    if (vectorLength2 .le. 0.0d0) then
      normalizeVectorDouble = 0.0d0
    else
      normalizeVectorDouble = argument / sqrt(vectorLength2)
    end if

  end function normalizeVectorDouble

  !---------------------------------------------------------------------
  subroutine findEdges(mask, figureEdges, groundEdges)
  !---------------------------------------------------------------------
  ! Given a logical array mask(), this subroutine returns an array
  ! figureEdges() which is true at all true pixels in mask() which are
  ! _not_ surrounded by nearest neighbours which are all true. The
  ! 'nearest neighbours' of a pixel at(x,y) are defined as those
  ! positions (x',y') for which x-1<=x'<=x+1 and y-1<=y'<=y+1. However,
  ! values of (x',y') that fall outside the boundaries of mask() are
  ! not counted as nearest neighbours: hence, no pixel on the boundary
  ! of mask() is set as an edge pixel just because it is on the boundary.
  !
  ! A second array groundEdges() is also returned, which is equivalent
  ! to figureEdges() calculated for .not.mask().
  !
  ! Example input mask:
  !
  !	11110000111
  !	11111000011
  !	11111100011
  !	11111100001
  !	00000000000
  !	11111110000
  !	11101111100
  !
  ! The returned array figureEdges should look as follows:
  !	
  !	00010000110
  !	00011000010
  !	00001100011
  !	11111100001
  !	00000000000
  !	11111110000
  !	00101011100
  !
  ! The returned array groundEdges should look as follows:
  !	
  !	00001101000
  !	00000111100
  !	00000010100
  !	00000010110
  !	11111111011
  !	00000001110
  !	00010000010

    implicit none

    logical(bool), intent(in)  :: mask(:,:)
    logical(bool), intent(out), optional ::&
      figureEdges(size(mask, 1),size(mask, 2)),&
      groundEdges(size(mask, 1),size(mask, 2))

    logical(bool), allocatable :: edgeMaskBig(:,:), edgeMaskSmall(:,:)
    integer :: xSize, ySize

    !-------------------------------------------------------------------
    ! Start

    xSize = size(mask, 1)
    ySize = size(mask, 2)

    if (.not. present(figureEdges) .and. .not. present(groundEdges)) then
  return
    end if

    allocate(edgeMaskBig(0:xSize+1,0:ySize+1), edgeMaskSmall(xSize, ySize))

    edgeMaskSmall = .false. ! default

    edgeMaskBig = .false. ! default
    edgeMaskBig(0:xSize-1, 0:ySize-1) = mask
    edgeMaskBig(0:xSize-1, 1:ySize)&
      = edgeMaskBig(0:xSize-1, 1:ySize)   .or. mask
    edgeMaskBig(0:xSize-1, 2:ySize+1)&
      = edgeMaskBig(0:xSize-1, 2:ySize+1) .or. mask
    edgeMaskBig(1:xSize,   0:ySize-1)&
      = edgeMaskBig(1:xSize  , 0:ySize-1) .or. mask
    edgeMaskBig(1:xSize,   1:ySize)&
      = edgeMaskBig(1:xSize  , 1:ySize)   .or. mask
    edgeMaskBig(1:xSize,   2:ySize+1)&
      = edgeMaskBig(1:xSize  , 2:ySize+1) .or. mask
    edgeMaskBig(2:xSize+1, 0:ySize-1)&
      = edgeMaskBig(2:xSize+1, 0:ySize-1) .or. mask
    edgeMaskBig(2:xSize+1, 1:ySize)&
      = edgeMaskBig(2:xSize+1, 1:ySize)   .or. mask
    edgeMaskBig(2:xSize+1, 2:ySize+1)&
      = edgeMaskBig(2:xSize+1, 2:ySize+1) .or. mask

    edgeMaskSmall = edgeMaskBig(1:xSize, 1:ySize)

    edgeMaskBig = .false. ! default
    edgeMaskBig(0:xSize-1, 0:ySize-1) = .not. mask
    edgeMaskBig(0:xSize-1, 1:ySize)&
      = edgeMaskBig(0:xSize-1, 1:ySize)   .or. .not. mask
    edgeMaskBig(0:xSize-1, 2:ySize+1)&
      = edgeMaskBig(0:xSize-1, 2:ySize+1) .or. .not. mask
    edgeMaskBig(1:xSize,   0:ySize-1)&
      = edgeMaskBig(1:xSize  , 0:ySize-1) .or. .not. mask
    edgeMaskBig(1:xSize,   1:ySize)&
      = edgeMaskBig(1:xSize  , 1:ySize)   .or. .not. mask
    edgeMaskBig(1:xSize,   2:ySize+1)&
      = edgeMaskBig(1:xSize  , 2:ySize+1) .or. .not. mask
    edgeMaskBig(2:xSize+1, 0:ySize-1)&
      = edgeMaskBig(2:xSize+1, 0:ySize-1) .or. .not. mask
    edgeMaskBig(2:xSize+1, 1:ySize)&
      = edgeMaskBig(2:xSize+1, 1:ySize)   .or. .not. mask
    edgeMaskBig(2:xSize+1, 2:ySize+1)&
      = edgeMaskBig(2:xSize+1, 2:ySize+1) .or. .not. mask

    edgeMaskSmall = edgeMaskSmall .and. edgeMaskBig(1:xSize, 1:ySize)

    if (present(figureEdges)) then
      figureEdges = .false.
      figureEdges = edgeMaskSmall .and. mask
    end if
    if (present(groundEdges)) then
      groundEdges = .false.
      groundEdges = edgeMaskSmall .and. .not. mask
    end if

  end subroutine findEdges

  !---------------------------------------------------------------------
  subroutine selectArraySubsetInt8(image&
  , xStartArg, xFinisArg, yStartArg, yFinisArg)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.

    implicit none

    integer(int8), pointer :: image(:,:)
    integer(int16), intent(in) :: xStartArg, xFinisArg, yStartArg, yFinisArg

    integer(int8), pointer :: tempSingleArray(:,:) => null()
    integer(int16) :: xStart, xFinis, yStart, yFinis

    !-------------------------------------------------------------------
    ! Start

    xStart = max(lbound(image, 1), int(xStartArg))
    xFinis = min(ubound(image, 1), int(xFinisArg))
    yStart = max(lbound(image, 2), int(yStartArg))
    yFinis = min(ubound(image, 2), int(yFinisArg))

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetInt8

  !---------------------------------------------------------------------
  subroutine selectArraySubsetInt16(image&
  , xStartArg, xFinisArg, yStartArg, yFinisArg)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.

    implicit none

    integer(int16), pointer :: image(:,:)
    integer(int16), intent(in) :: xStartArg, xFinisArg, yStartArg, yFinisArg

    integer(int16), pointer :: tempSingleArray(:,:) => null()
    integer(int16) :: xStart, xFinis, yStart, yFinis

    !-------------------------------------------------------------------
    ! Start

    xStart = max(lbound(image, 1), int(xStartArg))
    xFinis = min(ubound(image, 1), int(xFinisArg))
    yStart = max(lbound(image, 2), int(yStartArg))
    yFinis = min(ubound(image, 2), int(yFinisArg))

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetInt16

  !---------------------------------------------------------------------
  subroutine selectArraySubsetInt32(image&
  , xStartArg, xFinisArg, yStartArg, yFinisArg)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.

    implicit none

    integer(int32), pointer :: image(:,:)
    integer(int16), intent(in) :: xStartArg, xFinisArg, yStartArg, yFinisArg

    integer(int32), pointer :: tempSingleArray(:,:) => null()
    integer(int16) :: xStart, xFinis, yStart, yFinis

    !-------------------------------------------------------------------
    ! Start

    xStart = max(lbound(image, 1), int(xStartArg))
    xFinis = min(ubound(image, 1), int(xFinisArg))
    yStart = max(lbound(image, 2), int(yStartArg))
    yFinis = min(ubound(image, 2), int(yFinisArg))

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetInt32

  !---------------------------------------------------------------------
  subroutine selectArraySubsetSingle(image, xStart, xFinis, yStart, yFinis)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.

    implicit none

    real(single), pointer :: image(:,:)
    integer(int16), intent(in) :: xStart, xFinis, yStart, yFinis

    real(single), pointer :: tempSingleArray(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetSingle

  !---------------------------------------------------------------------
  subroutine selectArraySubsetDouble(image, xStart, xFinis, yStart, yFinis)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.

    implicit none

    real(double), pointer :: image(:,:)
    integer(int16), intent(in) :: xStart, xFinis, yStart, yFinis

    real(double), pointer :: tempSingleArray(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetDouble

  !---------------------------------------------------------------------
  subroutine selectArraySubsetLogical(image, xStart, xFinis, yStart, yFinis)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.

    implicit none

    logical,        pointer    :: image(:,:)
    integer(int16), intent(in) :: xStart, xFinis, yStart, yFinis

    logical, pointer :: tempSingleArray(:,:)

    !-------------------------------------------------------------------
    ! Start

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetLogical

  !---------------------------------------------------------------------
  subroutine selectArraySubsetBool(image, xStart, xFinis, yStart, yFinis)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.


!*** deprecated


    implicit none

    logical(bool), pointer :: image(:,:)
    integer(int16), intent(in) :: xStart, xFinis, yStart, yFinis

    logical(bool), pointer :: tempSingleArray(:,:) => null()

    !-------------------------------------------------------------------
    ! Start

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetBool

  !---------------------------------------------------------------------
  subroutine cyclicConvolution(vector, convolver, mask, padPlease&
  , normalise, interpolationLimit, interpolationMask, uncerts, status)
  !---------------------------------------------------------------------

    implicit none

    real,          intent(in)               :: convolver(:)
    real,          intent(inout)            :: vector(:)
    real,          intent(inout),  optional :: uncerts(:)
    logical(bool), intent(in),     optional :: mask(:), padPlease, normalise
    integer,       intent(in),     optional :: interpolationLimit
    integer,       intent(out),    optional :: status
    logical(bool), intent(out),    optional :: interpolationMask(:)

    integer :: vSize, csize, n, i, j, localStatus, psize, cHalfWidth,&
      interpolateHolesSmallerThan, holeStart, holeFinis
    real, allocatable :: paddedVector(:), output(:)
    real :: summ, weight, norm
    logical :: localPadPlease, localNormalise, inNullPatch
    logical(bool), allocatable :: localMask(:), localInterpolationMask(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    vSize = size(vector)
    csize = size(convolver)
    cHalfWidth = csize / 2

    if (present(padPlease)) then
      localPadPlease = padPlease
    else
      localPadPlease = .true.
    end if
    if (present(normalise)) then
      localNormalise = normalise
    else
      localNormalise = .true.
    end if
    if (present(interpolationLimit)) then
      interpolateHolesSmallerThan = interpolationLimit
    else
      interpolateHolesSmallerThan = 0
    end if

    if (interpolateHolesSmallerThan .gt. csize) then
      if (present(status)) status = -1
  return
    end if

    if (localPadPlease) then
      psize = vSize + csize ! Pad doesn't need to be quite this big, but
      ! it won't hurt.
    else
      psize = vSize
    end if

    if (localNormalise) then
      norm = 1.0
    else
      norm = sum(convolver)
    end if

    allocate(paddedVector(psize), output(psize), localMask(psize)&
    , localInterpolationMask(psize))

    localMask = .false.
    if (present(mask)) then
      if (size(mask) .ne. vSize) then
        if (present(status)) status = -3
        deallocate(paddedVector, output, localMask, localInterpolationMask)
  return
      end if
      localMask(1:vSize) = mask

    else
      localMask(1:vSize) = .true.
    end if

    ! Check all the holes to see if any are small enough to be interpolated:
    localInterpolationMask = localMask ! vector default
    if (.not. localMask(1)) then
      inNullPatch = .true.
      holeStart = 1
    else
      inNullPatch = .false.
    end if
    do i = 2, vSize
      if (inNullPatch .and. localMask(i)) then
        holeFinis = i - 1
        if (holeFinis - holeStart + 1 .lt. interpolateHolesSmallerThan) then
          localInterpolationMask(holeStart:holeFinis) = .true.
        end if
        inNullPatch = .false.

      else if (.not. inNullPatch .and. .not. localMask(i)) then
        holeStart = i
        inNullPatch = .true.
      end if
    end do
    if (inNullPatch) then
      holeFinis = vSize
      if (holeFinis - holeStart + 1 .lt. interpolateHolesSmallerThan) then
        localInterpolationMask(holeStart:holeFinis) = .true.
      end if
    end if

    paddedVector(1:vSize) = vector
    output = 0.0 ! vector default.
    do n = 1, vSize
    if (.not. localInterpolationMask(n)) cycle
      summ = 0.0
      weight = 0.0
      do i = 1, csize
        j = n - cHalfWidth + i - 1
        if (j .le. 0)     j = j + psize
        if (j .gt. psize) j = j - psize
      if (.not. localMask(j)) cycle
        summ = summ + convolver(i) * paddedVector(j)
        weight = weight + convolver(i)
      end do
      if (weight .le. 0.0)  then
        if (present(status)) status = -4
        deallocate(paddedVector, output, localMask, localInterpolationMask)
  return
      end if
      output(n) = summ * norm / weight
    end do

    vector = output(1:vSize)

    if (present(uncerts)) then
      paddedVector(1:vSize) = uncerts * uncerts
      output = 0.0 ! vector default.
      do n = 1, vSize
      if (.not. localInterpolationMask(n)) cycle
        summ = 0.0
        weight = 0.0
        do i = 1, csize
          j = n - cHalfWidth + i - 1
          if (j .le. 0)     j = j + psize
          if (j .gt. psize) j = j - psize
        if (.not. localMask(j)) cycle
          summ = summ + convolver(i) * convolver(i) * paddedVector(j)
          weight = weight + convolver(i)
        end do
        if (weight .le. 0.0)  then
          output(n) = 0.0
        else
          output(n) = sqrt(summ) * norm / weight
        end if
      end do

      uncerts = output(1:vSize)
    end if

    if (present(interpolationMask))&
      interpolationMask = localInterpolationMask(1:vSize)

    deallocate(paddedVector, output, localMask, localInterpolationMask)

    if (present(status)) status = localStatus

  end subroutine cyclicConvolution

  !---------------------------------------------------------------------
  function rotateArraySingle(inArray, numQuarterRevs) result(outArray)
  !---------------------------------------------------------------------
  ! Rotates the array clockwise by numQuarterRevs steps. The array must be square - if not, it is returned unchanged.

!*** NOT YET TESTED!!

    implicit none

    real(single), intent(in) :: inArray(:,:)
    integer, intent(in) :: numQuarterRevs
    real(single) :: outArray(size(inArray,1),size(inArray,2))

    integer :: aSize

    !-------------------------------------------------------------------
    ! Start

    aSize = size(inArray, 1)
    if (size(inArray, 2) .ne. aSize .or. aSize <= 0) then
      outArray = inArray
  return
    end if

    select case(modulo(numQuarterRevs, 4))
    case(0)
      outArray = inArray
    case(1)
      outArray(:,aSize:1:-1) = transpose(inArray)
    case(2)
      outArray(aSize:1:-1,aSize:1:-1) = inArray
    case(3)
      outArray(aSize:1:-1,:) = transpose(inArray)
    end select

  end function rotateArraySingle

  !---------------------------------------------------------------------
  subroutine addPatchSingle(image, patchTemplate, xi, yi, multiplier)
  !---------------------------------------------------------------------
  ! The arguments (xi,yi) should be the coords of the pixel of image() at
  ! which you want the patch to be centred. The centre of the patch,
  ! taking just the x axis for the sake of example, is best illustrated
  ! by a few examples:
  !
  !	patch size:	centre pixel:
  !	    1		     1
  !	    2		     2
  !	    3		     2
  !	    6		     4
  !	    7		     4
  !	    8		     5
  !
  ! So, suppose you have size(image) = (103, 44) and size(patch) = (7,5).
  ! The patch centre pixel is (4,3). If you want the patch to be added to
  ! the image such that patch pixel (4,3) is added to image pixel (86, 9)
  ! you should supply xi=86 and yi=9.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout)        :: image(:,:)
    real(single), intent(in)           :: patchTemplate(:,:)
    integer,      intent(in)           :: xi, yi
    real(single), intent(in), optional :: multiplier

    integer ::&
      patchXiSize, patchHalfXiSize, patchFinisXi,&
      patchYiSize, patchHalfYiSize, patchFinisYi
    real(single), pointer :: patch(:,:)

    !-------------------------------------------------------------------
    ! Start

    patchXiSize = size(patchTemplate, 1)
    patchHalfXiSize = patchXiSize / 2
    patchFinisXi = patchXiSize - patchHalfXiSize - 1

    patchYiSize = size(patchTemplate, 2)
    patchHalfYiSize = patchYiSize / 2
    patchFinisYi = patchYiSize - patchHalfYiSize - 1

    allocate(patch(xi-patchHalfXiSize:xi+patchFinisXi,yi-patchHalfYiSize:yi+patchFinisYi))
    patch = patchTemplate
    call addPatchPtrSingle(image, patch, multiplier)
    deallocate(patch)

  end subroutine addPatchSingle

  !---------------------------------------------------------------------
  subroutine addPatchPtrSingle(image, patch, multiplier)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout)        :: image(:,:)
    real(single), pointer              :: patch(:,:)
    real(single), intent(in), optional :: multiplier

    integer :: xStart, xFinis, yStart, yFinis, xSize, ySize
    real(single) :: localMultiplier

    !-------------------------------------------------------------------
    ! Start

    localMultiplier = decideOption(multiplier, 1.0)

    xStart = lbound(patch, 1)
    xFinis = ubound(patch, 1)
    yStart = lbound(patch, 2)
    yFinis = ubound(patch, 2)
    xSize = size(image, 1)
    ySize = size(image, 2)

    if (xStart > xSize .or. xFinis < 1 .or. yStart > ySize .or. yFinis < 1)&
  return

    xStart = max(1, xStart)
    xFinis = min(xSize, xFinis)
    yStart = max(1, yStart)
    yFinis = min(ySize, yFinis)

    image(xStart:xFinis, yStart:yFinis) = image(xStart:xFinis, yStart:yFinis)&
      + patch(xStart:xFinis, yStart:yFinis) * localMultiplier

  end subroutine addPatchPtrSingle

  !---------------------------------------------------------------------
  subroutine makeGaussConv_possibly_use(sigmaXPixels, sigmaYPixels, halfXiSize&
  , halfYiSize, convolver)
  !---------------------------------------------------------------------
!****** adapted from calcGaussianConvolver in asmooth.

    implicit none

    real(single), intent(in) :: sigmaXPixels, sigmaYPixels
    integer,      intent(in) :: halfXiSize, halfYiSize
    real(single), pointer    :: convolver(:,:)

    real(single), allocatable :: xVals(:), yVals(:)
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    allocate(xVals(halfXiSize:halfXiSize), yVals(-halfYiSize:halfYiSize))

    if (sigmaXPixels > 0.0) then
      xVals(0:halfXiSize) = (/(i/sigmaXPixels, i=0,halfXiSize)/)
      xVals(0:halfXiSize) = exp(-xVals(0:halfXiSize) * xVals(0:halfXiSize) / 2.0)
      xVals(-halfXiSize:-1) = xVals(1:halfXiSize:-1)
    else
      xVals = 0.0
      xVals(0) = 1.0
    end if

    if (sigmaYPixels > 0.0) then
      yVals(0:halfYiSize) = (/(i/sigmaYPixels, i=0,halfYiSize)/)
      yVals(0:halfYiSize) = exp(-yVals(0:halfYiSize) * yVals(0:halfYiSize) / 2.0)
      yVals(-halfYiSize:-1) = yVals(1:halfYiSize:-1)
    else
      yVals = 0.0
      yVals(0) = 1.0
    end if

    convolver = spread(xVals, 2, 1+2*halfYiSize)&
      * spread(yVals, 1, 1+2*halfXiSize)

  end subroutine makeGaussConv_possibly_use

  !---------------------------------------------------------------------
  subroutine makeGaussianConvolver(sigmaXPixels, sigmaYPixels&
  , xCentre, yCentre, convolver)
  !---------------------------------------------------------------------
!****** adapted from calcGaussianConvolver in asmooth.

    implicit none

    real(single), intent(in)  :: sigmaXPixels, sigmaYPixels
    real(single), intent(in)  :: xCentre, yCentre
    real(single), intent(out) :: convolver(:,:)

    real(single) :: convolverCentreX, convolverCentreY
    integer :: convolverXiSize, convolverYiSize, i
    real(single), allocatable :: xVals(:), yVals(:)

    !-------------------------------------------------------------------
    ! Start:

    convolverXiSize = size(convolver, 1)
    convolverYiSize = size(convolver, 2)
    convolverCentreX = decideOption(xCentre, convolverXiSize / 2.0)
    convolverCentreY = decideOption(yCentre, convolverYiSize / 2.0)

    allocate(xVals(convolverXiSize), yVals(convolverYiSize))

    if (sigmaXPixels > 0.0) then
      xVals = (/((i-convolverCentreX)/sigmaXPixels, i=1,convolverXiSize)/)
      xVals = exp(-xVals * xVals / 2.0)
    else
      xVals = 0.0
      xVals(nint(convolverCentreX)) = 1.0
    end if

    if (sigmaYPixels > 0.0) then
      yVals = (/((i-convolverCentreY)/sigmaYPixels, i=1,convolverYiSize)/)
      yVals = exp(-yVals * yVals / 2.0)
    else
      yVals = 0.0
      yVals(nint(convolverCentreY)) = 1.0
    end if

    convolver = spread(xVals, 2, convolverYiSize)&
      * spread(yVals, 1, convolverXiSize)

  end subroutine makeGaussianConvolver


  !---------------------------------------------------------------------
  subroutine invertMask2dLogical(mask)
  !---------------------------------------------------------------------
  ! Invert the 2-dim mask array.

    implicit none

    logical, intent(inout) :: mask(:,:)

    !-------------------------------------------------------------------
    ! Start

    where(mask)
       mask = .false.
    elsewhere
       mask = .true.
    end where

  end subroutine invertMask2dLogical

  !---------------------------------------------------------------------
  subroutine invertMask2dBool(mask)
  !---------------------------------------------------------------------
  ! Invert the 2-dim mask array.

    implicit none

    logical(bool), intent(inout) :: mask(:,:)

    !-------------------------------------------------------------------
    ! Start

    where(mask)
       mask = .false._bool
    elsewhere
       mask = .true._bool
    end where

  end subroutine invertMask2dBool

  !---------------------------------------------------------------------
  subroutine invertMask2dInt8(mask)
  !---------------------------------------------------------------------
  ! Invert the 2-dim mask array.

    implicit none

    integer(int8), intent(inout) :: mask(:,:)

    !-------------------------------------------------------------------
    ! Start

    where(mask == 0)
       mask = 1
    elsewhere
       mask = 0
    end where

  end subroutine invertMask2dInt8

  !---------------------------------------------------------------------
  subroutine invertMask3dLogical(mask)
  !---------------------------------------------------------------------
  ! Invert the 3-dim mask array.

    implicit none

    logical, intent(inout) :: mask(:,:,:)

    !-------------------------------------------------------------------
    ! Start

    where(mask)
       mask = .false.
    elsewhere
       mask = .true.
    end where

  end subroutine invertMask3dLogical

  !---------------------------------------------------------------------
  subroutine invertMask3dBool(mask)
  !---------------------------------------------------------------------
  ! Invert the 3-dim mask array.

    implicit none

    logical(bool), intent(inout) :: mask(:,:,:)

    !-------------------------------------------------------------------
    ! Start

    where(mask)
       mask = .false._bool
    elsewhere
       mask = .true._bool
    end where

  end subroutine invertMask3dBool

  !---------------------------------------------------------------------
  subroutine invertMask3dInt8(mask)
  !---------------------------------------------------------------------
  ! Invert the 3-dim mask array.

    implicit none

    integer(int8), intent(inout) :: mask(:,:,:)

    !-------------------------------------------------------------------
    ! Start

    where(mask == 0)
       mask = 1
    elsewhere
       mask = 0
    end where

  end subroutine invertMask3dInt8


  !---------------------------------------------------------------------
  function linearInterpolate(x, y, xSample) result(ySample)
  !---------------------------------------------------------------------
  ! Note that x values must be monotonically increasing.

!************* OBSOLETE! Use module interpolate.f90 now instead. IMS 2006-09-14.

    implicit none

    real(single), intent(in) :: x(:), y(size(x)), xSample
    real(single) :: ySample

    integer :: numSamples, bin

    !-------------------------------------------------------------------
    ! Start

    numSamples = size(x)

    if (numSamples <= 0) then
      ySample = 0.0
  return
    end if

    if (numSamples == 1) then
      if (xSample == x(1)) then
        ySample = y(1)
      else
        ySample = 0.0
      end if
  return
    end if

    if (any(x(:numSamples-1) > x(2:))) then
      ySample = 0.0
  return
    end if

    if (xSample < x(1) .or. xSample > x(numSamples)) then
      ySample = 0.0
  return
    end if

    bin = getBinNumber(x, xSample)
    ySample = y(bin) + (xSample - x(bin)) * (y(bin+1) - y(bin))&
      / (x(bin+1) - x(bin))

  end function linearInterpolate

  !---------------------------------------------------------------------
  function trapezoidSumSingle(x, y) result(summ)
  !---------------------------------------------------------------------
  ! Given two vectors x() and y() containing samples of a function F, this
  ! subroutine returns the 'trapezoidal-rule' estimate of the integral of F.

    implicit none

    real(single), intent(in) :: x(:), y(size(x))
    real(single)             :: summ

    integer(int32) :: i, n

    !-------------------------------------------------------------------
    ! Start

    summ = 0.0 ! default
    n = size(x)
  if (n <= 1) return

    do i = 1, n - 1
      summ = summ + x(i + 1) * y(i) - x(i) * y(i + 1)
    end do
    summ = 0.5 * (x(n) * y(n) - x(1) * y(1) + summ)

  end function trapezoidSumSingle

  !---------------------------------------------------------------------
  function trapezoidSumDouble(x, y) result(summ)
  !---------------------------------------------------------------------
  ! Given two vectors x() and y() containing samples of a function F, this
  ! subroutine returns the 'trapezoidal-rule' estimate of the integral of F.

    implicit none

    real(double), intent(in) :: x(:), y(size(x))
    real(double)             :: summ

    integer(int32) :: i, n

    !-------------------------------------------------------------------
    ! Start

    summ = 0.0d0 ! default
    n = size(x)
  if (n <= 1) return

    do i = 1, n - 1
      summ = summ + x(i + 1) * y(i) - x(i) * y(i + 1)
    end do
    summ = 0.5 * (x(n) * y(n) - x(1) * y(1) + summ)

  end function trapezoidSumDouble

  !---------------------------------------------------------------------
  subroutine getGridDiffs(arrayOfOffsets, xi, yi, xDiff, yDiff, status)
  !---------------------------------------------------------------------
  ! This and the sub interpolateGridDiff() are designed to cope with interpolations within a 2d grid, for each point of which there is a defined transformed location, ie a location on another grid. The position of the interpolated point on the second grid can be estimated using (i) the positional on the second grid of a base point on the first, and (ii) the differential relations between the two grids at that point.
  !
  ! The task of the present subroutine is, given the two grids and a position on one, to estimate these differentials.

    implicit none

    type(Point2dT), intent(in)            :: arrayOfOffsets(:,:)
    integer,        intent(in)            :: xi, yi
    type(Point2dT), intent(out)           :: xDiff, yDiff
    integer,        intent(out), optional :: status
    ! Point2dT type defined in geometric_types.

    integer :: xSize, ySize

    !-------------------------------------------------------------------
    ! Start

    xDiff%x = 0.0
    xDiff%y = 0.0
    yDiff%x = 0.0
    yDiff%y = 0.0

    xSize = size(arrayOfOffsets, 1)
    ySize = size(arrayOfOffsets, 2)
    if (xSize < 2 .or. ySize < 2) then
      if (present(status)) status = -1
  return
    end if
    if (xi < 1 .or. xi > xSize .or. yi < 1 .or. yi > ySize) then
      if (present(status)) status = -2
  return
    end if

    if (xi == 1) then
      xDiff%x = arrayOfOffsets(2, yi)%x - arrayOfOffsets(1, yi)%x
      xDiff%y = arrayOfOffsets(2, yi)%y - arrayOfOffsets(1, yi)%y
    else if (xi == xSize) then
      xDiff%x = arrayOfOffsets(xSize, yi)%x - arrayOfOffsets(xSize - 1, yi)%x
      xDiff%y = arrayOfOffsets(xSize, yi)%y - arrayOfOffsets(xSize - 1, yi)%y
    else
      xDiff%x = (arrayOfOffsets(xi + 1, yi)%x - arrayOfOffsets(xi - 1, yi)%x)&
        / 2.0
      xDiff%y = (arrayOfOffsets(xi + 1, yi)%y - arrayOfOffsets(xi - 1, yi)%y)&
        / 2.0
    end if

    if (yi == 1) then
      yDiff%x = arrayOfOffsets(xi, 2)%x - arrayOfOffsets(xi, 1)%x
      yDiff%y = arrayOfOffsets(xi, 2)%y - arrayOfOffsets(xi, 1)%y
    else if (yi == ySize) then
      yDiff%x = arrayOfOffsets(xi, ySize)%x - arrayOfOffsets(xi, ySize - 1)%x
      yDiff%y = arrayOfOffsets(xi, ySize)%y - arrayOfOffsets(xi, ySize - 1)%y
    else
      yDiff%x = (arrayOfOffsets(xi, yi + 1)%x - arrayOfOffsets(xi, yi - 1)%x)&
        / 2.0
      yDiff%y = (arrayOfOffsets(xi, yi + 1)%y - arrayOfOffsets(xi, yi - 1)%y)&
        / 2.0
    end if

    if (present(status)) status = 0

  end subroutine getGridDiffs

  !---------------------------------------------------------------------
  subroutine interpolateGridDiff(offset, xDiff, yDiff, fracVector, fracOffsetVector)
  !---------------------------------------------------------------------
  ! This and the sub interpolateGridDiff() are designed to cope with interpolations within a 2d grid, for each point of which there is a defined transformed location, ie a location on another grid. The position of the interpolated point on the second grid can be estimated using (i) the positional on the second grid of a base point on the first, and (ii) the differential relations between the two grids at that point.
  !
  ! The task of the present subroutine is to perform the interpolation.

    implicit none

    type(Point2dT), intent(in) :: offset, xDiff, yDiff, fracVector(:)
    type(Point2dT), intent(out) :: fracOffsetVector(size(fracVector))
    ! Point2dT type defined in geometric_types.

    !-------------------------------------------------------------------
    ! Start

    fracOffsetVector%x = offset%x + fracVector%x&
      * xDiff%x + fracVector%y * yDiff%x
    fracOffsetVector%y = offset%y + fracVector%x&
      * xDiff%y + fracVector%y * yDiff%y

  end subroutine interpolateGridDiff


!  !---------------------------------------------------------------------
!  real(single) function cohmMedianInt32Vector(vector, maskVector)&
!  result(cohmMedian)
!  !---------------------------------------------------------------------
!  ! A histogram h(n) as a function of integer n is first made of the values in the input integer vector. The result r is then calculated such that the sum of the histogram values for n<r equals the sum for n>r. For n such that n<=r and (n+1) > r, the values in this nth histogram bin are divided, h(n)*(r-n) being added to the 'lefthand' weights, h(n)*(n+1-r) to the 'righthand'. This is approximately the same as a median where typical values of n are large. However the bin-splitting property makes it a useful extension of the median concept to cases where most n are small (ie of order 1).
!
!    implicit none
!
!    integer(int32), intent(in)           :: vector(:)
!    logical,        intent(in), optional :: maskVector(:)
!
!    real(single)                         :: cohmMedian
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    call makeHistogram()
!
!    medianBin = medianInt32Vector(vector, maskVector)
!    lhArea = 
!    do while()
!      medianBin = medianBin + 1
!    end do
!
!  end function cohmMedianInt32Vector


  !---------------------------------------------------------------------
  real(single) function medianSingleArray(array, maskArray)
  !---------------------------------------------------------------------
  ! NOTE that for EVEN-numbered size(array,1)*size(array,2) the function
  ! returns the value in the lower central bin of the sorted vector, rather
  ! than the (more correct) mean of the lower and upper central bins. For
  ! ODD-numbered size the function returns the value in the central bin of
  ! the sorted vector, as expected.

    implicit none

    real(single), intent(in)           :: array(:,:)
    logical,      intent(in), optional :: maskArray(:,:)

    integer :: vectorShape(1)

    !-------------------------------------------------------------------
    ! Start

    vectorShape(1) = size(array, 1) * size(array, 2)
    if (present(maskArray)) then
      medianSingleArray = valueAtHistoFracSingleVec(reshape(array, vectorShape)&
        , 0.5, reshape(maskArray, vectorShape))
    else
      medianSingleArray = valueAtHistoFracSingleVec(reshape(array, vectorShape)&
        , 0.5)
    end if

  end function medianSingleArray

  !---------------------------------------------------------------------
  real(double) function medianDoubleArray(array, maskArray)
  !---------------------------------------------------------------------
  ! NOTE that for EVEN-numbered size(array,1)*size(array,2) the function
  ! returns the value in the lower central bin of the sorted vector, rather
  ! than the (more correct) mean of the lower and upper central bins. For
  ! ODD-numbered size the function returns the value in the central bin of
  ! the sorted vector, as expected.

    implicit none

    real(double), intent(in)           :: array(:,:)
    logical,      intent(in), optional :: maskArray(:,:)

    integer :: vectorShape(1)

    !-------------------------------------------------------------------
    ! Start

    vectorShape(1) = size(array, 1) * size(array, 2)
    if (present(maskArray)) then
      medianDoubleArray = valueAtHistoFracDoubleVec(reshape(array, vectorShape)&
        , 0.5d0, reshape(maskArray, vectorShape))
    else
      medianDoubleArray = valueAtHistoFracDoubleVec(reshape(array, vectorShape)&
        , 0.5d0)
    end if

  end function medianDoubleArray

  !---------------------------------------------------------------------
  real(single) function medianSingleVector(vector, maskVector)
  !---------------------------------------------------------------------
  ! NOTE that for EVEN-numbered size(vector) the function returns the value
  ! in the lower central bin of the sorted vector, rather than the (more
  ! correct) mean of the lower and upper central bins. For ODD-numbered size
  ! the function returns the value in the central bin of the sorted vector,
  ! as expected.

    implicit none

    real(single), intent(in)           :: vector(:)
    logical,      intent(in), optional :: maskVector(:)

    !-------------------------------------------------------------------
    ! Start

    medianSingleVector = valueAtHistoFracSingleVec(vector, 0.5, maskVector)

  end function medianSingleVector

  !---------------------------------------------------------------------
  real(double) function medianDoubleVector(vector, maskVector)
  !---------------------------------------------------------------------
  ! NOTE that for EVEN-numbered size(vector) the function returns the value
  ! in the lower central bin of the sorted vector, rather than the (more
  ! correct) mean of the lower and upper central bins. For ODD-numbered size
  ! the function returns the value in the central bin of the sorted vector,
  ! as expected.

    implicit none

    real(double), intent(in)           :: vector(:)
    logical,      intent(in), optional :: maskVector(:)

    !-------------------------------------------------------------------
    ! Start

    medianDoubleVector = valueAtHistoFracDoubleVec(vector, 0.5d0, maskVector)

  end function medianDoubleVector

  !---------------------------------------------------------------------
  real(single) function valueAtHistoFracSingleVec(vector, binFraction&
    , maskVector)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: vector(:), binFraction
    logical,      intent(in), optional :: maskVector(:)

    !-------------------------------------------------------------------
    ! Start

    valueAtHistoFracSingleVec = real(valueAtHistoFracDoubleVec(&
      real(vector, double), real(binFraction, double), maskVector), single)

  end function valueAtHistoFracSingleVec

  !---------------------------------------------------------------------
  real(double) function valueAtHistoFracDoubleVec(vector, binFraction&
    , maskVector)
  !---------------------------------------------------------------------
  ! This function takes a vector of real values, sorts it in increasing
  ! order, and returns that element whose bin index (starting at 1) in
  ! the sorted vector, when divided by the total number of bins in the
  ! vector, is closest to the argument binFraction. Eg to find the median
  ! value of a vector one could call this function with binFraction set
  ! to 0.5.

!****maskVector mods not yet tested

    implicit none

    real(double), intent(in)           :: vector(:), binFraction
    logical,      intent(in), optional :: maskVector(:)

    integer :: nearestBin, numBins
    real(double), allocatable :: repackedVector(:)
    real(double) :: adjustedBinFraction

    !-------------------------------------------------------------------
    ! Start

    adjustedBinFraction = min(1.0d0, max(0.0d0, binFraction))

    if (present(maskVector)) then
      numBins = count(maskVector)
    else
      numBins = size(vector)
    end if

    if (adjustedBinFraction > 0.5) then
      nearestBin = 1 + numBins - ceiling(numBins * (1.0 - adjustedBinFraction))
    else
      nearestBin = ceiling(numBins * adjustedBinFraction)
    end if
    nearestBin = min(numBins, max(1, nearestBin))

    if (present(maskVector)) then
      allocate(repackedVector(count(maskVector)))
      repackedVector = pack(vector, maskVector)
      valueAtHistoFracDoubleVec = valueAtHistoFractionKernel(repackedVector&
        , nearestBin)

    else
      valueAtHistoFracDoubleVec = valueAtHistoFractionKernel(vector&
        , nearestBin)
    end if

  end function valueAtHistoFracDoubleVec

  !---------------------------------------------------------------------
  real(single) function valueAtHistoFracSingleArr(array, binFraction, maskArray)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: array(:,:), binFraction
    logical,      intent(in), optional :: maskArray(size(array,1)&
                                                   ,size(array,2))


    !-------------------------------------------------------------------
    ! Start

    valueAtHistoFracSingleArr = real(valueAtHistoFracDoubleArr(&
      real(array, double), real(binFraction, double), maskArray), single)

  end function valueAtHistoFracSingleArr

  !---------------------------------------------------------------------
  real(double) function valueAtHistoFracDoubleArr(array, binFraction, maskArray)
  !---------------------------------------------------------------------
  ! This function takes an array of real values, reshapes it to a vector, sorts the vector in increasing order, and returns the value of that element whose bin index (starting at 1) in the sorted vector, when divided by the total number of bins in the vector, is closest to the argument binFraction. Eg to find the median value of an array one could call this function with binFraction set to 0.5.
  !
  ! If maskArray() is supplied the non-masked pixels are ignored in the calculation.

!*** mods from _20030817 not yet tested

    implicit none

    real(double), intent(in)           :: array(:,:), binFraction
    logical,      intent(in), optional :: maskArray(size(array,1)&
                                                   ,size(array,2))

    integer :: xSize, ySize, vSize, nearestBin, numBins
    real(double), allocatable :: vector(:)
    real(double) :: adjustedBinFraction

    !-------------------------------------------------------------------
    ! Start

    xSize = size(array, 1)
    ySize = size(array, 2)
    vSize = xSize * ySize

    if (present(maskArray)) then
      allocate(vector(count(maskArray)))
      vector = pack(array, maskArray)
      numBins = count(maskArray)

    else
      allocate(vector(vSize))
      vector = reshape(array, (/vSize/))
      numBins = vSize
    end if

    adjustedBinFraction = min(1.0d0, max(0.0d0, binFraction))
    if (adjustedBinFraction > 0.5) then
      nearestBin = 1 + numBins - ceiling(numBins * (1.0 - adjustedBinFraction))
    else
      nearestBin = ceiling(numBins * adjustedBinFraction)
    end if
    nearestBin = min(numBins, max(1, nearestBin))

    valueAtHistoFracDoubleArr = valueAtHistoFractionKernel(vector, nearestBin)

  end function valueAtHistoFracDoubleArr

  !---------------------------------------------------------------------
  real(double) function valueAtHistoFractionKernel(vector, nearestBin)
  !---------------------------------------------------------------------
  ! Adapted from the Numerical Recipes function select().

    implicit none

    real(double), intent(in) :: vector(:)
    integer, intent(in) :: nearestBin

    integer :: vSize, i, ir, j, l, mid
    real(double) :: tempVal
    real(double), allocatable :: tempVec(:)

    !-------------------------------------------------------------------
    ! Start

    vSize = size(vector)

    if (vSize <= 0) then
      valueAtHistoFractionKernel = 0.0d0
  return
    end if

    allocate(tempVec(vSize))
    tempVec = vector

    l = 1
    ir = vSize

    do
      if (ir - l <= 1) then
        if (ir - l == 1) then
          if (tempVec(l) > tempVec(ir)) call swap(tempVec(l), tempVec(ir))
          ! swap() is in ssc_misc.f90
        end if
        valueAtHistoFractionKernel = tempVec(nearestBin)
    exit

      else
        mid = (l + ir) / 2
        call swap(tempVec(mid), tempVec(l + 1))

        if (tempVec(l)   > tempVec(ir))  call swap(tempVec(l),   tempVec(ir))
        if (tempVec(l+1) > tempVec(ir))  call swap(tempVec(l+1), tempVec(ir))
        if (tempVec(l)   > tempVec(l+1)) call swap(tempVec(l),   tempVec(l+1))
        ! swap() is in ssc_misc.f90

        i = l + 1
        j = ir
        tempVal = tempVec(l + 1)
        do
          do
            i = i + 1
          if (tempVec(i) >= tempVal) exit
          end do

          do
            j = j - 1
          if (tempVec(j) <= tempVal) exit
          end do

        if (j < i) exit

          call swap(tempVec(i), tempVec(j))
          ! swap() is in ssc_misc.f90
        end do

        tempVec(l + 1) = tempVec(j)
        tempVec(j) = tempVal
        if (j >= nearestBin) ir = j - 1
        if (j <= nearestBin) l = i
      end if
    end do

  end function valueAtHistoFractionKernel

  !---------------------------------------------------------------------
  real(single) function angleBetweenSingle(vectorA, vectorB)
  !---------------------------------------------------------------------

!****** check that these give a more precise value than the cosine rule!! It's by no means obvious.


  ! As is well known, the cosine of the angle between two vectors a_ and
  ! b_ is just (a_.b_)/|a_|/|b_|. Direct inversion of this is however not
  ! very precise when a_ and b_ are nearly parallel. Better is to also
  ! find the sine, then use atan2 to get the angle. This does incur a
  ! penalty however, since it involves calculation of 2 square roots.
  !
  ! The sine can be shown to equal
  !
  !	              |   b_       a_ . b_      a_   |
  !	sine(angle) = | ------ - ---------- * ------ |.
  !	              |  |b_|     |a_||b_|     |a_|  |
  !
  ! However it is more convenient to calculate |a_|*|b_|*cos and
  ! |a_|*|b_|*sin instead of cos and sin.

    implicit none

    real(single), intent(in) :: vectorA(:), vectorB(size(vectorA))

    real(single) :: normA, abCos, abSinVector(size(vectorA)), abSin

    !-------------------------------------------------------------------
    ! Start

    normA = sqrt(sum(vectorA * vectorA))
    abCos = sum(vectorA * vectorB)
    abSinVector = vectorB * normA - abCos * vectorA / normA
    abSin = sqrt(sum(abSinVector * abSinVector))

    angleBetweenSingle = atan2(abSin, abCos)

  end function angleBetweenSingle

  !---------------------------------------------------------------------
  real(double) function angleBetweenDouble(vectorA, vectorB)
  !---------------------------------------------------------------------
  ! As is well known, the cosine of the angle between two vectors a_ and
  ! b_ is just (a_.b_)/|a_|/|b_|. Direct inversion of this is however not
  ! very precise when a_ and b_ are nearly parallel. Better is to also
  ! find the sine, then use atan2 to get the angle. This does incur a
  ! penalty however, since it involves calculation of 2 square roots.
  !
  ! The sine can be shown to equal
  !
  !	              |   b_       a_ . b_      a_   |
  !	sine(angle) = | ------ - ---------- * ------ |.
  !	              |  |b_|     |a_||b_|     |a_|  |
  !
  ! However it is more convenient to calculate |a_|*|b_|*cos and
  ! |a_|*|b_|*sin instead of cos and sin.

    implicit none

    real(double), intent(in) :: vectorA(:), vectorB(size(vectorA))

    real(double) :: normA, abCos, abSinVector(size(vectorA)), abSin

    !-------------------------------------------------------------------
    ! Start

    normA = sqrt(sum(vectorA * vectorA))
    abCos = sum(vectorA * vectorB)
    abSinVector = vectorB * normA - abCos * vectorA / normA
    abSin = sqrt(sum(abSinVector * abSinVector))

    angleBetweenDouble = atan2(abSin, abCos)

  end function angleBetweenDouble


  !---------------------------------------------------------------------
  function sumInt32Int81d(ary) result(retInt32)
  !---------------------------------------------------------------------
  ! sum(variableInt8) => int32

    implicit none

    integer(int8), intent(in)  :: ary(:)
    integer(int32) :: retInt32

    retInt32 = sumInt32Int321d(int(ary, int32))

  end function sumInt32Int81d

  !---------------------------------------------------------------------
  function sumInt32Int161d(ary) result(retInt32)
  !---------------------------------------------------------------------
  ! sum(variableInt16) => int32

    implicit none

    integer(int16), intent(in)  :: ary(:)
    integer(int32) :: retInt32

    retInt32 = sumInt32Int321d(int(ary, int32))

  end function sumInt32Int161d

  !---------------------------------------------------------------------
  function sumInt32Int321d(ary) result(retInt32)
  !---------------------------------------------------------------------
  ! sum(variableInt32) => int32 (or INTEGER_NULL if it is too large/small).

    implicit none

    integer(int32), intent(in)  :: ary(:)
    integer(int32) :: retInt32

    call sumInt32DblInt321d(ary, retInt32)

  end function sumInt32Int321d

  !---------------------------------------------------------------------
  function sumInt32Int82d(ary) result(retInt32)
  !---------------------------------------------------------------------
  ! sum(variableInt8) => int32

    implicit none

    integer(int8), intent(in)  :: ary(:,:)
    integer(int32) :: retInt32

    retInt32 = sumInt32Int322d(int(ary, int32))

  end function sumInt32Int82d

  !---------------------------------------------------------------------
  function sumInt32Int162d(ary) result(retInt32)
  !---------------------------------------------------------------------
  ! sum(variableInt16) => int32

    implicit none

    integer(int16), intent(in)  :: ary(:,:)
    integer(int32) :: retInt32

    retInt32 = sumInt32Int322d(int(ary, int32))

  end function sumInt32Int162d

  !---------------------------------------------------------------------
  function sumInt32Int322d(ary) result(retInt32)
  !---------------------------------------------------------------------
  ! sum(variableInt32) => int32 (or INTEGER_NULL if it is too large/small).

    implicit none

    integer(int32), intent(in)  :: ary(:,:)
    integer(int32) :: retInt32

    call sumInt32DblInt322d(ary, retInt32)

  end function sumInt32Int322d

  !---------------------------------------------------------------------
  function sumInt32Int83d(ary) result(retInt32)
  !---------------------------------------------------------------------
  ! sum(variableInt8) => int32

    implicit none

    integer(int8), intent(in)  :: ary(:,:,:)
    integer(int32) :: retInt32

    retInt32 = sumInt32Int323d(int(ary, int32))

  end function sumInt32Int83d

  !---------------------------------------------------------------------
  function sumInt32Int163d(ary) result(retInt32)
  !---------------------------------------------------------------------
  ! sum(variableInt16) => int32

    implicit none

    integer(int16), intent(in)  :: ary(:,:,:)
    integer(int32) :: retInt32

    retInt32 = sumInt32Int323d(int(ary, int32))

  end function sumInt32Int163d

  !---------------------------------------------------------------------
  function sumInt32Int323d(ary) result(retInt32)
  !---------------------------------------------------------------------
  ! sum(variableInt32) => int32 (or INTEGER_NULL if it is too large/small).

    implicit none

    integer(int32), intent(in)  :: ary(:,:,:)
    integer(int32) :: retInt32

    call sumInt32DblInt323d(ary, retInt32)

  end function sumInt32Int323d

  !---------------------------------------------------------------------
  function sumInt32Int84d(ary) result(retInt32)
  !---------------------------------------------------------------------
  ! sum(variableInt8) => int32

    implicit none

    integer(int8), intent(in)  :: ary(:,:,:,:)
    integer(int32) :: retInt32

    retInt32 = sumInt32Int324d(int(ary, int32))

  end function sumInt32Int84d

  !---------------------------------------------------------------------
  function sumInt32Int164d(ary) result(retInt32)
  !---------------------------------------------------------------------
  ! sum(variableInt16) => int32

    implicit none

    integer(int16), intent(in)  :: ary(:,:,:,:)
    integer(int32) :: retInt32

    retInt32 = sumInt32Int324d(int(ary, int32))

  end function sumInt32Int164d

  !---------------------------------------------------------------------
  function sumInt32Int324d(ary) result(retInt32)
  !---------------------------------------------------------------------
  ! sum(variableInt32) => int32 (or INTEGER_NULL if it is too large/small).

    implicit none

    integer(int32), intent(in)  :: ary(:,:,:,:)
    integer(int32) :: retInt32

    call sumInt32DblInt324d(ary, retInt32)

  end function sumInt32Int324d


  !---------------------------------------------------------------------
  subroutine sumInt32DblInt81d(ary, sumInInt32, sumInDbl)
  !---------------------------------------------------------------------
  ! sum(variableInt8) => int32 & double

    implicit none

    integer(int8),  intent(in) :: ary(:)
    integer(int32), intent(out), optional :: sumInInt32
    real(double),   intent(out), optional :: sumInDbl

    call sumInt32DblInt321d(int(ary, int32), sumInInt32, sumInDbl)

  end subroutine sumInt32DblInt81d

  !---------------------------------------------------------------------
  subroutine sumInt32DblInt161d(ary, sumInInt32, sumInDbl)
  !---------------------------------------------------------------------
  ! sum(variableInt16) => int32 & double

    implicit none

    integer(int16), intent(in) :: ary(:)
    integer(int32), intent(out), optional :: sumInInt32
    real(double),   intent(out), optional :: sumInDbl

    call sumInt32DblInt321d(int(ary, int32), sumInInt32, sumInDbl)

  end subroutine sumInt32DblInt161d

  !---------------------------------------------------------------------
  subroutine sumInt32DblInt321d(ary, sumInInt32, sumInDbl)
  !---------------------------------------------------------------------
  ! sum(variableInt32) => int32 & double (or INTEGER_NULL if it is too large/small).

    implicit none

    integer(int32), intent(in) :: ary(:)
    integer(int32), intent(out), optional :: sumInInt32
    real(double),   intent(out), optional :: sumInDbl
    integer(int32) :: sumValInt32
    real(double)   :: sumValDouble

    if (size(ary) == 0) then
      sumValInt32  = INTEGER_NULL
      sumValDouble = REAL_NULL
    else
      sumValDouble = sum(real(ary, double))
      if ( abs(sumValDouble) > huge(ary(1)) ) then
        sumValInt32 = INTEGER_NULL
      else
        sumValInt32 = sum(ary)
        ! sumValInt32 = nint(sumValDouble, int32)	! Faster, but possibly less accurate
      end if
    end if

    if (present(sumInInt32)) sumInInt32 = sumValInt32 
    if (present(sumInDbl))   sumInDbl   = sumValDouble

  end subroutine sumInt32DblInt321d

  !---------------------------------------------------------------------
  subroutine sumInt32DblInt82d(ary, sumInInt32, sumInDbl)
  !---------------------------------------------------------------------
  ! sum(variableInt8) => int32 & double

    implicit none

    integer(int8),  intent(in) :: ary(:,:)
    integer(int32), intent(out), optional :: sumInInt32
    real(double),   intent(out), optional :: sumInDbl

    call sumInt32DblInt322d(int(ary, int32), sumInInt32, sumInDbl)

  end subroutine sumInt32DblInt82d

  !---------------------------------------------------------------------
  subroutine sumInt32DblInt162d(ary, sumInInt32, sumInDbl)
  !---------------------------------------------------------------------
  ! sum(variableInt16) => int32 & double

    implicit none

    integer(int16), intent(in) :: ary(:,:)
    integer(int32), intent(out), optional :: sumInInt32
    real(double),   intent(out), optional :: sumInDbl

    call sumInt32DblInt322d(int(ary, int32), sumInInt32, sumInDbl)

  end subroutine sumInt32DblInt162d

  !---------------------------------------------------------------------
  subroutine sumInt32DblInt322d(ary, sumInInt32, sumInDbl)
  !---------------------------------------------------------------------
  ! sum(variableInt32) => int32 & double (or INTEGER_NULL if it is too large/small).

    implicit none

    integer(int32), intent(in) :: ary(:,:)
    integer(int32), intent(out), optional :: sumInInt32
    real(double),   intent(out), optional :: sumInDbl
    integer(int32) :: sumValInt32
    real(double)   :: sumValDouble

    if (size(ary) == 0) then
      sumValInt32  = INTEGER_NULL
      sumValDouble = REAL_NULL
    else
      sumValDouble = sum(real(ary, double))
      if ( abs(sumValDouble) > huge(ary(1,1)) ) then
        sumValInt32 = INTEGER_NULL
      else
        sumValInt32 = sum(ary)
        ! sumValInt32 = nint(sumValDouble, int32)	! Faster, but possibly less accurate
      end if
    end if

    if (present(sumInInt32)) sumInInt32 = sumValInt32 
    if (present(sumInDbl))   sumInDbl   = sumValDouble

  end subroutine sumInt32DblInt322d

  !---------------------------------------------------------------------
  subroutine sumInt32DblInt83d(ary, sumInInt32, sumInDbl)
  !---------------------------------------------------------------------
  ! sum(variableInt8) => int32 & double

    implicit none

    integer(int8),  intent(in) :: ary(:,:,:)
    integer(int32), intent(out), optional :: sumInInt32
    real(double),   intent(out), optional :: sumInDbl

    call sumInt32DblInt323d(int(ary, int32), sumInInt32, sumInDbl)

  end subroutine sumInt32DblInt83d

  !---------------------------------------------------------------------
  subroutine sumInt32DblInt163d(ary, sumInInt32, sumInDbl)
  !---------------------------------------------------------------------
  ! sum(variableInt16) => int32 & double

    implicit none

    integer(int16), intent(in) :: ary(:,:,:)
    integer(int32), intent(out), optional :: sumInInt32
    real(double),   intent(out), optional :: sumInDbl

    call sumInt32DblInt323d(int(ary, int32), sumInInt32, sumInDbl)

  end subroutine sumInt32DblInt163d

  !---------------------------------------------------------------------
  subroutine sumInt32DblInt323d(ary, sumInInt32, sumInDbl)
  !---------------------------------------------------------------------
  ! sum(variableInt32) => int32 & double (or INTEGER_NULL if it is too large/small).

    implicit none

    integer(int32), intent(in) :: ary(:,:,:)
    integer(int32), intent(out), optional :: sumInInt32
    real(double),   intent(out), optional :: sumInDbl
    integer(int32) :: sumValInt32
    real(double)   :: sumValDouble

    if (size(ary) == 0) then
      sumValInt32  = INTEGER_NULL
      sumValDouble = REAL_NULL
    else
      sumValDouble = sum(real(ary, double))
      if ( abs(sumValDouble) > huge(ary(1,1,1)) ) then
        sumValInt32 = INTEGER_NULL
      else
        sumValInt32 = sum(ary)
        ! sumValInt32 = nint(sumValDouble, int32)	! Faster, but possibly less accurate
      end if
    end if

    if (present(sumInInt32)) sumInInt32 = sumValInt32 
    if (present(sumInDbl))   sumInDbl   = sumValDouble

  end subroutine sumInt32DblInt323d

  !---------------------------------------------------------------------
  subroutine sumInt32DblInt84d(ary, sumInInt32, sumInDbl)
  !---------------------------------------------------------------------
  ! sum(variableInt8) => int32 & double

    implicit none

    integer(int8),  intent(in) :: ary(:,:,:,:)
    integer(int32), intent(out), optional :: sumInInt32
    real(double),   intent(out), optional :: sumInDbl

    call sumInt32DblInt324d(int(ary, int32), sumInInt32, sumInDbl)

  end subroutine sumInt32DblInt84d

  !---------------------------------------------------------------------
  subroutine sumInt32DblInt164d(ary, sumInInt32, sumInDbl)
  !---------------------------------------------------------------------
  ! sum(variableInt16) => int32 & double

    implicit none

    integer(int16), intent(in) :: ary(:,:,:,:)
    integer(int32), intent(out), optional :: sumInInt32
    real(double),   intent(out), optional :: sumInDbl

    call sumInt32DblInt324d(int(ary, int32), sumInInt32, sumInDbl)

  end subroutine sumInt32DblInt164d

  !---------------------------------------------------------------------
  subroutine sumInt32DblInt324d(ary, sumInInt32, sumInDbl)
  !---------------------------------------------------------------------
  ! sum(variableInt32) => int32 & double (or INTEGER_NULL if it is too large/small).

    implicit none

    integer(int32), intent(in) :: ary(:,:,:,:)
    integer(int32), intent(out), optional :: sumInInt32
    real(double),   intent(out), optional :: sumInDbl
    integer(int32) :: sumValInt32
    real(double)   :: sumValDouble

    if (size(ary) == 0) then
      sumValInt32  = INTEGER_NULL
      sumValDouble = REAL_NULL
    else
      sumValDouble = sum(real(ary, double))
      if ( abs(sumValDouble) > huge(ary(1,1,1,1)) ) then
        sumValInt32 = INTEGER_NULL
      else
        sumValInt32 = sum(ary)
        ! sumValInt32 = nint(sumValDouble, int32)	! Faster, but possibly less accurate
      end if
    end if

    if (present(sumInInt32)) sumInInt32 = sumValInt32 
    if (present(sumInDbl))   sumInDbl   = sumValDouble

  end subroutine sumInt32DblInt324d


  !---------------------------------------------------------------------
  function getUboundFromEdgesInfo(aryEdgesInfo) result(outUbound)
  !---------------------------------------------------------------------
  ! Returns ubound() for a given aryEdgesInfoT variable.

    implicit none

    type(AryEdgesInfoT), intent(in) :: aryEdgesInfo
    integer :: outUbound(aryEdgesInfo%aryDimension)	! return

    !-------------------------------------------------------------------
    ! Start:

    outUbound(:) = aryEdgesInfo%lboundIndex + aryEdgesInfo%arySize - 1	! Array operation.

  end function getUboundFromEdgesInfo

  !---------------------------------------------------------------------
  function getUboundFromScalars(inLbound, arySize) result(outUbound)
  !---------------------------------------------------------------------
  ! Returns ubound() for a given pair of lbound and size.

    implicit none

    integer, intent(in) :: inLbound(:), arySize(size(inLbound))
    integer :: outUbound(size(inLbound))	! return

    !-------------------------------------------------------------------
    ! Start:

    outUbound(:) = inLbound + arySize - 1	! Array operation.

  end function getUboundFromScalars


  !---------------------------------------------------------------------
  subroutine printAryEdgesInfo(aryEdgesInfo, varName)
  !---------------------------------------------------------------------
  ! print the debug info for AryEdgesInfoT

    implicit none

    type(AryEdgesInfoT), intent(in) :: aryEdgesInfo
    character(*), intent(in), optional :: varName

    !-------------------------------------------------------------------
    ! Start:

    if (present(varName)) then
      print *, '*** AryEdgesInfoT: ', trim(varName)
    else
      print *, '*** AryEdgesInfoT: '
    end if

    print *, ' aryDimension = ', aryEdgesInfo%aryDimension
    if (allocated(aryEdgesInfo%arySize)) then
      print *, ' arySize     = ', aryEdgesInfo%arySize
    else
      print *, ' arySize     : NOT allocated.'
    end if
    if (allocated(aryEdgesInfo%lboundIndex)) then
      print *, ' lboundIndex = ', aryEdgesInfo%lboundIndex
    else
      print *, ' lboundIndex : NOT allocated.'
    end if
    if (allocated(aryEdgesInfo%lEdge)) then
      print *, ' lEdge = ', aryEdgesInfo%lEdge
    else
      print *, ' lEdge : NOT allocated.'
    end if
    if (allocated(aryEdgesInfo%uEdge)) then
      print *, ' uEdge = ', aryEdgesInfo%uEdge
    else
      print *, ' uEdge : NOT allocated.'
    end if

  end subroutine printAryEdgesInfo


  !---------------------------------------------------------------------
  function getAryEdgesInfoDbl1d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 1d array of Double.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    real(double), intent(in)  :: ary(:)
    real(double), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! return:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    aryEdgesInfo%arySize = ubound(ary)
    call setAryEdgesInfoCommon(rankArin, aryEdgesInfo, lEdge, uEdge, lboundIndex, style)

  end function getAryEdgesInfoDbl1d


  !---------------------------------------------------------------------
  function getAryEdgesInfoDbl2d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 2d array of Double.
  ! lboundIndex is lbound(ary) (in the caller); or 1 if unspecified.
  ! If style=='wcs', then it is allowed neither lEdge nor uEdge are given
  ! (nb., if you for some reason choose to give one, please give the both);
  ! in that case lEdge is lboundIndex-0.5 and uEdge is accordingly defined.
  ! Otherwise lEdge and uEdge MUST be given.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    real(double), intent(in)  :: ary(:,:)
    real(double), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! Start:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    ! aryEdgesInfo%arySize(1) = size(ary, 1)
    ! aryEdgesInfo%arySize(2) = size(ary, 2)
    aryEdgesInfo%arySize = ubound(ary)
    call setAryEdgesInfoCommon(rankArin, aryEdgesInfo, lEdge, uEdge, lboundIndex, style)

  end function getAryEdgesInfoDbl2d


  !---------------------------------------------------------------------
  function getAryEdgesInfoDbl3d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 3d array of Double.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 3	! = size(lbound(ary))

    real(double), intent(in)  :: ary(:,:,:)
    real(double), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! Start:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    aryEdgesInfo%arySize = ubound(ary)
    call setAryEdgesInfoCommon(rankArin, aryEdgesInfo, lEdge, uEdge, lboundIndex, style)

  end function getAryEdgesInfoDbl3d


  !---------------------------------------------------------------------
  function getAryEdgesInfoSgl1d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 1d array of Single.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    real(single), intent(in)  :: ary(:)
    real(single), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! return:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    aryEdgesInfo%arySize = ubound(ary)
    if (present(lEdge) .and. present(uEdge)) then
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , real(lEdge, double), real(uEdge, double), lboundIndex, style)
    else
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , lboundIndex=lboundIndex, style=style)
    end if

  end function getAryEdgesInfoSgl1d


  !---------------------------------------------------------------------
  function getAryEdgesInfoSgl2d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 2d array of Single.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    real(single), intent(in)  :: ary(:,:)
    real(single), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! Start:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    ! aryEdgesInfo%arySize(1) = size(ary, 1)
    ! aryEdgesInfo%arySize(2) = size(ary, 2)
    aryEdgesInfo%arySize = ubound(ary)
    if (present(lEdge) .and. present(uEdge)) then
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , real(lEdge, double), real(uEdge, double), lboundIndex, style)
    else
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , lboundIndex=lboundIndex, style=style)
    end if

  end function getAryEdgesInfoSgl2d


  !---------------------------------------------------------------------
  function getAryEdgesInfoSgl3d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 3d array of Single.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 3	! = size(lbound(ary))

    real(single), intent(in)  :: ary(:,:,:)
    real(single), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! Start:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    aryEdgesInfo%arySize = ubound(ary)
    if (present(lEdge) .and. present(uEdge)) then
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , real(lEdge, double), real(uEdge, double), lboundIndex, style)
    else
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , lboundIndex=lboundIndex, style=style)
    end if

  end function getAryEdgesInfoSgl3d


  !---------------------------------------------------------------------
  function getAryEdgesInfoInt321d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 1d array of Single.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    integer(int32), intent(in)  :: ary(:)
    real(single), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! return:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    aryEdgesInfo%arySize = ubound(ary)
    if (present(lEdge) .and. present(uEdge)) then
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , real(lEdge, double), real(uEdge, double), lboundIndex, style)
    else
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , lboundIndex=lboundIndex, style=style)
    end if

  end function getAryEdgesInfoInt321d


  !---------------------------------------------------------------------
  function getAryEdgesInfoInt322d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 2d array of Single.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    integer(int32), intent(in)  :: ary(:,:)
    real(single), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! Start:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    ! aryEdgesInfo%arySize(1) = size(ary, 1)
    ! aryEdgesInfo%arySize(2) = size(ary, 2)
    aryEdgesInfo%arySize = ubound(ary)
    if (present(lEdge) .and. present(uEdge)) then
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , real(lEdge, double), real(uEdge, double), lboundIndex, style)
    else
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , lboundIndex=lboundIndex, style=style)
    end if

  end function getAryEdgesInfoInt322d


  !---------------------------------------------------------------------
  function getAryEdgesInfoInt323d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 3d array of Single.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 3	! = size(lbound(ary))

    integer(int32), intent(in)  :: ary(:,:,:)
    real(single), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! Start:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    aryEdgesInfo%arySize = ubound(ary)
    if (present(lEdge) .and. present(uEdge)) then
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , real(lEdge, double), real(uEdge, double), lboundIndex, style)
    else
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , lboundIndex=lboundIndex, style=style)
    end if

  end function getAryEdgesInfoInt323d


  !---------------------------------------------------------------------
  function getAryEdgesInfoInt161d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 1d array of Single.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    integer(int16), intent(in)  :: ary(:)
    real(single), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! return:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    aryEdgesInfo%arySize = ubound(ary)
    if (present(lEdge) .and. present(uEdge)) then
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , real(lEdge, double), real(uEdge, double), lboundIndex, style)
    else
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , lboundIndex=lboundIndex, style=style)
    end if

  end function getAryEdgesInfoInt161d


  !---------------------------------------------------------------------
  function getAryEdgesInfoInt162d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 2d array of Single.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    integer(int16), intent(in)  :: ary(:,:)
    real(single), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! Start:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    ! aryEdgesInfo%arySize(1) = size(ary, 1)
    ! aryEdgesInfo%arySize(2) = size(ary, 2)
    aryEdgesInfo%arySize = ubound(ary)
    if (present(lEdge) .and. present(uEdge)) then
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , real(lEdge, double), real(uEdge, double), lboundIndex, style)
    else
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , lboundIndex=lboundIndex, style=style)
    end if

  end function getAryEdgesInfoInt162d


  !---------------------------------------------------------------------
  function getAryEdgesInfoInt163d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 3d array of Single.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 3	! = size(lbound(ary))

    integer(int16), intent(in)  :: ary(:,:,:)
    real(single), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! Start:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    aryEdgesInfo%arySize = ubound(ary)
    if (present(lEdge) .and. present(uEdge)) then
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , real(lEdge, double), real(uEdge, double), lboundIndex, style)
    else
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , lboundIndex=lboundIndex, style=style)
    end if

  end function getAryEdgesInfoInt163d


  !---------------------------------------------------------------------
  function getAryEdgesInfoInt81d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 1d array of Single.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    integer(int8), intent(in)  :: ary(:)
    real(single), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! return:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    aryEdgesInfo%arySize = ubound(ary)
    if (present(lEdge) .and. present(uEdge)) then
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , real(lEdge, double), real(uEdge, double), lboundIndex, style)
    else
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , lboundIndex=lboundIndex, style=style)
    end if

  end function getAryEdgesInfoInt81d


  !---------------------------------------------------------------------
  function getAryEdgesInfoInt82d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 2d array of Single.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    integer(int8), intent(in)  :: ary(:,:)
    real(single), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! Start:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    ! aryEdgesInfo%arySize(1) = size(ary, 1)
    ! aryEdgesInfo%arySize(2) = size(ary, 2)
    aryEdgesInfo%arySize = ubound(ary)
    if (present(lEdge) .and. present(uEdge)) then
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , real(lEdge, double), real(uEdge, double), lboundIndex, style)
    else
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , lboundIndex=lboundIndex, style=style)
    end if

  end function getAryEdgesInfoInt82d


  !---------------------------------------------------------------------
  function getAryEdgesInfoInt83d(ary, lEdge, uEdge, lboundIndex, style) result(aryEdgesInfo)
  !---------------------------------------------------------------------
  ! Return aryEdgesInfo for 3d array of Single.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, parameter :: rankArin = 3	! = size(lbound(ary))

    integer(int8), intent(in)  :: ary(:,:,:)
    real(single), intent(in), optional :: lEdge(rankArin), uEdge(rankArin)
    integer,      intent(in), optional :: lboundIndex(rankArin)
    character(*), intent(in), optional :: style
    type(AryEdgesInfoT) :: aryEdgesInfo	! return

    !-------------------------------------------------------------------
    ! Start:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%arySize(rankArin))
    aryEdgesInfo%arySize = ubound(ary)
    if (present(lEdge) .and. present(uEdge)) then
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , real(lEdge, double), real(uEdge, double), lboundIndex, style)
    else
      call setAryEdgesInfoCommon(rankArin, aryEdgesInfo &
         , lboundIndex=lboundIndex, style=style)
    end if

  end function getAryEdgesInfoInt83d


  !---------------------------------------------------------------------
  subroutine setAryEdgesInfoCommon(rankArin, aryEdgesInfo, lEdge, uEdge, lboundIndex, style)
  !---------------------------------------------------------------------
  ! Modify the input aryEdgesInfo (type AryEdgesInfoT) for setting up.
  ! aryEdgesInfo%arySize(:) is assumed to be set.
  ! See the comments in getAryEdgesInfoDbl2d for detail.

    implicit none

    integer, intent(in) :: rankArin	! eg., 2 for 2-dimension
    type(AryEdgesInfoT), intent(inout) :: aryEdgesInfo
    real(double), intent(in), optional :: lEdge(:), uEdge(:)
    integer,      intent(in), optional :: lboundIndex(:)
    character(*), intent(in), optional :: style

    character(64) :: localStyle

    !-------------------------------------------------------------------
    ! Start:

    aryEdgesInfo%aryDimension = rankArin
    allocate(aryEdgesInfo%lboundIndex(rankArin) &
           , aryEdgesInfo%lEdge(rankArin),   aryEdgesInfo%uEdge(rankArin))

    if (present(lboundIndex)) then
      aryEdgesInfo%lboundIndex = lboundIndex
    else
      aryEdgesInfo%lboundIndex = 1
    end if

    if (present(style)) then
      localStyle = style
    else
      localStyle = ''
    end if

    select case(trim(localStyle))
    case('wcs')
      if (present(lEdge) .and. present(uEdge)) then
        aryEdgesInfo%lEdge = lEdge
        aryEdgesInfo%uEdge = uEdge
      else
        aryEdgesInfo%lEdge = aryEdgesInfo%lboundIndex - 0.5
        aryEdgesInfo%uEdge = aryEdgesInfo%lboundIndex + aryEdgesInfo%arySize - 0.5
      end if

    case default
      if (present(lEdge)) then
        aryEdgesInfo%lEdge = lEdge
      else
        call error('lEdgeUndefined', 'ssclib:array_utils:setAryEdgesInfoCommon(): &
           &style="wcs" should be given in getAryEdgesInfo().  Contact the code developer.')
      end if
      if (present(uEdge)) then
        aryEdgesInfo%uEdge = uEdge
      else
        call error('uEdgeUndefined', 'ssclib:array_utils:setAryEdgesInfoCommon(): &
           &style="wcs" should be given in getAryEdgesInfo().  Contact the code developer.')
      end if
    end select

  end subroutine setAryEdgesInfoCommon


  !---------------------------------------------------------------------
  subroutine calcAryStatInfoMaskDouble1d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)
  !---------------------------------------------------------------------
  ! Return aryStatInfo _and_ the final mask used in the calculation.
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    real(double), intent(in)  :: arin(:)
    type(aryStatInfoDoubleT), intent(out) :: retInfo
    logical, intent(out) :: arMaskOut(:)	! Must be predefined.
    logical, intent(in), optional :: arMaskIn(:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    type(aryStatInfoFlagT) :: usedInfo
    logical, dimension(:), allocatable :: arMaskArea, arMaskVal
    real(double), dimension(:), allocatable :: arWork
    integer :: ubound_arin(rankArin)
    integer :: i !, j

    !-------------------------------------------------------------------
    ! Start:

    ubound_arin(1) = ubound(arin,1)

    allocate(retInfo%minindices(rankArin), retInfo%maxindices(rankArin))
    allocate(arMaskArea( ubound_arin(1)))
    allocate(arMaskVal(  ubound_arin(1)))
    allocate(arWork(     ubound_arin(1)))
    arMaskArea = .true.	! Default is ALL true.
    arMaskVal  = .true.	! Default is ALL true.

    !! Sort out usedInfo
    if (present(flagInfo)) then
      usedInfo = flagInfo
      if (.not. present(minAreaIndices)) usedInfo%isAreaLowerUsed = .false.
      if (.not. present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .false.
      if (.not. present(valLower))       usedInfo%isValLowerUsed  = .false.
      if (.not. present(valUpper))       usedInfo%isValUpperUsed  = .false.
      if (.not. present(arMaskIn))       usedInfo%isMaskUsed      = .false.
    else
      if (present(minAreaIndices)) usedInfo%isAreaLowerUsed = .true.
      if (present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .true.
      if (present(valLower))       usedInfo%isValLowerUsed  = .true.
      if (present(valUpper))       usedInfo%isValUpperUsed  = .true.
      if (present(arMaskIn))       usedInfo%isMaskUsed      = .true.
    end if

    !! Sanity check for minAreaIndices and maxAreaIndices.
    if (usedInfo%isAreaLowerUsed .and. usedInfo%isAreaUpperUsed) then
      call chkAreaIndices(minAreaIndices, maxAreaIndices, retInfo%flag%status)
      retInfo%flag%status = retInfo%flag%status * 4	! The 3rd and 4th bit.
    end if

    !! AreaIndices --- Define (min|max)areaindices if _either_ of them is specified.
    if (usedInfo%isAreaLowerUsed .or. usedInfo%isAreaUpperUsed) then
      allocate(retInfo%minareaindices(rankArin), retInfo%maxareaindices(rankArin))
      if (usedInfo%isAreaLowerUsed) then
        retInfo%flag%isAreaLowerUsed = .true.
        retInfo%minareaindices = minAreaIndices
      else
        retInfo%minareaindices(:) = 1
      end if
      if (usedInfo%isAreaUpperUsed) then
        retInfo%flag%isAreaUpperUsed = .true.
        retInfo%maxareaindices = maxAreaIndices
      else
        retInfo%maxareaindices = ubound(arin)
      end if

      do i=1,ubound_arin(1)
          if (     (i < retInfo%minareaindices(1)) &
              .or. (i > retInfo%maxareaindices(1))) then
            arMaskArea(i) = .false.
          end if
      end do
    else
      ! Default
    end if

    !! val(Lower|Upper)
    if (usedInfo%isValLowerUsed) then
      retInfo%flag%isValLowerUsed = .true.
      retInfo%vallower = valLower
    else
      ! Do nothing (Default)
    end if
    if (usedInfo%isValUpperUsed) then
      retInfo%flag%isValUpperUsed = .true.
      retInfo%valupper = valUpper
    else
      ! Do nothing (Default)
    end if

    if (retInfo%flag%isValLowerUsed .and. retInfo%flag%isValUpperUsed) then
      where (arin < valLower .or. arin > valUpper) arMaskVal = .false.
    else if (retInfo%flag%isValLowerUsed) then
      where (arin < valLower) arMaskVal = .false.
    else if (retInfo%flag%isValUpperUsed) then
      where (arin > valUpper) arMaskVal = .false.
    end if

    arMaskOut = arMaskArea .and. arMaskVal

    if (usedInfo%isMaskUsed) then
      retInfo%flag%isMaskUsed = .true.
      arMaskOut = arMaskOut .and. arMaskIn
    end if

    arWork = 0.0
    where (arMaskOut) arWork = arin

    retInfo%totalentry = ubound_arin(1)
    retInfo%validentry = count(arMaskOut)

    retInfo%minindices = minloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    retInfo%maxindices = maxloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    ! retInfo%minval = arin(retInfo%minindices(1))
    ! retInfo%maxval = arin(retInfo%maxindices(1))

    retInfo%totalsum = sum(arWork)
    retInfo%realsum  = retInfo%totalsum

    if (retInfo%validentry > 0) then
      retInfo%minval = arin(retInfo%minindices(1))
      retInfo%maxval = arin(retInfo%maxindices(1))
      retInfo%mean = retInfo%realsum / retInfo%validentry
      where (arMaskOut) arWork = (arin - retInfo%mean)**2	! arWork is modified.
      ! nb., arWork == 0.0 elsewhere
      ! Now arWork is an array of squared values.
      retInfo%sigma = sqrt(sum(arWork) / retInfo%validentry)
    else
      ! NOTE: %minindices==(0), hence those indices can not be used.
      retInfo%minval = minval(arin, arMaskOut)	! eg.  3.4028235E+38
      retInfo%maxval = maxval(arin, arMaskOut)	! eg. -3.4028235E+38
      retInfo%flag%status = retInfo%flag%status - 1
      call message(AppMsg, Noisy, 'ssclib/array_utils: No valid entry in the input array.')
    end if

    ! if (retInfo%validentry > 0) then
    !   retInfo%mean = retInfo%totalsum / retInfo%validentry
    !   where (arMaskOut) arWork = (arin - retInfo%mean)**2
    !   ! nb., arWork == 0.0 elsewhere
    !   ! Now arWork is an array of squared values.
    !   retInfo%sigma = sqrt(sum(arWork) / retInfo%validentry)
    ! else
    !   retInfo%flag%status = retInfo%flag%status - 1
    !   call message(AppMsg, Noisy, 'ssclib: No valid entry in the input array.')
    ! end if

    deallocate(arMaskArea, arMaskVal)

  end subroutine calcAryStatInfoMaskDouble1d


  !---------------------------------------------------------------------
  subroutine calcAryStatInfoMaskDouble2d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)
  !---------------------------------------------------------------------
  ! Return aryStatInfo _and_ the final mask used in the calculation.
  ! NOTE:
  !  * flagInfo is, if given in call, used for its logical elements,
  !    that is, if an argument, such as valLower, is present
  !    but .false. in flagInfo, then that parameter is ignored
  !    in the calculation.
  !  * For (min|max)AreaIndices, the boundary pixels are valid (.true.),
  !   namely used in the calculation
  !  * For val(Lower|Upper), the boundary pixels are also valid (.true.).
  !  * When val(Lower|Upper) is not given here, the %val(lower|upper)
  !   elements in the returned aryStatInfoDoubleT (retInfo) is undefined.
  !   Accordingly if you try to refer to it when undefined, it may cause
  !   a trouble or Fortran error.  Hence if you are not sure, make sure
  !   to check retInfo%flag%isVal(Lower|Upper)Used beforehand.
  !  * The ranks of arin, arMaskOut and arMaskIn have to be identical.
  !  * (The rank of) arMaskOut has to be defined in the caller side.
  !  * retInfo%flag%status gives the returned status.
  !    0 if normal ends.
  !    (negative) 1st bit (-1) is given if validentry is zero (or negative).
  !    (negative) 3rd and 4th bits expresses the number of invalid
  !    axes in minAreaIndices and maxAreaIndices, 
  !    where an index in minAreaIndices is larger than that in maxAreaIndices.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    real(double), intent(in)  :: arin(:,:)
    type(aryStatInfoDoubleT), intent(out) :: retInfo
    logical, intent(out) :: arMaskOut(:,:)	! Must be predefined.
    logical, intent(in), optional :: arMaskIn(:,:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    type(aryStatInfoFlagT) :: usedInfo
    logical, dimension(:,:), allocatable :: arMaskArea, arMaskVal
    real(double), dimension(:,:), allocatable :: arWork
    integer :: ubound_arin(rankArin)
    integer :: i, j

    !-------------------------------------------------------------------
    ! Start:

    ubound_arin(1) = ubound(arin,1)
    ubound_arin(2) = ubound(arin,2)

    allocate(retInfo%minindices(rankArin), retInfo%maxindices(rankArin))
    allocate(arMaskArea( ubound_arin(1), ubound_arin(2)))
    allocate(arMaskVal(  ubound_arin(1), ubound_arin(2)))
    allocate(arWork(     ubound_arin(1), ubound_arin(2)))
    arMaskArea = .true.	! Default is ALL true.
    arMaskVal  = .true.	! Default is ALL true.

    !! Sort out usedInfo
    if (present(flagInfo)) then
      usedInfo = flagInfo
      if (.not. present(minAreaIndices)) usedInfo%isAreaLowerUsed = .false.
      if (.not. present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .false.
      if (.not. present(valLower))       usedInfo%isValLowerUsed  = .false.
      if (.not. present(valUpper))       usedInfo%isValUpperUsed  = .false.
      if (.not. present(arMaskIn))       usedInfo%isMaskUsed      = .false.
    else
      if (present(minAreaIndices)) usedInfo%isAreaLowerUsed = .true.
      if (present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .true.
      if (present(valLower))       usedInfo%isValLowerUsed  = .true.
      if (present(valUpper))       usedInfo%isValUpperUsed  = .true.
      if (present(arMaskIn))       usedInfo%isMaskUsed      = .true.
    end if

    !! Sanity check for minAreaIndices and maxAreaIndices.
    if (usedInfo%isAreaLowerUsed .and. usedInfo%isAreaUpperUsed) then
      call chkAreaIndices(minAreaIndices, maxAreaIndices, retInfo%flag%status)
      retInfo%flag%status = retInfo%flag%status * 4	! The 3rd and 4th bit.
    end if

    !! AreaIndices --- Define (min|max)areaindices if _either_ of them is specified.
    if (usedInfo%isAreaLowerUsed .or. usedInfo%isAreaUpperUsed) then
      allocate(retInfo%minareaindices(rankArin), retInfo%maxareaindices(rankArin))
      if (usedInfo%isAreaLowerUsed) then
        retInfo%flag%isAreaLowerUsed = .true.
        retInfo%minareaindices = minAreaIndices
      else
        retInfo%minareaindices(:) = 1
      end if
      if (usedInfo%isAreaUpperUsed) then
        retInfo%flag%isAreaUpperUsed = .true.
        retInfo%maxareaindices = maxAreaIndices
      else
        retInfo%maxareaindices = ubound(arin)
      end if

      do i=1,ubound_arin(1)
        do j=1,ubound_arin(2)
          if (     (i < retInfo%minareaindices(1)) &
              .or. (i > retInfo%maxareaindices(1)) &
              .or. (j < retInfo%minareaindices(2)) &
              .or. (j > retInfo%maxareaindices(2))) then
            arMaskArea(i,j) = .false.
          end if
        end do
      end do
      ! arMaskArea = (arin(i,j) == arin(i,j) &	! This line is always .true.
      !    .and. (i >= retInfo%minareaindices(1)) &
      !    .and. (i <= retInfo%maxareaindices(1)) &
      !    .and. (j >= retInfo%minareaindices(2)) &
      !    .and. (j <= retInfo%maxareaindices(2))), &
      !    i=1,ubound_arin(1); j=1,ubound_arin(2)
    else
      ! Default
    end if

    !! val(Lower|Upper)
    if (usedInfo%isValLowerUsed) then
      retInfo%flag%isValLowerUsed = .true.
      retInfo%vallower = valLower
    else
      ! Do nothing (Default)
    end if
    if (usedInfo%isValUpperUsed) then
      retInfo%flag%isValUpperUsed = .true.
      retInfo%valupper = valUpper
    else
      ! Do nothing (Default)
    end if

    if (retInfo%flag%isValLowerUsed .and. retInfo%flag%isValUpperUsed) then
      where (arin < valLower .or. arin > valUpper) arMaskVal = .false.
    else if (retInfo%flag%isValLowerUsed) then
      where (arin < valLower) arMaskVal = .false.
    else if (retInfo%flag%isValUpperUsed) then
      where (arin > valUpper) arMaskVal = .false.
    end if

    arMaskOut = arMaskArea .and. arMaskVal

    if (usedInfo%isMaskUsed) then
      retInfo%flag%isMaskUsed = .true.
      arMaskOut = arMaskOut .and. arMaskIn
    end if

    arWork = 0.0
    where (arMaskOut) arWork = arin

    retInfo%totalentry = ubound_arin(1) * ubound_arin(2)
    retInfo%validentry = count(arMaskOut)

    retInfo%minindices = minloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    retInfo%maxindices = maxloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    ! retInfo%minval = arin(retInfo%minindices(1), retInfo%minindices(2))
    ! retInfo%maxval = arin(retInfo%maxindices(1), retInfo%maxindices(2))

    retInfo%totalsum = sum(arWork)
    retInfo%realsum  = retInfo%totalsum

    if (retInfo%validentry > 0) then
      retInfo%minval = arin(retInfo%minindices(1), retInfo%minindices(2))
      retInfo%maxval = arin(retInfo%maxindices(1), retInfo%maxindices(2))
      retInfo%mean = retInfo%realsum / retInfo%validentry
      where (arMaskOut) arWork = (arin - retInfo%mean)**2	! arWork is modified.
      ! nb., arWork == 0.0 elsewhere
      ! Now arWork is an array of squared values.
      retInfo%sigma = sqrt(sum(arWork) / retInfo%validentry)
    else
      ! NOTE: %minindices==(0, 0), hence those indices can not be used.
      retInfo%minval = minval(arin, arMaskOut)	! eg.  3.4028235E+38
      retInfo%maxval = maxval(arin, arMaskOut)	! eg. -3.4028235E+38
      retInfo%flag%status = retInfo%flag%status - 1
      call message(AppMsg, Noisy, 'ssclib/array_utils: No valid entry in the input array.')
    end if

    ! if (retInfo%validentry > 0) then
    !   retInfo%mean = retInfo%totalsum / retInfo%validentry
    !   where (arMaskOut) arWork = (arin - retInfo%mean)**2
    !   ! nb., arWork == 0.0 elsewhere
    !   ! Now arWork is an array of squared values.
    !   retInfo%sigma = sqrt(sum(arWork) / retInfo%validentry)
    ! else
    !   retInfo%flag%status = retInfo%flag%status - 1
    !   call message(AppMsg, Noisy, 'ssclib: No valid entry in the input array.')
    ! end if

    deallocate(arMaskArea, arMaskVal)

  end subroutine calcAryStatInfoMaskDouble2d


  !---------------------------------------------------------------------
  function getAryStatInfoDouble1d(arin, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo) result(retInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    type(aryStatInfoDoubleT) :: retInfo
    real(double), intent(in)  :: arin(:)
    logical, intent(in), optional :: arMaskIn(:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    logical, dimension(:), allocatable :: arMaskOut

    !-------------------------------------------------------------------
    ! Start:

    allocate(arMaskOut(ubound(arin,1)))

    call calcAryStatInfoMaskDouble1d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)

    deallocate(arMaskOut)

  end function getAryStatInfoDouble1d

  !---------------------------------------------------------------------
  function getAryStatInfoDouble2d(arin, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo) result(retInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    type(aryStatInfoDoubleT) :: retInfo
    real(double), intent(in)  :: arin(:,:)
    logical, intent(in), optional :: arMaskIn(:,:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    logical, dimension(:,:), allocatable :: arMaskOut

    !-------------------------------------------------------------------
    ! Start:

    allocate(arMaskOut(ubound(arin,1), ubound(arin,2)))

    call calcAryStatInfoMaskDouble2d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)

    deallocate(arMaskOut)

  end function getAryStatInfoDouble2d


  !---------------------------------------------------------------------
  subroutine calcAryStatInfoMaskSingle1d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    real(single), intent(in)  :: arin(:)
    type(aryStatInfoSingleT), intent(out) :: retInfo
    logical, intent(out) :: arMaskOut(:)	! Must be predefined.
    logical, intent(in), optional :: arMaskIn(:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(single), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    type(aryStatInfoFlagT) :: usedInfo
    logical, dimension(:), allocatable :: arMaskArea, arMaskVal
    real(single), dimension(:), allocatable :: arWork
    integer :: ubound_arin(rankArin)
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    ubound_arin(1) = ubound(arin,1)

    allocate(retInfo%minindices(rankArin), retInfo%maxindices(rankArin))
    allocate(arMaskArea( ubound_arin(1)))
    allocate(arMaskVal(  ubound_arin(1)))
    allocate(arWork(     ubound_arin(1)))
    arMaskArea = .true.	! Default is ALL true.
    arMaskVal  = .true.	! Default is ALL true.

    !! Sort out usedInfo
    if (present(flagInfo)) then
      usedInfo = flagInfo
      if (.not. present(minAreaIndices)) usedInfo%isAreaLowerUsed = .false.
      if (.not. present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .false.
      if (.not. present(valLower))       usedInfo%isValLowerUsed  = .false.
      if (.not. present(valUpper))       usedInfo%isValUpperUsed  = .false.
      if (.not. present(arMaskIn))       usedInfo%isMaskUsed      = .false.
    else
      if (present(minAreaIndices)) usedInfo%isAreaLowerUsed = .true.
      if (present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .true.
      if (present(valLower))       usedInfo%isValLowerUsed  = .true.
      if (present(valUpper))       usedInfo%isValUpperUsed  = .true.
      if (present(arMaskIn))       usedInfo%isMaskUsed      = .true.
    end if

    !! Sanity check for minAreaIndices and maxAreaIndices.
    if (usedInfo%isAreaLowerUsed .and. usedInfo%isAreaUpperUsed) then
      call chkAreaIndices(minAreaIndices, maxAreaIndices, retInfo%flag%status)
      retInfo%flag%status = retInfo%flag%status * 4	! The 3rd and 4th bit.
    end if

    !! AreaIndices --- Define (min|max)areaindices if _either_ of them is specified.
    if (usedInfo%isAreaLowerUsed .or. usedInfo%isAreaUpperUsed) then
      allocate(retInfo%minareaindices(rankArin), retInfo%maxareaindices(rankArin))
      if (usedInfo%isAreaLowerUsed) then
        retInfo%flag%isAreaLowerUsed = .true.
        retInfo%minareaindices = minAreaIndices
      else
        retInfo%minareaindices(:) = 1
      end if
      if (usedInfo%isAreaUpperUsed) then
        retInfo%flag%isAreaUpperUsed = .true.
        retInfo%maxareaindices = maxAreaIndices
      else
        retInfo%maxareaindices = ubound(arin)
      end if

      do i=1,ubound_arin(1)
          if (     (i < retInfo%minareaindices(1)) &
              .or. (i > retInfo%maxareaindices(1))) then
            arMaskArea(i) = .false.
          end if
      end do
    else
      ! Default
    end if

    !! val(Lower|Upper)
    if (usedInfo%isValLowerUsed) then
      retInfo%flag%isValLowerUsed = .true.
      retInfo%vallower = valLower
    else
      ! Do nothing (Default)
    end if
    if (usedInfo%isValUpperUsed) then
      retInfo%flag%isValUpperUsed = .true.
      retInfo%valupper = valUpper
    else
      ! Do nothing (Default)
    end if

    if (retInfo%flag%isValLowerUsed .and. retInfo%flag%isValUpperUsed) then
      where (arin < valLower .or. arin > valUpper) arMaskVal = .false.
    else if (retInfo%flag%isValLowerUsed) then
      where (arin < valLower) arMaskVal = .false.
    else if (retInfo%flag%isValUpperUsed) then
      where (arin > valUpper) arMaskVal = .false.
    end if

    arMaskOut = arMaskArea .and. arMaskVal

    if (present(arMaskIn)) then
      retInfo%flag%isMaskUsed = .true.
      arMaskOut = arMaskOut .and. arMaskIn
    end if

    arWork = 0.0
    where (arMaskOut) arWork = arin

    retInfo%totalentry = ubound_arin(1)
    retInfo%validentry = count(arMaskOut)

    retInfo%minindices = minloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    retInfo%maxindices = maxloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    ! retInfo%minval = arin(retInfo%minindices(1))
    ! retInfo%maxval = arin(retInfo%maxindices(1))

    retInfo%totalsum = sum(arWork)
    retInfo%realsum  = sum(real(arWork, double))	! to minimise rounded errors
    ! retInfo%realsum  = real(retInfo%totalsum, double)

    if (retInfo%validentry > 0) then
      retInfo%minval = arin(retInfo%minindices(1))
      retInfo%maxval = arin(retInfo%maxindices(1))
      retInfo%mean = retInfo%realsum / retInfo%validentry
      where (arMaskOut) arWork = (arin - retInfo%mean)**2	! arWork is modified.
      ! nb., arWork == 0.0 elsewhere
      ! Now arWork is an array of squared values.
      retInfo%sigma = sqrt(sum(real(arWork,double)) / retInfo%validentry)
    else
      ! NOTE: %minindices==(0), hence those indices can not be used.
      retInfo%minval = minval(arin, arMaskOut)	! eg.  3.4028235E+38
      retInfo%maxval = maxval(arin, arMaskOut)	! eg. -3.4028235E+38
      retInfo%flag%status = retInfo%flag%status - 1
      call message(AppMsg, Noisy, 'ssclib/array_utils: No valid entry in the input array.')
    end if

    deallocate(arMaskArea, arMaskVal)

  end subroutine calcAryStatInfoMaskSingle1d

  !---------------------------------------------------------------------
  subroutine calcAryStatInfoMaskSingle2d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    real(single), intent(in)  :: arin(:,:)
    type(aryStatInfoSingleT), intent(out) :: retInfo
    logical, intent(out) :: arMaskOut(:,:)	! Must be predefined.
    logical, intent(in), optional :: arMaskIn(:,:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(single), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    type(aryStatInfoFlagT) :: usedInfo
    logical, dimension(:,:), allocatable :: arMaskArea, arMaskVal
    real(single), dimension(:,:), allocatable :: arWork
    integer :: ubound_arin(rankArin)
    integer :: i, j

    !-------------------------------------------------------------------
    ! Start:

    ubound_arin(1) = ubound(arin,1)
    ubound_arin(2) = ubound(arin,2)

    allocate(retInfo%minindices(rankArin), retInfo%maxindices(rankArin))
    allocate(arMaskArea( ubound_arin(1), ubound_arin(2)))
    allocate(arMaskVal(  ubound_arin(1), ubound_arin(2)))
    allocate(arWork(     ubound_arin(1), ubound_arin(2)))
    arMaskArea = .true.	! Default is ALL true.
    arMaskVal  = .true.	! Default is ALL true.

    !! Sort out usedInfo
    if (present(flagInfo)) then
      usedInfo = flagInfo
      if (.not. present(minAreaIndices)) usedInfo%isAreaLowerUsed = .false.
      if (.not. present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .false.
      if (.not. present(valLower))       usedInfo%isValLowerUsed  = .false.
      if (.not. present(valUpper))       usedInfo%isValUpperUsed  = .false.
      if (.not. present(arMaskIn))       usedInfo%isMaskUsed      = .false.
    else
      if (present(minAreaIndices)) usedInfo%isAreaLowerUsed = .true.
      if (present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .true.
      if (present(valLower))       usedInfo%isValLowerUsed  = .true.
      if (present(valUpper))       usedInfo%isValUpperUsed  = .true.
      if (present(arMaskIn))       usedInfo%isMaskUsed      = .true.
    end if

    !! Sanity check for minAreaIndices and maxAreaIndices.
    if (usedInfo%isAreaLowerUsed .and. usedInfo%isAreaUpperUsed) then
      call chkAreaIndices(minAreaIndices, maxAreaIndices, retInfo%flag%status)
      retInfo%flag%status = retInfo%flag%status * 4	! The 3rd and 4th bit.
    end if

    !! AreaIndices --- Define (min|max)areaindices if _either_ of them is specified.
    if (usedInfo%isAreaLowerUsed .or. usedInfo%isAreaUpperUsed) then
      allocate(retInfo%minareaindices(rankArin), retInfo%maxareaindices(rankArin))
      if (usedInfo%isAreaLowerUsed) then
        retInfo%flag%isAreaLowerUsed = .true.
        retInfo%minareaindices = minAreaIndices
      else
        retInfo%minareaindices(:) = 1
      end if
      if (usedInfo%isAreaUpperUsed) then
        retInfo%flag%isAreaUpperUsed = .true.
        retInfo%maxareaindices = maxAreaIndices
      else
        retInfo%maxareaindices = ubound(arin)
      end if

      do i=1,ubound_arin(1)
        do j=1,ubound_arin(2)
          if (     (i < retInfo%minareaindices(1)) &
              .or. (i > retInfo%maxareaindices(1)) &
              .or. (j < retInfo%minareaindices(2)) &
              .or. (j > retInfo%maxareaindices(2))) then
            arMaskArea(i,j) = .false.
          end if
        end do
      end do
      ! arMaskArea = (arin(i,j) == arin(i,j) &	! This line is always .true.
      !    .and. (i >= retInfo%minareaindices(1)) &
      !    .and. (i <= retInfo%maxareaindices(1)) &
      !    .and. (j >= retInfo%minareaindices(2)) &
      !    .and. (j <= retInfo%maxareaindices(2))), &
      !    i=1,ubound_arin(1); j=1,ubound_arin(2)
    else
      ! Default
    end if

    !! val(Lower|Upper)
    if (usedInfo%isValLowerUsed) then
      retInfo%flag%isValLowerUsed = .true.
      retInfo%vallower = valLower
    else
      ! Do nothing (Default)
    end if
    if (usedInfo%isValUpperUsed) then
      retInfo%flag%isValUpperUsed = .true.
      retInfo%valupper = valUpper
    else
      ! Do nothing (Default)
    end if

    if (retInfo%flag%isValLowerUsed .and. retInfo%flag%isValUpperUsed) then
      where (arin < valLower .or. arin > valUpper) arMaskVal = .false.
    else if (retInfo%flag%isValLowerUsed) then
      where (arin < valLower) arMaskVal = .false.
    else if (retInfo%flag%isValUpperUsed) then
      where (arin > valUpper) arMaskVal = .false.
    end if

    arMaskOut = arMaskArea .and. arMaskVal

    if (present(arMaskIn)) then
      retInfo%flag%isMaskUsed = .true.
      arMaskOut = arMaskOut .and. arMaskIn
    end if

    arWork = 0.0
    where (arMaskOut) arWork = arin

    retInfo%totalentry = ubound_arin(1) * ubound_arin(2)
    retInfo%validentry = count(arMaskOut)

    retInfo%minindices = minloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    retInfo%maxindices = maxloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0

    retInfo%totalsum = sum(arWork)
    retInfo%realsum  = sum(real(arWork, double))	! to minimise rounded errors
    ! retInfo%realsum  = real(retInfo%totalsum, double)

    if (retInfo%validentry > 0) then
      retInfo%minval = arin(retInfo%minindices(1), retInfo%minindices(2))
      retInfo%maxval = arin(retInfo%maxindices(1), retInfo%maxindices(2))
      retInfo%mean = retInfo%realsum / retInfo%validentry
      where (arMaskOut) arWork = (arin - retInfo%mean)**2	! arWork is modified.
      ! nb., arWork == 0.0 elsewhere
      ! Now arWork is an array of squared values.
      retInfo%sigma = sqrt(sum(real(arWork,double)) / retInfo%validentry)
    else
      ! NOTE: %minindices==(0, 0), hence those indices can not be used.
      retInfo%minval = minval(arin, arMaskOut)	! eg.  3.4028235E+38
      retInfo%maxval = maxval(arin, arMaskOut)	! eg. -3.4028235E+38
      retInfo%flag%status = retInfo%flag%status - 1
      call message(AppMsg, Noisy, 'ssclib/array_utils: No valid entry in the input array.')
    end if

    deallocate(arMaskArea, arMaskVal)

  end subroutine calcAryStatInfoMaskSingle2d


  !---------------------------------------------------------------------
  function getAryStatInfoSingle1d(arin, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo) result(retInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    type(aryStatInfoSingleT) :: retInfo
    real(single), intent(in)  :: arin(:)
    logical, intent(in), optional :: arMaskIn(:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(single), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    logical, dimension(:), allocatable :: arMaskOut

    !-------------------------------------------------------------------
    ! Start:

    allocate(arMaskOut(ubound(arin,1)))

    call calcAryStatInfoMaskSingle1d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)

    deallocate(arMaskOut)

  end function getAryStatInfoSingle1d

  !---------------------------------------------------------------------
  function getAryStatInfoSingle2d(arin, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo) result(retInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    type(aryStatInfoSingleT) :: retInfo
    real(single), intent(in)  :: arin(:,:)
    logical, intent(in), optional :: arMaskIn(:,:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(single), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    logical, dimension(:,:), allocatable :: arMaskOut

    !-------------------------------------------------------------------
    ! Start:

    allocate(arMaskOut(ubound(arin,1), ubound(arin,2)))

    call calcAryStatInfoMaskSingle2d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)

    deallocate(arMaskOut)

  end function getAryStatInfoSingle2d


  !---------------------------------------------------------------------
  subroutine calcAryStatInfoMaskInt321d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    integer(int32), intent(in)  :: arin(:)
    type(aryStatInfoInt32T), intent(out) :: retInfo
    logical, intent(out) :: arMaskOut(:)	! Must be predefined.
    logical, intent(in), optional :: arMaskIn(:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    type(aryStatInfoFlagT) :: usedInfo
    logical, dimension(:), allocatable :: arMaskArea, arMaskVal
    integer(int32), dimension(:), allocatable :: arWork
    real(double),   dimension(:), allocatable :: arWorkDbl
    integer :: ubound_arin(rankArin)
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    ubound_arin(1) = ubound(arin,1)

    allocate(retInfo%minindices(rankArin), retInfo%maxindices(rankArin))
    allocate(arMaskArea( ubound_arin(1)))
    allocate(arMaskVal(  ubound_arin(1)))
    allocate(arWork(     ubound_arin(1)))
    allocate(arWorkDbl(  ubound_arin(1)))
    arMaskArea = .true.	! Default is ALL true.
    arMaskVal  = .true.	! Default is ALL true.

    !! Sort out usedInfo
    if (present(flagInfo)) then
      usedInfo = flagInfo
      if (.not. present(minAreaIndices)) usedInfo%isAreaLowerUsed = .false.
      if (.not. present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .false.
      if (.not. present(valLower))       usedInfo%isValLowerUsed  = .false.
      if (.not. present(valUpper))       usedInfo%isValUpperUsed  = .false.
      if (.not. present(arMaskIn))       usedInfo%isMaskUsed      = .false.
    else
      if (present(minAreaIndices)) usedInfo%isAreaLowerUsed = .true.
      if (present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .true.
      if (present(valLower))       usedInfo%isValLowerUsed  = .true.
      if (present(valUpper))       usedInfo%isValUpperUsed  = .true.
      if (present(arMaskIn))       usedInfo%isMaskUsed      = .true.
    end if

    !! Sanity check for minAreaIndices and maxAreaIndices.
    if (usedInfo%isAreaLowerUsed .and. usedInfo%isAreaUpperUsed) then
      call chkAreaIndices(minAreaIndices, maxAreaIndices, retInfo%flag%status)
      retInfo%flag%status = retInfo%flag%status * 4	! The 3rd and 4th bit.
    end if

    !! AreaIndices --- Define (min|max)areaindices if _either_ of them is specified.
    if (usedInfo%isAreaLowerUsed .or. usedInfo%isAreaUpperUsed) then
      allocate(retInfo%minareaindices(rankArin), retInfo%maxareaindices(rankArin))
      if (usedInfo%isAreaLowerUsed) then
        retInfo%flag%isAreaLowerUsed = .true.
        retInfo%minareaindices = minAreaIndices
      else
        retInfo%minareaindices(:) = 1
      end if
      if (usedInfo%isAreaUpperUsed) then
        retInfo%flag%isAreaUpperUsed = .true.
        retInfo%maxareaindices = maxAreaIndices
      else
        retInfo%maxareaindices = ubound(arin)
      end if

      do i=1,ubound_arin(1)
          if (     (i < retInfo%minareaindices(1)) &
              .or. (i > retInfo%maxareaindices(1))) then
            arMaskArea(i) = .false.
          end if
      end do
    else
      ! Default
    end if

    !! val(Lower|Upper)
    if (usedInfo%isValLowerUsed) then
      retInfo%flag%isValLowerUsed = .true.
      retInfo%vallower = valLower
    else
      ! Do nothing (Default)
    end if
    if (usedInfo%isValUpperUsed) then
      retInfo%flag%isValUpperUsed = .true.
      retInfo%valupper = valUpper
    else
      ! Do nothing (Default)
    end if

    if (retInfo%flag%isValLowerUsed .and. retInfo%flag%isValUpperUsed) then
      where (arin < valLower .or. arin > valUpper) arMaskVal = .false.
    else if (retInfo%flag%isValLowerUsed) then
      where (arin < valLower) arMaskVal = .false.
    else if (retInfo%flag%isValUpperUsed) then
      where (arin > valUpper) arMaskVal = .false.
    end if

    arMaskOut = arMaskArea .and. arMaskVal

    if (present(arMaskIn)) then
      retInfo%flag%isMaskUsed = .true.
      arMaskOut = arMaskOut .and. arMaskIn
    end if

    arWork = 0
    where (arMaskOut) arWork = arin

    retInfo%totalentry = ubound_arin(1)
    retInfo%validentry = count(arMaskOut)

    retInfo%minindices = minloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    retInfo%maxindices = maxloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    ! retInfo%minval = arin(retInfo%minindices(1))
    ! retInfo%maxval = arin(retInfo%maxindices(1))

    call sumInt32DblInt321d(arWork, retInfo%totalsum, retInfo%realsum)	! or call sumInt32Dbl()
    ! retInfo%totalsum = sumInt32(arWork)

    if (retInfo%validentry > 0) then
      retInfo%minval = arin(retInfo%minindices(1))
      retInfo%maxval = arin(retInfo%maxindices(1))
      !if (retInfo%totalsum == INTEGER_NULL) then
      !  arWorkDbl = real(arWork, double)
      !  sumDbl = sum(arWorkDbl)
      !else
      !  arWorkDbl = 0.0d0
      !  sumDbl = real(retInfo%totalsum, double)
      !end if
      !retInfo%mean = sumDbl / retInfo%validentry
      retInfo%mean = retInfo%realsum / retInfo%validentry
      arWorkDbl = 0.0d0
      where (arMaskOut) arWorkDbl = (arin - retInfo%mean)**2	! Double === retInfo%mean
      ! nb., arWorkDbl == 0 elsewhere
      ! Now arWorkDbl is an array of squared values.
      retInfo%sigma = sqrt(sum(arWorkDbl) / retInfo%validentry)
    else
      ! NOTE: %minindices==(0), hence those indices can not be used.
      retInfo%minval = minval(arin, arMaskOut)	! eg.  3.4028235E+38
      retInfo%maxval = maxval(arin, arMaskOut)	! eg. -3.4028235E+38
      retInfo%flag%status = retInfo%flag%status - 1
      call message(AppMsg, Noisy, 'ssclib/array_utils: No valid entry in the input array.')
    end if

    deallocate(arMaskArea, arMaskVal)

  end subroutine calcAryStatInfoMaskInt321d

  !---------------------------------------------------------------------
  subroutine calcAryStatInfoMaskInt322d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    integer(int32), intent(in)  :: arin(:,:)
    type(aryStatInfoInt32T), intent(out) :: retInfo
    logical, intent(out) :: arMaskOut(:,:)	! Must be predefined.
    logical, intent(in), optional :: arMaskIn(:,:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    type(aryStatInfoFlagT) :: usedInfo
    logical, dimension(:,:), allocatable :: arMaskArea, arMaskVal
    integer(int32), dimension(:,:), allocatable :: arWork
    real(double),   dimension(:,:), allocatable :: arWorkDbl
    integer :: ubound_arin(rankArin)
    integer :: i, j

    !-------------------------------------------------------------------
    ! Start:

    ubound_arin(1) = ubound(arin,1)
    ubound_arin(2) = ubound(arin,2)

    allocate(retInfo%minindices(rankArin), retInfo%maxindices(rankArin))
    allocate(arMaskArea( ubound_arin(1), ubound_arin(2)))
    allocate(arMaskVal(  ubound_arin(1), ubound_arin(2)))
    allocate(arWork(     ubound_arin(1), ubound_arin(2)))
    allocate(arWorkDbl(  ubound_arin(1), ubound_arin(2)))
    arMaskArea = .true.	! Default is ALL true.
    arMaskVal  = .true.	! Default is ALL true.

    !! Sort out usedInfo
    if (present(flagInfo)) then
      usedInfo = flagInfo
      if (.not. present(minAreaIndices)) usedInfo%isAreaLowerUsed = .false.
      if (.not. present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .false.
      if (.not. present(valLower))       usedInfo%isValLowerUsed  = .false.
      if (.not. present(valUpper))       usedInfo%isValUpperUsed  = .false.
      if (.not. present(arMaskIn))       usedInfo%isMaskUsed      = .false.
    else
      if (present(minAreaIndices)) usedInfo%isAreaLowerUsed = .true.
      if (present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .true.
      if (present(valLower))       usedInfo%isValLowerUsed  = .true.
      if (present(valUpper))       usedInfo%isValUpperUsed  = .true.
      if (present(arMaskIn))       usedInfo%isMaskUsed      = .true.
    end if

    !! Sanity check for minAreaIndices and maxAreaIndices.
    if (usedInfo%isAreaLowerUsed .and. usedInfo%isAreaUpperUsed) then
      call chkAreaIndices(minAreaIndices, maxAreaIndices, retInfo%flag%status)
      retInfo%flag%status = retInfo%flag%status * 4	! The 3rd and 4th bit.
    end if

    !! AreaIndices --- Define (min|max)areaindices if _either_ of them is specified.
    if (usedInfo%isAreaLowerUsed .or. usedInfo%isAreaUpperUsed) then
      allocate(retInfo%minareaindices(rankArin), retInfo%maxareaindices(rankArin))
      if (usedInfo%isAreaLowerUsed) then
        retInfo%flag%isAreaLowerUsed = .true.
        retInfo%minareaindices = minAreaIndices
      else
        retInfo%minareaindices(:) = 1
      end if
      if (usedInfo%isAreaUpperUsed) then
        retInfo%flag%isAreaUpperUsed = .true.
        retInfo%maxareaindices = maxAreaIndices
      else
        retInfo%maxareaindices = ubound(arin)
      end if

      do i=1,ubound_arin(1)
        do j=1,ubound_arin(2)
          if (     (i < retInfo%minareaindices(1)) &
              .or. (i > retInfo%maxareaindices(1)) &
              .or. (j < retInfo%minareaindices(2)) &
              .or. (j > retInfo%maxareaindices(2))) then
            arMaskArea(i,j) = .false.
          end if
        end do
      end do
      ! arMaskArea = (arin(i,j) == arin(i,j) &	! This line is always .true.
      !    .and. (i >= retInfo%minareaindices(1)) &
      !    .and. (i <= retInfo%maxareaindices(1)) &
      !    .and. (j >= retInfo%minareaindices(2)) &
      !    .and. (j <= retInfo%maxareaindices(2))), &
      !    i=1,ubound_arin(1); j=1,ubound_arin(2)
    else
      ! Default
    end if

    !! val(Lower|Upper)
    if (usedInfo%isValLowerUsed) then
      retInfo%flag%isValLowerUsed = .true.
      retInfo%vallower = valLower
    else
      ! Do nothing (Default)
    end if
    if (usedInfo%isValUpperUsed) then
      retInfo%flag%isValUpperUsed = .true.
      retInfo%valupper = valUpper
    else
      ! Do nothing (Default)
    end if

    if (retInfo%flag%isValLowerUsed .and. retInfo%flag%isValUpperUsed) then
      where (arin < valLower .or. arin > valUpper) arMaskVal = .false.
    else if (retInfo%flag%isValLowerUsed) then
      where (arin < valLower) arMaskVal = .false.
    else if (retInfo%flag%isValUpperUsed) then
      where (arin > valUpper) arMaskVal = .false.
    end if

    arMaskOut = arMaskArea .and. arMaskVal

    if (present(arMaskIn)) then
      retInfo%flag%isMaskUsed = .true.
      arMaskOut = arMaskOut .and. arMaskIn
    end if

    arWork = 0
    where (arMaskOut) arWork = arin

    retInfo%totalentry = ubound_arin(1) * ubound_arin(2)
    retInfo%validentry = count(arMaskOut)

    retInfo%minindices = minloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    retInfo%maxindices = maxloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0

    call sumInt32DblInt322d(arWork, retInfo%totalsum, retInfo%realsum)	! or call sumInt32Dbl()
    ! retInfo%totalsum = sumInt32(arWork)
    if (retInfo%validentry > 0) then
      retInfo%minval = arin(retInfo%minindices(1), retInfo%minindices(2))
      retInfo%maxval = arin(retInfo%maxindices(1), retInfo%maxindices(2))
      !if (retInfo%totalsum == INTEGER_NULL) then
      !  arWorkDbl = real(arWork, double)
      !  sumDbl = sum(arWorkDbl)
      !else
      !  arWorkDbl = 0.0d0
      !  sumDbl = real(retInfo%totalsum, double)
      !end if
      !retInfo%mean = sumDbl / retInfo%validentry
      retInfo%mean = retInfo%realsum / retInfo%validentry
      arWorkDbl = 0.0d0
      where (arMaskOut) arWorkDbl = (arin - retInfo%mean)**2	! Double === retInfo%mean
      ! nb., arWorkDbl == 0 elsewhere
      ! Now arWorkDbl is an array of squared values.
      retInfo%sigma = sqrt(sum(arWorkDbl) / retInfo%validentry)
    else
      ! NOTE: %minindices==(0, 0), hence those indices can not be used.
      retInfo%minval = minval(arin, arMaskOut)	! eg. 
      retInfo%maxval = maxval(arin, arMaskOut)	! eg. 
      retInfo%flag%status = retInfo%flag%status - 1
      call message(AppMsg, Noisy, 'ssclib/array_utils: No valid entry in the input array.')
    end if

    deallocate(arMaskArea, arMaskVal)

  end subroutine calcAryStatInfoMaskInt322d


  !---------------------------------------------------------------------
  function getAryStatInfoInt321d(arin, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo) result(retInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    type(aryStatInfoInt32T) :: retInfo
    integer(int32), intent(in)  :: arin(:)
    logical, intent(in), optional :: arMaskIn(:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    logical, dimension(:), allocatable :: arMaskOut

    !-------------------------------------------------------------------
    ! Start:

    allocate(arMaskOut(ubound(arin,1)))

    call calcAryStatInfoMaskInt321d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)

    deallocate(arMaskOut)

  end function getAryStatInfoInt321d

  !---------------------------------------------------------------------
  function getAryStatInfoInt322d(arin, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo) result(retInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    type(aryStatInfoInt32T) :: retInfo
    integer(int32), intent(in)  :: arin(:,:)
    logical, intent(in), optional :: arMaskIn(:,:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    logical, dimension(:,:), allocatable :: arMaskOut

    !-------------------------------------------------------------------
    ! Start:

    allocate(arMaskOut(ubound(arin,1), ubound(arin,2)))

    call calcAryStatInfoMaskInt322d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)

    deallocate(arMaskOut)

  end function getAryStatInfoInt322d


  !---------------------------------------------------------------------
  subroutine calcAryStatInfoMaskInt161d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    integer(int16), intent(in)  :: arin(:)
    type(aryStatInfoInt16T), intent(out) :: retInfo
    logical, intent(out) :: arMaskOut(:)	! Must be predefined.
    logical, intent(in), optional :: arMaskIn(:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    type(aryStatInfoFlagT) :: usedInfo
    logical, dimension(:), allocatable :: arMaskArea, arMaskVal
    integer(int16), dimension(:), allocatable :: arWork
    real(double),   dimension(:), allocatable :: arWorkDbl
    integer :: ubound_arin(rankArin)
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    ubound_arin(1) = ubound(arin,1)

    allocate(retInfo%minindices(rankArin), retInfo%maxindices(rankArin))
    allocate(arMaskArea( ubound_arin(1)))
    allocate(arMaskVal(  ubound_arin(1)))
    allocate(arWork(     ubound_arin(1)))
    allocate(arWorkDbl(  ubound_arin(1)))
    arMaskArea = .true.	! Default is ALL true.
    arMaskVal  = .true.	! Default is ALL true.

    !! Sort out usedInfo
    if (present(flagInfo)) then
      usedInfo = flagInfo
      if (.not. present(minAreaIndices)) usedInfo%isAreaLowerUsed = .false.
      if (.not. present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .false.
      if (.not. present(valLower))       usedInfo%isValLowerUsed  = .false.
      if (.not. present(valUpper))       usedInfo%isValUpperUsed  = .false.
      if (.not. present(arMaskIn))       usedInfo%isMaskUsed      = .false.
    else
      if (present(minAreaIndices)) usedInfo%isAreaLowerUsed = .true.
      if (present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .true.
      if (present(valLower))       usedInfo%isValLowerUsed  = .true.
      if (present(valUpper))       usedInfo%isValUpperUsed  = .true.
      if (present(arMaskIn))       usedInfo%isMaskUsed      = .true.
    end if

    !! Sanity check for minAreaIndices and maxAreaIndices.
    if (usedInfo%isAreaLowerUsed .and. usedInfo%isAreaUpperUsed) then
      call chkAreaIndices(minAreaIndices, maxAreaIndices, retInfo%flag%status)
      retInfo%flag%status = retInfo%flag%status * 4	! The 3rd and 4th bit.
    end if

    !! AreaIndices --- Define (min|max)areaindices if _either_ of them is specified.
    if (usedInfo%isAreaLowerUsed .or. usedInfo%isAreaUpperUsed) then
      allocate(retInfo%minareaindices(rankArin), retInfo%maxareaindices(rankArin))
      if (usedInfo%isAreaLowerUsed) then
        retInfo%flag%isAreaLowerUsed = .true.
        retInfo%minareaindices = minAreaIndices
      else
        retInfo%minareaindices(:) = 1
      end if
      if (usedInfo%isAreaUpperUsed) then
        retInfo%flag%isAreaUpperUsed = .true.
        retInfo%maxareaindices = maxAreaIndices
      else
        retInfo%maxareaindices = ubound(arin)
      end if

      do i=1,ubound_arin(1)
          if (     (i < retInfo%minareaindices(1)) &
              .or. (i > retInfo%maxareaindices(1))) then
            arMaskArea(i) = .false.
          end if
      end do
    else
      ! Default
    end if

    !! val(Lower|Upper)
    if (usedInfo%isValLowerUsed) then
      retInfo%flag%isValLowerUsed = .true.
      retInfo%vallower = valLower
    else
      ! Do nothing (Default)
    end if
    if (usedInfo%isValUpperUsed) then
      retInfo%flag%isValUpperUsed = .true.
      retInfo%valupper = valUpper
    else
      ! Do nothing (Default)
    end if

    if (retInfo%flag%isValLowerUsed .and. retInfo%flag%isValUpperUsed) then
      where (arin < valLower .or. arin > valUpper) arMaskVal = .false.
    else if (retInfo%flag%isValLowerUsed) then
      where (arin < valLower) arMaskVal = .false.
    else if (retInfo%flag%isValUpperUsed) then
      where (arin > valUpper) arMaskVal = .false.
    end if

    arMaskOut = arMaskArea .and. arMaskVal

    if (present(arMaskIn)) then
      retInfo%flag%isMaskUsed = .true.
      arMaskOut = arMaskOut .and. arMaskIn
    end if

    arWork = 0
    where (arMaskOut) arWork = arin

    retInfo%totalentry = ubound_arin(1)
    retInfo%validentry = count(arMaskOut)

    retInfo%minindices = minloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    retInfo%maxindices = maxloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    ! retInfo%minval = arin(retInfo%minindices(1))
    ! retInfo%maxval = arin(retInfo%maxindices(1))

    call sumInt32DblInt161d(arWork, retInfo%totalsum, retInfo%realsum)	! or call sumInt32Dbl()

    if (retInfo%validentry > 0) then
      retInfo%minval = arin(retInfo%minindices(1))
      retInfo%maxval = arin(retInfo%maxindices(1))
      retInfo%mean = retInfo%realsum / retInfo%validentry
      arWorkDbl = 0.0d0
      where (arMaskOut) arWorkDbl = (arin - retInfo%mean)**2	! Double === retInfo%mean
      ! nb., arWorkDbl == 0 elsewhere
      ! Now arWorkDbl is an array of squared values.
      retInfo%sigma = sqrt(sum(arWorkDbl) / retInfo%validentry)
    else
      ! NOTE: %minindices==(0), hence those indices can not be used.
      retInfo%minval = minval(arin, arMaskOut)	! eg.  3.4028235E+38
      retInfo%maxval = maxval(arin, arMaskOut)	! eg. -3.4028235E+38
      retInfo%flag%status = retInfo%flag%status - 1
      call message(AppMsg, Noisy, 'ssclib/array_utils: No valid entry in the input array.')
    end if

    deallocate(arMaskArea, arMaskVal)

  end subroutine calcAryStatInfoMaskInt161d

  !---------------------------------------------------------------------
  subroutine calcAryStatInfoMaskInt162d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    integer(int16), intent(in)  :: arin(:,:)
    type(aryStatInfoInt16T), intent(out) :: retInfo
    logical, intent(out) :: arMaskOut(:,:)	! Must be predefined.
    logical, intent(in), optional :: arMaskIn(:,:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    type(aryStatInfoFlagT) :: usedInfo
    logical, dimension(:,:), allocatable :: arMaskArea, arMaskVal
    integer(int16), dimension(:,:), allocatable :: arWork
    real(double),   dimension(:,:), allocatable :: arWorkDbl
    integer :: ubound_arin(rankArin)
    integer :: i, j

    !-------------------------------------------------------------------
    ! Start:

    ubound_arin(1) = ubound(arin,1)
    ubound_arin(2) = ubound(arin,2)

    allocate(retInfo%minindices(rankArin), retInfo%maxindices(rankArin))
    allocate(arMaskArea( ubound_arin(1), ubound_arin(2)))
    allocate(arMaskVal(  ubound_arin(1), ubound_arin(2)))
    allocate(arWork(     ubound_arin(1), ubound_arin(2)))
    allocate(arWorkDbl(  ubound_arin(1), ubound_arin(2)))
    arMaskArea = .true.	! Default is ALL true.
    arMaskVal  = .true.	! Default is ALL true.

    !! Sort out usedInfo
    if (present(flagInfo)) then
      usedInfo = flagInfo
      if (.not. present(minAreaIndices)) usedInfo%isAreaLowerUsed = .false.
      if (.not. present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .false.
      if (.not. present(valLower))       usedInfo%isValLowerUsed  = .false.
      if (.not. present(valUpper))       usedInfo%isValUpperUsed  = .false.
      if (.not. present(arMaskIn))       usedInfo%isMaskUsed      = .false.
    else
      if (present(minAreaIndices)) usedInfo%isAreaLowerUsed = .true.
      if (present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .true.
      if (present(valLower))       usedInfo%isValLowerUsed  = .true.
      if (present(valUpper))       usedInfo%isValUpperUsed  = .true.
      if (present(arMaskIn))       usedInfo%isMaskUsed      = .true.
    end if

    !! Sanity check for minAreaIndices and maxAreaIndices.
    if (usedInfo%isAreaLowerUsed .and. usedInfo%isAreaUpperUsed) then
      call chkAreaIndices(minAreaIndices, maxAreaIndices, retInfo%flag%status)
      retInfo%flag%status = retInfo%flag%status * 4	! The 3rd and 4th bit.
    end if

    !! AreaIndices --- Define (min|max)areaindices if _either_ of them is specified.
    if (usedInfo%isAreaLowerUsed .or. usedInfo%isAreaUpperUsed) then
      allocate(retInfo%minareaindices(rankArin), retInfo%maxareaindices(rankArin))
      if (usedInfo%isAreaLowerUsed) then
        retInfo%flag%isAreaLowerUsed = .true.
        retInfo%minareaindices = minAreaIndices
      else
        retInfo%minareaindices(:) = 1
      end if
      if (usedInfo%isAreaUpperUsed) then
        retInfo%flag%isAreaUpperUsed = .true.
        retInfo%maxareaindices = maxAreaIndices
      else
        retInfo%maxareaindices = ubound(arin)
      end if

      do i=1,ubound_arin(1)
        do j=1,ubound_arin(2)
          if (     (i < retInfo%minareaindices(1)) &
              .or. (i > retInfo%maxareaindices(1)) &
              .or. (j < retInfo%minareaindices(2)) &
              .or. (j > retInfo%maxareaindices(2))) then
            arMaskArea(i,j) = .false.
          end if
        end do
      end do
      ! arMaskArea = (arin(i,j) == arin(i,j) &	! This line is always .true.
      !    .and. (i >= retInfo%minareaindices(1)) &
      !    .and. (i <= retInfo%maxareaindices(1)) &
      !    .and. (j >= retInfo%minareaindices(2)) &
      !    .and. (j <= retInfo%maxareaindices(2))), &
      !    i=1,ubound_arin(1); j=1,ubound_arin(2)
    else
      ! Default
    end if

    !! val(Lower|Upper)
    if (usedInfo%isValLowerUsed) then
      retInfo%flag%isValLowerUsed = .true.
      retInfo%vallower = valLower
    else
      ! Do nothing (Default)
    end if
    if (usedInfo%isValUpperUsed) then
      retInfo%flag%isValUpperUsed = .true.
      retInfo%valupper = valUpper
    else
      ! Do nothing (Default)
    end if

    if (retInfo%flag%isValLowerUsed .and. retInfo%flag%isValUpperUsed) then
      where (arin < valLower .or. arin > valUpper) arMaskVal = .false.
    else if (retInfo%flag%isValLowerUsed) then
      where (arin < valLower) arMaskVal = .false.
    else if (retInfo%flag%isValUpperUsed) then
      where (arin > valUpper) arMaskVal = .false.
    end if

    arMaskOut = arMaskArea .and. arMaskVal

    if (present(arMaskIn)) then
      retInfo%flag%isMaskUsed = .true.
      arMaskOut = arMaskOut .and. arMaskIn
    end if

    arWork = 0
    where (arMaskOut) arWork = arin

    retInfo%totalentry = ubound_arin(1) * ubound_arin(2)
    retInfo%validentry = count(arMaskOut)

    retInfo%minindices = minloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    retInfo%maxindices = maxloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0

    call sumInt32DblInt162d(arWork, retInfo%totalsum, retInfo%realsum)	! or call sumInt32Dbl()
    if (retInfo%validentry > 0) then
      retInfo%minval = arin(retInfo%minindices(1), retInfo%minindices(2))
      retInfo%maxval = arin(retInfo%maxindices(1), retInfo%maxindices(2))
      retInfo%mean = retInfo%realsum / retInfo%validentry
      arWorkDbl = 0.0d0
      where (arMaskOut) arWorkDbl = (arin - retInfo%mean)**2	! Double === retInfo%mean
      ! nb., arWorkDbl == 0 elsewhere
      ! Now arWorkDbl is an array of squared values.
      retInfo%sigma = sqrt(sum(arWorkDbl) / retInfo%validentry)
    else
      ! NOTE: %minindices==(0, 0), hence those indices can not be used.
      retInfo%minval = minval(arin, arMaskOut)	! eg. 
      retInfo%maxval = maxval(arin, arMaskOut)	! eg. 
      retInfo%flag%status = retInfo%flag%status - 1
      call message(AppMsg, Noisy, 'ssclib/array_utils: No valid entry in the input array.')
    end if

    deallocate(arMaskArea, arMaskVal)

  end subroutine calcAryStatInfoMaskInt162d


  !---------------------------------------------------------------------
  function getAryStatInfoInt161d(arin, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo) result(retInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    type(aryStatInfoInt16T) :: retInfo
    integer(int16), intent(in)  :: arin(:)
    logical, intent(in), optional :: arMaskIn(:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    logical, dimension(:), allocatable :: arMaskOut

    !-------------------------------------------------------------------
    ! Start:

    allocate(arMaskOut(ubound(arin,1)))

    call calcAryStatInfoMaskInt161d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)

    deallocate(arMaskOut)

  end function getAryStatInfoInt161d

  !---------------------------------------------------------------------
  function getAryStatInfoInt162d(arin, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo) result(retInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    type(aryStatInfoInt16T) :: retInfo
    integer(int16), intent(in)  :: arin(:,:)
    logical, intent(in), optional :: arMaskIn(:,:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    logical, dimension(:,:), allocatable :: arMaskOut

    !-------------------------------------------------------------------
    ! Start:

    allocate(arMaskOut(ubound(arin,1), ubound(arin,2)))

    call calcAryStatInfoMaskInt162d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)

    deallocate(arMaskOut)

  end function getAryStatInfoInt162d


  !---------------------------------------------------------------------
  subroutine calcAryStatInfoMaskInt81d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    integer(int8), intent(in)  :: arin(:)
    type(aryStatInfoInt8T), intent(out) :: retInfo
    logical, intent(out) :: arMaskOut(:)	! Must be predefined.
    logical, intent(in), optional :: arMaskIn(:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    type(aryStatInfoFlagT) :: usedInfo
    logical, dimension(:), allocatable :: arMaskArea, arMaskVal
    integer(int8), dimension(:), allocatable :: arWork
    real(double),  dimension(:), allocatable :: arWorkDbl
    integer :: ubound_arin(rankArin)
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    ubound_arin(1) = ubound(arin,1)

    allocate(retInfo%minindices(rankArin), retInfo%maxindices(rankArin))
    allocate(arMaskArea( ubound_arin(1)))
    allocate(arMaskVal(  ubound_arin(1)))
    allocate(arWork(     ubound_arin(1)))
    allocate(arWorkDbl(  ubound_arin(1)))
    arMaskArea = .true.	! Default is ALL true.
    arMaskVal  = .true.	! Default is ALL true.

    !! Sort out usedInfo
    if (present(flagInfo)) then
      usedInfo = flagInfo
      if (.not. present(minAreaIndices)) usedInfo%isAreaLowerUsed = .false.
      if (.not. present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .false.
      if (.not. present(valLower))       usedInfo%isValLowerUsed  = .false.
      if (.not. present(valUpper))       usedInfo%isValUpperUsed  = .false.
      if (.not. present(arMaskIn))       usedInfo%isMaskUsed      = .false.
    else
      if (present(minAreaIndices)) usedInfo%isAreaLowerUsed = .true.
      if (present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .true.
      if (present(valLower))       usedInfo%isValLowerUsed  = .true.
      if (present(valUpper))       usedInfo%isValUpperUsed  = .true.
      if (present(arMaskIn))       usedInfo%isMaskUsed      = .true.
    end if

    !! Sanity check for minAreaIndices and maxAreaIndices.
    if (usedInfo%isAreaLowerUsed .and. usedInfo%isAreaUpperUsed) then
      call chkAreaIndices(minAreaIndices, maxAreaIndices, retInfo%flag%status)
      retInfo%flag%status = retInfo%flag%status * 4	! The 3rd and 4th bit.
    end if

    !! AreaIndices --- Define (min|max)areaindices if _either_ of them is specified.
    if (usedInfo%isAreaLowerUsed .or. usedInfo%isAreaUpperUsed) then
      allocate(retInfo%minareaindices(rankArin), retInfo%maxareaindices(rankArin))
      if (usedInfo%isAreaLowerUsed) then
        retInfo%flag%isAreaLowerUsed = .true.
        retInfo%minareaindices = minAreaIndices
      else
        retInfo%minareaindices(:) = 1
      end if
      if (usedInfo%isAreaUpperUsed) then
        retInfo%flag%isAreaUpperUsed = .true.
        retInfo%maxareaindices = maxAreaIndices
      else
        retInfo%maxareaindices = ubound(arin)
      end if

      do i=1,ubound_arin(1)
          if (     (i < retInfo%minareaindices(1)) &
              .or. (i > retInfo%maxareaindices(1))) then
            arMaskArea(i) = .false.
          end if
      end do
    else
      ! Default
    end if

    !! val(Lower|Upper)
    if (usedInfo%isValLowerUsed) then
      retInfo%flag%isValLowerUsed = .true.
      retInfo%vallower = valLower
    else
      ! Do nothing (Default)
    end if
    if (usedInfo%isValUpperUsed) then
      retInfo%flag%isValUpperUsed = .true.
      retInfo%valupper = valUpper
    else
      ! Do nothing (Default)
    end if

    if (retInfo%flag%isValLowerUsed .and. retInfo%flag%isValUpperUsed) then
      where (arin < valLower .or. arin > valUpper) arMaskVal = .false.
    else if (retInfo%flag%isValLowerUsed) then
      where (arin < valLower) arMaskVal = .false.
    else if (retInfo%flag%isValUpperUsed) then
      where (arin > valUpper) arMaskVal = .false.
    end if

    arMaskOut = arMaskArea .and. arMaskVal

    if (present(arMaskIn)) then
      retInfo%flag%isMaskUsed = .true.
      arMaskOut = arMaskOut .and. arMaskIn
    end if

    arWork = 0
    where (arMaskOut) arWork = arin

    retInfo%totalentry = ubound_arin(1)
    retInfo%validentry = count(arMaskOut)

    retInfo%minindices = minloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    retInfo%maxindices = maxloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    ! retInfo%minval = arin(retInfo%minindices(1))
    ! retInfo%maxval = arin(retInfo%maxindices(1))

    call sumInt32DblInt81d(arWork, retInfo%totalsum, retInfo%realsum)	! or call sumInt32Dbl()

    if (retInfo%validentry > 0) then
      retInfo%minval = arin(retInfo%minindices(1))
      retInfo%maxval = arin(retInfo%maxindices(1))
      retInfo%mean = retInfo%realsum / retInfo%validentry
      arWorkDbl = 0.0d0
      where (arMaskOut) arWorkDbl = (arin - retInfo%mean)**2	! Double === retInfo%mean
      ! nb., arWorkDbl == 0 elsewhere
      ! Now arWorkDbl is an array of squared values.
      retInfo%sigma = sqrt(sum(arWorkDbl) / retInfo%validentry)
    else
      ! NOTE: %minindices==(0), hence those indices can not be used.
      retInfo%minval = minval(arin, arMaskOut)	! eg.  3.4028235E+38
      retInfo%maxval = maxval(arin, arMaskOut)	! eg. -3.4028235E+38
      retInfo%flag%status = retInfo%flag%status - 1
      call message(AppMsg, Noisy, 'ssclib/array_utils: No valid entry in the input array.')
    end if

    deallocate(arMaskArea, arMaskVal)

  end subroutine calcAryStatInfoMaskInt81d

  !---------------------------------------------------------------------
  subroutine calcAryStatInfoMaskInt82d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    integer(int8), intent(in)  :: arin(:,:)
    type(aryStatInfoInt8T), intent(out) :: retInfo
    logical, intent(out) :: arMaskOut(:,:)	! Must be predefined.
    logical, intent(in), optional :: arMaskIn(:,:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    type(aryStatInfoFlagT) :: usedInfo
    logical, dimension(:,:), allocatable :: arMaskArea, arMaskVal
    integer(int8), dimension(:,:), allocatable :: arWork
    real(double),  dimension(:,:), allocatable :: arWorkDbl
    integer :: ubound_arin(rankArin)
    integer :: i, j

    !-------------------------------------------------------------------
    ! Start:

    ubound_arin(1) = ubound(arin,1)
    ubound_arin(2) = ubound(arin,2)

    allocate(retInfo%minindices(rankArin), retInfo%maxindices(rankArin))
    allocate(arMaskArea( ubound_arin(1), ubound_arin(2)))
    allocate(arMaskVal(  ubound_arin(1), ubound_arin(2)))
    allocate(arWork(     ubound_arin(1), ubound_arin(2)))
    allocate(arWorkDbl(  ubound_arin(1), ubound_arin(2)))
    arMaskArea = .true.	! Default is ALL true.
    arMaskVal  = .true.	! Default is ALL true.

    !! Sort out usedInfo
    if (present(flagInfo)) then
      usedInfo = flagInfo
      if (.not. present(minAreaIndices)) usedInfo%isAreaLowerUsed = .false.
      if (.not. present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .false.
      if (.not. present(valLower))       usedInfo%isValLowerUsed  = .false.
      if (.not. present(valUpper))       usedInfo%isValUpperUsed  = .false.
      if (.not. present(arMaskIn))       usedInfo%isMaskUsed      = .false.
    else
      if (present(minAreaIndices)) usedInfo%isAreaLowerUsed = .true.
      if (present(maxAreaIndices)) usedInfo%isAreaUpperUsed = .true.
      if (present(valLower))       usedInfo%isValLowerUsed  = .true.
      if (present(valUpper))       usedInfo%isValUpperUsed  = .true.
      if (present(arMaskIn))       usedInfo%isMaskUsed      = .true.
    end if

    !! Sanity check for minAreaIndices and maxAreaIndices.
    if (usedInfo%isAreaLowerUsed .and. usedInfo%isAreaUpperUsed) then
      call chkAreaIndices(minAreaIndices, maxAreaIndices, retInfo%flag%status)
      retInfo%flag%status = retInfo%flag%status * 4	! The 3rd and 4th bit.
    end if

    !! AreaIndices --- Define (min|max)areaindices if _either_ of them is specified.
    if (usedInfo%isAreaLowerUsed .or. usedInfo%isAreaUpperUsed) then
      allocate(retInfo%minareaindices(rankArin), retInfo%maxareaindices(rankArin))
      if (usedInfo%isAreaLowerUsed) then
        retInfo%flag%isAreaLowerUsed = .true.
        retInfo%minareaindices = minAreaIndices
      else
        retInfo%minareaindices(:) = 1
      end if
      if (usedInfo%isAreaUpperUsed) then
        retInfo%flag%isAreaUpperUsed = .true.
        retInfo%maxareaindices = maxAreaIndices
      else
        retInfo%maxareaindices = ubound(arin)
      end if

      do i=1,ubound_arin(1)
        do j=1,ubound_arin(2)
          if (     (i < retInfo%minareaindices(1)) &
              .or. (i > retInfo%maxareaindices(1)) &
              .or. (j < retInfo%minareaindices(2)) &
              .or. (j > retInfo%maxareaindices(2))) then
            arMaskArea(i,j) = .false.
          end if
        end do
      end do
      ! arMaskArea = (arin(i,j) == arin(i,j) &	! This line is always .true.
      !    .and. (i >= retInfo%minareaindices(1)) &
      !    .and. (i <= retInfo%maxareaindices(1)) &
      !    .and. (j >= retInfo%minareaindices(2)) &
      !    .and. (j <= retInfo%maxareaindices(2))), &
      !    i=1,ubound_arin(1); j=1,ubound_arin(2)
    else
      ! Default
    end if

    !! val(Lower|Upper)
    if (usedInfo%isValLowerUsed) then
      retInfo%flag%isValLowerUsed = .true.
      retInfo%vallower = valLower
    else
      ! Do nothing (Default)
    end if
    if (usedInfo%isValUpperUsed) then
      retInfo%flag%isValUpperUsed = .true.
      retInfo%valupper = valUpper
    else
      ! Do nothing (Default)
    end if

    if (retInfo%flag%isValLowerUsed .and. retInfo%flag%isValUpperUsed) then
      where (arin < valLower .or. arin > valUpper) arMaskVal = .false.
    else if (retInfo%flag%isValLowerUsed) then
      where (arin < valLower) arMaskVal = .false.
    else if (retInfo%flag%isValUpperUsed) then
      where (arin > valUpper) arMaskVal = .false.
    end if

    arMaskOut = arMaskArea .and. arMaskVal

    if (present(arMaskIn)) then
      retInfo%flag%isMaskUsed = .true.
      arMaskOut = arMaskOut .and. arMaskIn
    end if

    arWork = 0
    where (arMaskOut) arWork = arin

    retInfo%totalentry = ubound_arin(1) * ubound_arin(2)
    retInfo%validentry = count(arMaskOut)

    retInfo%minindices = minloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0
    retInfo%maxindices = maxloc(arin, arMaskOut)	! 0 if retInfo%validentry == 0

    call sumInt32DblInt82d(arWork, retInfo%totalsum, retInfo%realsum)	! or call sumInt32Dbl()
    if (retInfo%validentry > 0) then
      retInfo%minval = arin(retInfo%minindices(1), retInfo%minindices(2))
      retInfo%maxval = arin(retInfo%maxindices(1), retInfo%maxindices(2))
      retInfo%mean = retInfo%realsum / retInfo%validentry
      arWorkDbl = 0.0d0
      where (arMaskOut) arWorkDbl = (arin - retInfo%mean)**2	! Double === retInfo%mean
      ! nb., arWorkDbl == 0 elsewhere
      ! Now arWorkDbl is an array of squared values.
      retInfo%sigma = sqrt(sum(arWorkDbl) / retInfo%validentry)
    else
      ! NOTE: %minindices==(0, 0), hence those indices can not be used.
      retInfo%minval = minval(arin, arMaskOut)	! eg. 
      retInfo%maxval = maxval(arin, arMaskOut)	! eg. 
      retInfo%flag%status = retInfo%flag%status - 1
      call message(AppMsg, Noisy, 'ssclib/array_utils: No valid entry in the input array.')
    end if

    deallocate(arMaskArea, arMaskVal)

  end subroutine calcAryStatInfoMaskInt82d


  !---------------------------------------------------------------------
  function getAryStatInfoInt81d(arin, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo) result(retInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 1	! = size(lbound(arin))

    type(aryStatInfoInt8T) :: retInfo
    integer(int8), intent(in)  :: arin(:)
    logical, intent(in), optional :: arMaskIn(:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    logical, dimension(:), allocatable :: arMaskOut

    !-------------------------------------------------------------------
    ! Start:

    allocate(arMaskOut(ubound(arin,1)))

    call calcAryStatInfoMaskInt81d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)

    deallocate(arMaskOut)

  end function getAryStatInfoInt81d

  !---------------------------------------------------------------------
  function getAryStatInfoInt82d(arin, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo) result(retInfo)
  !---------------------------------------------------------------------
  ! See comment lines in subroutine calcAryStatInfoMaskDouble2d() for detail.

    implicit none

    integer, parameter :: rankArin = 2	! = size(lbound(arin))

    type(aryStatInfoInt8T) :: retInfo
    integer(int8), intent(in)  :: arin(:,:)
    logical, intent(in), optional :: arMaskIn(:,:)
    integer(int32), intent(in), optional :: minAreaIndices(rankArin), maxAreaIndices(rankArin)
    real(double), intent(in), optional :: valLower, valUpper
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    logical, dimension(:,:), allocatable :: arMaskOut

    !-------------------------------------------------------------------
    ! Start:

    allocate(arMaskOut(ubound(arin,1), ubound(arin,2)))

    call calcAryStatInfoMaskInt82d(arin, retInfo, arMaskOut, arMaskIn &
     , minAreaIndices, maxAreaIndices, valLower, valUpper, flagInfo)

    deallocate(arMaskOut)

  end function getAryStatInfoInt82d


  !---------------------------------------------------------------------
  subroutine chkAreaIndices(minAreaIndices, maxAreaIndices, status, flagInfo)
  !---------------------------------------------------------------------
  ! Check the validity of (minAreaIndices, maxAreaIndices),
  ! only when both minAreaIndices and maxAreaIndices are given,
  ! and if (flagInfo%isAreaLowerUsed .and. flagInfo%isAreaUpperUsed) if given.
  ! Returns integer status if desired (0 if normal ends or if not both are given).

    implicit none

    integer(int32), intent(in),  optional :: minAreaIndices(:), maxAreaIndices(:)
    integer,        intent(out), optional :: status
    type(aryStatInfoFlagT), intent(in), optional :: flagInfo

    integer :: localStatus, i
    logical :: doCheck

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    doCheck = .false.

    if (present(minAreaIndices) .and. present(maxAreaIndices)) then
      if (present(flagInfo)) then
        if (flagInfo%isAreaLowerUsed .and. flagInfo%isAreaUpperUsed) &
           doCheck = .true.
      else
           doCheck = .true.
      end if
    end if

    if (doCheck) then
      if (size(minAreaIndices) .ne. size(maxAreaIndices)) & 
         call error('bug_chkAreaIndices', 'Contact the code developer.')
      do i=1, size(minAreaIndices)
        if (minAreaIndices(i) > maxAreaIndices(i)) then
          localStatus = localStatus - 1	! Negative.
          write(errStr, '(i2,a,i8,a,i8,a)') &
             i,"-th index in the minimum area (=", minAreaIndices(i) &
             , ") is larger than the maximum (=",  maxAreaIndices(i) &
             , ")."
          call warning('inconsistentAreaIndices')
        end if
      end do
    end if

    if (present(status)) status = localStatus

  end subroutine chkAreaIndices


  !---------------------------------------------------------------------
  function convertAryStatInt16ToInt32(aryInfo) result(aryInfoInt32)
  !---------------------------------------------------------------------
  ! Return type(aryStatInfoInt32T).

    implicit none

    !integer, parameter :: rankArin = 2
    type(aryStatInfoInt32T) :: aryInfoInt32
    type(aryStatInfoInt16T), intent(in) :: aryInfo
    integer :: rankArin

    !-------------------------------------------------------------------
    ! Start:

    rankArin = size(aryInfo%minindices)
    allocate(aryInfoInt32%minindices(rankArin) &
           , aryInfoInt32%maxindices(rankArin))

    aryInfoInt32%totalsum       = aryInfo%totalsum
    aryInfoInt32%realsum        = aryInfo%realsum
    aryInfoInt32%mean           = aryInfo%mean
    aryInfoInt32%sigma          = aryInfo%sigma
    aryInfoInt32%totalentry     = aryInfo%totalentry
    aryInfoInt32%validentry     = aryInfo%validentry
    aryInfoInt32%minval         = int(aryInfo%minval, int32)
    aryInfoInt32%maxval         = int(aryInfo%maxval, int32)
    aryInfoInt32%minindices     = aryInfo%minindices
    aryInfoInt32%maxindices     = aryInfo%maxindices
    aryInfoInt32%vallower       = aryInfo%vallower
    aryInfoInt32%valupper       = aryInfo%valupper
    if (allocated(aryInfo%minareaindices)) then
      allocate(aryInfoInt32%minareaindices(rankArin))
      aryInfoInt32%minareaindices = aryInfo%minareaindices
    end if
    if (allocated(aryInfo%maxareaindices)) then
      allocate(aryInfoInt32%maxareaindices(rankArin))
      aryInfoInt32%maxareaindices = aryInfo%maxareaindices
    end if
    aryInfoInt32%flag           = aryInfo%flag

  end function convertAryStatInt16ToInt32


  !---------------------------------------------------------------------
  function convertAryStatInt8ToInt32(aryInfo) result(aryInfoInt32)
  !---------------------------------------------------------------------
  ! Return type(aryStatInfoInt32T).

    implicit none

    !integer, parameter :: rankArin = 2
    type(aryStatInfoInt32T) :: aryInfoInt32
    type(aryStatInfoInt8T), intent(in) :: aryInfo
    integer :: rankArin

    !-------------------------------------------------------------------
    ! Start:

    rankArin = size(aryInfo%minindices)
    allocate(aryInfoInt32%minindices(rankArin) &
           , aryInfoInt32%maxindices(rankArin))

    aryInfoInt32%totalsum       = aryInfo%totalsum
    aryInfoInt32%realsum        = aryInfo%realsum
    aryInfoInt32%mean           = aryInfo%mean
    aryInfoInt32%sigma          = aryInfo%sigma
    aryInfoInt32%totalentry     = aryInfo%totalentry
    aryInfoInt32%validentry     = aryInfo%validentry
    aryInfoInt32%minval         = int(aryInfo%minval, int32)
    aryInfoInt32%maxval         = int(aryInfo%maxval, int32)
    aryInfoInt32%minindices     = aryInfo%minindices
    aryInfoInt32%maxindices     = aryInfo%maxindices
    aryInfoInt32%vallower       = aryInfo%vallower
    aryInfoInt32%valupper       = aryInfo%valupper
    if (allocated(aryInfo%minareaindices)) then
      allocate(aryInfoInt32%minareaindices(rankArin))
      aryInfoInt32%minareaindices = aryInfo%minareaindices
    end if
    if (allocated(aryInfo%maxareaindices)) then
      allocate(aryInfoInt32%maxareaindices(rankArin))
      aryInfoInt32%maxareaindices = aryInfo%maxareaindices
    end if
    aryInfoInt32%flag           = aryInfo%flag

  end function convertAryStatInt8ToInt32


  !---------------------------------------------------------------------
  function getStr4AryStatInfoIntMultiDim(arInt, sTitle, fmt) result(retStr)
  !---------------------------------------------------------------------
  ! Private.  Called from printAryStatInfo*() .
  ! Return the string of Integer expression for printAryStatInfo*().
  ! Usage:
  !  character(1024) :: s
  !  s=printAryStatInfoIntMultiDim(aryInfo%minindices, 'Minimum location')
  !  write(*,'(A)') trim(s)
  ! Note: 
  !   write(*,*) trim(printAryStatInfoIntMultiDim(aryInfo%minindices, 'Minimum location'))
  ! would cause fatal error due to "Recursive I/O reference".

    implicit none

    character(NCHAR4GETSTR4ASI) :: retStr
    integer(int32), intent(in) :: arInt(:)
    character(*), intent(in) :: sTitle
    character(*), intent(in), optional :: fmt	! Default is 'I7'

    character(16) :: fm
    character(128) :: partStr
    integer :: i
    ! namelist /INTARY/ arInt	! Not allowed in Fortran90

    !-------------------------------------------------------------------
    ! Start

    if (present(fmt)) then
      fm = trim(fmt)
    else
      fm = 'I7'	! Default
    end if

    retStr = ''
!print *, 'ret010=',trim(retStr)
    write(retStr, HPASIALL) trim(sTitle)
!print *, 'ret020=',trim(retStr)
    retStr = trim(retStr)//' = ('
!print *, 'ret030=',trim(retStr)

    write(partStr, '('//trim(fm)//',)') arInt(1)
    retStr = trim(retStr)//trim(partStr)
    do i=2, size(arInt)		! Skip the loop if 1-dimensional.
      write(partStr, '(",",'//trim(fm)//')') arInt(i)
      retStr = trim(retStr)//trim(partStr)
!print *, 'ret040=',trim(retStr)
    end do

    retStr = trim(retStr)//')'
!print *, 'ret050=',trim(retStr)

    ! wholeFmt = '("'//trim(wholeFmt)//' = (",'
    ! do i=1, size(arInt)
    !   wholeFmt = trim(wholeFmt)//trim(fm)//','
    ! end do
    ! wholeFmt = trim(wholeFmt)//',")")'
    ! 
    !   write(retStr, wholeFmt) arInt
    ! ! write(retStr, wholeFmt, NML=INTARY)
    ! select case(size(arInt))
    ! case(1)
    !   write(retStr, wholeFmt) arInt(1)
    ! case(2)
    !   write(retStr, wholeFmt) arInt(1), arInt(2)
    !  !write(retStr, HPASI//'" = (",'//fm//',",",'//fm//',")")') trim(sTitle) &
    !  !   , arInt(1), arInt(2)
    ! case default
    !   call error('undefinedDimension', 'For '//trim(sTitle))
    ! end select

  end function getStr4AryStatInfoIntMultiDim


  !---------------------------------------------------------------------
  subroutine printAryStatInfoDouble(aryInfo)
  !---------------------------------------------------------------------
  ! Print AryStatInfo to STDOUT.

    implicit none

    type(aryStatInfoDoubleT), intent(in) :: aryInfo
    character(NCHAR4GETSTR4ASI) :: s

    !-------------------------------------------------------------------
    ! Start

    write(*,HPASI//'" = ", ES22.14)') "Total sum",  aryInfo%totalsum
    write(*,HPASI//'" = ", ES22.14)') "Real sum",   aryInfo%realsum
    write(*,HPASI//'" = ", ES22.14)') "Mean",       aryInfo%mean
    write(*,HPASI//'" = ", ES22.14)') "Sigma",      aryInfo%sigma
    write(*,HPASI//'" = ", I18)') "Total area", aryInfo%totalentry
    write(*,HPASI//'" = ", I18)') "Valid entry", aryInfo%validentry
    write(*,HPASI//'" = ", ES22.14)') "Minimum value", aryInfo%minval
    write(*,HPASI//'" = ", ES22.14)') "Maximum value", aryInfo%maxval
    s=getStr4AryStatInfoIntMultiDim(aryInfo%minindices, 'Minimum location')
    write(*,'(A)') trim(s)
    ! write(*,'(A)') trim(printAryStatInfoIntMultiDim(aryInfo%minindices, 'Minimum location'))	! Causes fatal error due to "Recursive I/O reference"
    ! write(*,HPASI//'" = (",I7,",",I7,")")') "Minimum location" &
    !      , aryInfo%minindices(1), aryInfo%minindices(2)
    s=getStr4AryStatInfoIntMultiDim(aryInfo%maxindices, 'Maximum location')
    write(*,'(A)') trim(s)
    ! write(*,HPASI//'" = (",I7,",",I7,")")') "Maximum location" &
    !      , aryInfo%maxindices(1), aryInfo%maxindices(2)
    write(*,HPASI//'" = ",L3)')   "Lower threshold used? (T/F)", aryInfo%flag%isVallowerUsed
    if (aryInfo%flag%isValLowerUsed) then
      write(*,HPASI//'" = ",ES22.14)')   "Lower threshold", aryInfo%vallower
    end if
    write(*,HPASI//'" = ",L3)')   "Upper threshold used? (T/F)", aryInfo%flag%isValupperUsed
    if (aryInfo%flag%isValUpperUsed) then
      write(*,HPASI//'" = ",ES22.14)')   "Upper threshold", aryInfo%valupper
    end if
    write(*,HPASI//'" = ",L3)')   "Lower area constrain used? (T/F)", aryInfo%flag%isAreaLowerUsed
    if (aryInfo%flag%isAreaLowerUsed) then
      s=getStr4AryStatInfoIntMultiDim(aryInfo%minareaindices, 'Lower area boundary')
      write(*,'(A)') trim(s)
      ! write(*,HPASI//'" = (",I7,",",I7,")")') "Lower area boundary" &
      !    , aryInfo%minareaindices(1), aryInfo%minareaindices(2)
    end if
    write(*,HPASI//'" = ",L3)')   "Upper area constrain used? (T/F)", aryInfo%flag%isAreaUpperUsed
    if (aryInfo%flag%isAreaUpperUsed) then
      s=getStr4AryStatInfoIntMultiDim(aryInfo%maxareaindices, 'Upper area boundary')
      write(*,'(A)') trim(s)
      ! write(*,HPASI//'" = (",I7,",",I7,")")') "Upper area boundary" &
      !    , aryInfo%maxareaindices(1), aryInfo%maxareaindices(2)
    end if
    write(*,HPASI//'" = ",L3)')   "External mask used? (T/F)", aryInfo%flag%isMaskUsed
    write(*,HPASI//'" = ",I5)')   "Return status (normal:0)", aryInfo%flag%status

  end subroutine printAryStatInfoDouble


  !---------------------------------------------------------------------
  subroutine printAryStatInfoSingle(aryInfo)
  !---------------------------------------------------------------------
  ! See the comments in  subroutine printAryStatInfoDouble().

    implicit none

    type(aryStatInfoSingleT), intent(in) :: aryInfo
    character(NCHAR4GETSTR4ASI) :: s

    !-------------------------------------------------------------------
    ! Start

    write(*,HPASI//'" = ", ES18.10)') "Total sum",  aryInfo%totalsum
    write(*,HPASI//'" = ", ES22.14)') "Real sum",   aryInfo%realsum
    write(*,HPASI//'" = ", ES18.10)') "Mean",       aryInfo%mean
    write(*,HPASI//'" = ", ES18.10)') "Sigma",      aryInfo%sigma
    write(*,HPASI//'" = ", I18)') "Total area", aryInfo%totalentry
    write(*,HPASI//'" = ", I18)') "Valid entry", aryInfo%validentry
    write(*,HPASI//'" = ", ES18.10)') "Minimum value", aryInfo%minval
    write(*,HPASI//'" = ", ES18.10)') "Maximum value", aryInfo%maxval
    s=getStr4AryStatInfoIntMultiDim(aryInfo%minindices, 'Minimum location')
    write(*,'(A)') trim(s)
    s=getStr4AryStatInfoIntMultiDim(aryInfo%maxindices, 'Maximum location')
    write(*,'(A)') trim(s)
    write(*,HPASI//'" = ",L3)')   "Lower threshold used? (T/F)", aryInfo%flag%isVallowerUsed
    if (aryInfo%flag%isValLowerUsed) then
      write(*,HPASI//'" = ",ES18.10)')   "Lower threshold", aryInfo%vallower
    end if
    write(*,HPASI//'" = ",L3)')   "Upper threshold used? (T/F)", aryInfo%flag%isValupperUsed
    if (aryInfo%flag%isValUpperUsed) then
      write(*,HPASI//'" = ",ES18.10)')   "Upper threshold", aryInfo%valupper
    end if
    write(*,HPASI//'" = ",L3)')   "Lower area constrain used? (T/F)", aryInfo%flag%isAreaLowerUsed
    if (aryInfo%flag%isAreaLowerUsed) then
      s=getStr4AryStatInfoIntMultiDim(aryInfo%minareaindices, 'Lower area boundary')
      write(*,'(A)') trim(s)
    end if
    write(*,HPASI//'" = ",L3)')   "Upper area constrain used? (T/F)", aryInfo%flag%isAreaUpperUsed
    if (aryInfo%flag%isAreaUpperUsed) then
      s=getStr4AryStatInfoIntMultiDim(aryInfo%maxareaindices, 'Upper area boundary')
      write(*,'(A)') trim(s)
    end if
    write(*,HPASI//'" = ",L3)')   "External mask used? (T/F)", aryInfo%flag%isMaskUsed
    write(*,HPASI//'" = ",I5)')   "Return status (normal:0)", aryInfo%flag%status

  end subroutine printAryStatInfoSingle


  !---------------------------------------------------------------------
  subroutine printAryStatInfoInt32(aryInfo)
  !---------------------------------------------------------------------
  ! See the comments in  subroutine printAryStatInfoDouble().

    implicit none

    type(aryStatInfoInt32T), intent(in) :: aryInfo
    character(NCHAR4GETSTR4ASI) :: s

    !-------------------------------------------------------------------
    ! Start

    write(*,HPASI//'" = ", I18)') "Total sum",  aryInfo%totalsum
    write(*,HPASI//'" = ", ES22.14)') "Real sum",   aryInfo%realsum
    write(*,HPASI//'" = ", ES18.10)') "Mean",       aryInfo%mean
    write(*,HPASI//'" = ", ES18.10)') "Sigma",      aryInfo%sigma
    write(*,HPASI//'" = ", I18)') "Total area", aryInfo%totalentry
    write(*,HPASI//'" = ", I18)') "Valid entry", aryInfo%validentry
    write(*,HPASI//'" = ", I18)') "Minimum value", aryInfo%minval
    write(*,HPASI//'" = ", I18)') "Maximum value", aryInfo%maxval
    s=getStr4AryStatInfoIntMultiDim(aryInfo%minindices, 'Minimum location')
    write(*,'(A)') trim(s)
    s=getStr4AryStatInfoIntMultiDim(aryInfo%maxindices, 'Maximum location')
    write(*,'(A)') trim(s)
    write(*,HPASI//'" = ",L3)')   "Lower threshold used? (T/F)", aryInfo%flag%isVallowerUsed
    if (aryInfo%flag%isValLowerUsed) then
      write(*,HPASI//'" = ",ES18.10)')   "Lower threshold", aryInfo%vallower
    end if
    write(*,HPASI//'" = ",L3)')   "Upper threshold used? (T/F)", aryInfo%flag%isValupperUsed
    if (aryInfo%flag%isValUpperUsed) then
      write(*,HPASI//'" = ",ES18.10)')   "Upper threshold", aryInfo%valupper
    end if
    write(*,HPASI//'" = ",L3)')   "Lower area constrain used? (T/F)", aryInfo%flag%isAreaLowerUsed
    if (aryInfo%flag%isAreaLowerUsed) then
      s=getStr4AryStatInfoIntMultiDim(aryInfo%minareaindices, 'Lower area boundary')
      write(*,'(A)') trim(s)
    end if
    write(*,HPASI//'" = ",L3)')   "Upper area constrain used? (T/F)", aryInfo%flag%isAreaUpperUsed
    if (aryInfo%flag%isAreaUpperUsed) then
      s=getStr4AryStatInfoIntMultiDim(aryInfo%maxareaindices, 'Upper area boundary')
      write(*,'(A)') trim(s)
    end if
    write(*,HPASI//'" = ",L3)')   "External mask used? (T/F)", aryInfo%flag%isMaskUsed
    write(*,HPASI//'" = ",I5)')   "Return status (normal:0)", aryInfo%flag%status

  end subroutine printAryStatInfoInt32


  !---------------------------------------------------------------------
  subroutine printAryStatInfoInt16(aryInfo)
  !---------------------------------------------------------------------
  ! See the comments in  subroutine printAryStatInfoDouble().

    implicit none

    type(aryStatInfoInt16T), intent(in) :: aryInfo

    call printAryStatInfoInt32(convertAryStatInt16ToInt32(aryInfo))

  end subroutine printAryStatInfoInt16


  !---------------------------------------------------------------------
  subroutine printAryStatInfoInt8(aryInfo)
  !---------------------------------------------------------------------
  ! See the comments in  subroutine printAryStatInfoDouble().

    implicit none

    type(aryStatInfoInt8T), intent(in) :: aryInfo

    call printAryStatInfoInt32(convertAryStatInt8ToInt32(aryInfo))

  end subroutine printAryStatInfoInt8


  !---------------------------------------------------------------------
  function getAnnularMaskAry(sizeX, sizeY, centX, centY &
     , rOuter, rInner) result(arMask)
  !---------------------------------------------------------------------
  ! Return Logical arMask(sizeX,sizeY) (allocatable), where
  !  the pixels at (rInner <= radius < rOuter) are True.
  ! NOTE: make sure to deallocate the returned array after use.
  ! TODO: deal with 3-dimensional array!

    implicit none
    integer, parameter :: rankArin = 2

    logical, allocatable :: arMask(:,:)	! result
    integer(int32), intent(in) :: sizeX, sizeY
    real(double),   intent(in) :: centX, centY, rOuter
    real(double),   intent(in), optional :: rInner	! 0 in default.

    real(double) :: rInner2, rOuter2, dist2
    integer(int32) :: arySize(rankArin), i, j

    !-------------------------------------------------------------------
    ! Start

    rOuter2 = rOuter**2
    rInner2 = -1	! Default: Should be 0, but just to be safe at the boundary.
    if (present(rInner)) then
      if (rInner > 0) then
        rInner2 = rInner**2
      end if
    end if

    arySize(1) = sizeX
    arySize(2) = sizeY
    allocate(arMask(arySize(1), arySize(2)))
    arMask = .false.

    ! ------- Main -------
    do   i=max(int(centX-rOuter-1),1), min(int(centX+rOuter+1),arySize(1))
      do j=max(int(centY-rOuter-1),1), min(int(centY+rOuter+1),arySize(2))
        dist2 = (i-centX)**2 + (j-centY)**2
        if ((dist2 < rOuter2) .and. (dist2 >= rInner2)) then
          arMask(i,j) = .true.
        end if
      end do
    end do

  end function getAnnularMaskAry

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module array_utils
